(*
 * Copyright (C) 2013 National ICT Australia Ltd ('NICTA')
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Author: David Cock - David.Cock@nicta.com.au *)

header "The Monty Hall Problem"

theory Monty imports "../pGCL/pGCL" "../pGCL/pVCG" begin

section {* The Game *}

text {*
The setting is a game show:  There is a prize hidden behind one
of three doors, and the contestant is invited to choose one.  Once
the guess is made, the host than opens one of the remaining two 
doors, to reveal a goat, showing that the prize is elsewhere.  The
contestent is then given the choice of switching their guess to the
other unopened door, or sticking to their first guess.

The puzzle is whether the contestant is better of switching, or
sticking; or whether it makes a difference at all.  Most people's
intuition suggests that it make no difference, whereas, in fact,
switching raises the chance of success from 1/3 to 2/3.
*}

subsection {* State Space *}

text {*
The game state consists of a prize location, a guess, and a clue
(the door the host opens).  These are not constrained to the range
{1,2,3}:  We instead provide an invariant that this constraint
occurs.
*}

record game =
  prize  :: nat
 "guess" :: nat
  clue   :: nat

definition player_wins :: "game \<Rightarrow> bool"
where "player_wins g \<equiv> guess g = prize g"

subsection "Invariants"

definition inv_doors :: "game \<Rightarrow> bool"
where "inv_doors g \<equiv> {guess g, clue g, prize g} \<subseteq> {1,2,3}"

definition inv_prize :: "game \<Rightarrow> bool"
where "inv_prize g \<equiv> prize g \<in> {1,2,3}"

definition inv_clue :: "game \<Rightarrow> bool"
where "inv_clue g \<equiv> clue g \<in> {1,2,3}"

definition inv_guess :: "game \<Rightarrow> bool"
where "inv_guess g \<equiv> guess g \<in> {1,2,3}"

subsection "The Game"

definition hide_behind :: "nat \<Rightarrow> game prog"
where "hide_behind D \<equiv> Apply (prize_update (\<lambda>x. D))"

definition guess_behind :: "nat \<Rightarrow> game prog"
where "guess_behind D \<equiv> Apply (guess_update (\<lambda>x. D))"

definition open_door :: "nat \<Rightarrow> game prog"
where "open_door D \<equiv> Apply (clue_update (\<lambda>x. D))"

definition hide_prize :: "game prog"
where "hide_prize \<equiv> hide_behind 1 \<Sqinter> hide_behind 2 \<Sqinter> hide_behind 3"

definition make_guess :: "game prog"
where "make_guess \<equiv> guess_behind 1 \<^bsub>(\<lambda>s. 1/3)\<^esub>\<oplus>
                    guess_behind 2 \<^bsub>(\<lambda>s. 1/2)\<^esub>\<oplus> guess_behind 3"

definition reveal :: "game prog"
where "reveal \<equiv> \<Sqinter>d\<in>(\<lambda>s. {1,2,3} - {prize s, guess s}). open_door d"

definition switch_guess :: "game prog"
where "switch_guess \<equiv> \<Sqinter>d\<in>(\<lambda>s. {1,2,3} - {clue s, guess s}). guess_behind d"

definition monty :: "bool \<Rightarrow> game prog"
where
  "monty switch \<equiv> hide_prize ;;
                  make_guess ;;
                  reveal ;;
                  (if switch then switch_guess else Skip)"

section "A Brute Force Approach"

text {*
For simple programs, we can show the exact WP equation by unfolding.
*}

lemma eval_win[simp]:
  "p = g \<Longrightarrow> \<guillemotleft>player_wins\<guillemotright> (s\<lparr> prize := p, guess := g, clue := c \<rparr>) = 1"
  by(simp add:embed_bool_def player_wins_def)

lemma eval_loss[simp]:
  "p \<noteq> g \<Longrightarrow> \<guillemotleft>player_wins\<guillemotright> (s\<lparr> prize := p, guess := g, clue := c \<rparr>) = 0"
  by(simp add:embed_bool_def player_wins_def)

text {* If they stick to their guns, the player wins with p=1/3 *}
lemma wp_monty_noswitch:
  "(\<lambda>s. 1/3) = wp (monty False) \<guillemotleft>player_wins\<guillemotright>"
  unfolding monty_def hide_prize_def make_guess_def reveal_def
            hide_behind_def guess_behind_def open_door_def
            switch_guess_def
  by(simp add:wp_eval insert_Diff_if o_def)

lemma swap_upd:
  "s\<lparr> prize := p, clue := c, guess := g \<rparr> =
   s\<lparr> prize := p, guess := g, clue := c \<rparr>"
  by(simp)

text {* If they switch, they win with p=2/3.  Brute force here
        takes longer, but it's still possible.  Any bigger and this
        wouldn't work.  *}

lemma wp_monty_switch_bruteforce:
  "(\<lambda>s. 2/3) = wp (monty True) \<guillemotleft>player_wins\<guillemotright>"
  unfolding monty_def hide_prize_def make_guess_def reveal_def
            hide_behind_def guess_behind_def open_door_def
            switch_guess_def
  -- "Note that this is getting slow"
  by(simp add:wp_eval insert_Diff_if swap_upd o_def)

section {* A Modular Approach *}

text {*
Here, we break the problem up, and prove an annotation for each step.
*}

subsection {* Healthiness *}

text {* We first need healthiness for each step. *}

lemma healthy_wp_hide_prize:
  "healthy (wp hide_prize)"
  unfolding hide_prize_def hide_behind_def
  by(simp add:healthy_intros)

lemma healthy_wp_make_guess:
  "healthy (wp make_guess)"
  unfolding make_guess_def guess_behind_def
  by(simp add:healthy_intros bounded_by_const)

lemma healthy_wp_reveal:
  "healthy (wp reveal)"
  unfolding reveal_def
  apply(rule healthy_wp_SetDC)
   unfolding open_door_def apply(rule healthy_wp_Apply)
  apply(simp add:insert_Diff_if)
  apply(safe, blast+)
  done

lemma healthy_wp_switch_guess:
  "healthy (wp switch_guess)"
  unfolding switch_guess_def guess_behind_def
  apply(rule healthy_wp_SetDC)
   apply(simp add:healthy_intros)
  apply(simp add:insert_Diff_if)
  apply(safe, blast+)
  done

lemmas monty_healthy =
  healthy_wp_switch_guess healthy_wp_reveal
  healthy_wp_make_guess healthy_wp_hide_prize

subsection "Annotations for the four phases"

text {* @{term hide_prize} establishes the prize invariant. *}
lemma wp_hide_prize:
  "(\<lambda>s. 1) \<tturnstile> wp hide_prize \<guillemotleft>inv_prize\<guillemotright>"
  unfolding hide_prize_def hide_behind_def
  unfolding wp_eval o_def
  by(simp add:embed_bool_def inv_prize_def)

text {* Given the prize invariant, @{term make_guess} establishes the
 guess invariant and guesses wrongly with probability at least 2/3. *}
lemma wp_make_guess:
  "(\<lambda>s. 2/3 * \<guillemotleft>\<lambda>g. inv_prize g\<guillemotright> s) \<tturnstile>
   wp make_guess \<guillemotleft>\<lambda>g. guess g \<noteq> prize g \<and> inv_prize g \<and> inv_guess g\<guillemotright>"
  unfolding make_guess_def guess_behind_def
  unfolding wp_eval o_def
  apply(rule le_funI)
  apply(simp add:embed_bool_def inv_prize_def inv_guess_def)
  done

lemma last_one:
  "\<lbrakk> a \<noteq> b; a \<in> {1::nat,2,3}; b \<in> {1,2,3} \<rbrakk> \<Longrightarrow> \<exists>!c. {1,2,3} - {b,a} = {c}"
  apply(simp)
  apply(case_tac "a=1", simp_all)
   apply(case_tac "b=2", simp_all add:insert_Diff_if)
  apply(case_tac "a=2", simp_all)
   apply(case_tac "b=1", simp_all add:insert_Diff_if)
  apply(case_tac "b=1", simp_all)
  done

text {* Given the previous invariants, and an incorrect guess, @{term reveal}
  will give a clue that is neither the prize, nor the guess. *}
lemma wp_reveal:
  "\<guillemotleft>\<lambda>g. guess g \<noteq> prize g \<and> inv_prize g \<and> inv_guess g\<guillemotright> \<tturnstile>
   wp reveal \<guillemotleft>\<lambda>g. guess g \<noteq> prize g \<and>
                  clue g \<noteq> prize g \<and>
                  clue g \<noteq> guess g \<and>
                  inv_prize g \<and> inv_guess g \<and> inv_clue g\<guillemotright>"
  (is "?X \<tturnstile> wp reveal ?Y")
  -- "use_premise gives you the LHS of a predicate implication as an assumption"
proof(safe intro!:use_premise healthy_wp_reveal)
  fix s
  assume "guess s \<noteq> prize s"
     and "inv_prize s"
     and "inv_guess s"
  moreover then obtain c
    where singleton: "{Suc 0,2,3} -  {prize s, guess s} = {c}"
      and "c \<noteq> prize s"
      and "c \<noteq> guess s"
      and "c \<in> {Suc 0,2,3}"
    unfolding inv_prize_def inv_guess_def
    by(force dest:last_one elim!:ex1E)
  ultimately show "1 \<le> wp reveal ?Y s"
    by(simp add:reveal_def open_door_def wp_eval singleton o_def
                embed_bool_def inv_prize_def inv_guess_def inv_clue_def)
qed

lemma distinct_game:
  "\<lbrakk> guess g \<noteq> prize g; clue g \<noteq> prize g; clue g \<noteq> guess g;
     inv_prize g; inv_guess g; inv_clue g \<rbrakk> \<Longrightarrow> 
   {1, 2, 3} = {guess g, prize g, clue g}"
  unfolding inv_prize_def inv_guess_def inv_clue_def
  by(safe intro!:set_eqI, (simp, metis?)+)

text {* Given the invariants, switching the wrong guess gives
the right one. *}
lemma wp_switch_guess:
  "\<guillemotleft>\<lambda>g. guess g \<noteq> prize g \<and> clue g \<noteq> prize g \<and> clue g \<noteq> guess g \<and>
        inv_prize g \<and> inv_guess g \<and> inv_clue g\<guillemotright> \<tturnstile>
   wp switch_guess \<guillemotleft>player_wins\<guillemotright>"
  apply(safe intro!:use_premise healthy_wp_switch_guess)
  unfolding switch_guess_def guess_behind_def wp_eval
  apply(subst distinct_game, assumption+)
  apply(simp add:insert_Diff_if)
  apply(case_tac s, simp add:player_wins_def)
  done

text {* Given componentwise specifications, we can glue them together
with calculational reasoning to get our result. *}
lemma wp_monty_switch_modular:
  "(\<lambda>s. 2/3) \<tturnstile> wp (monty True) \<guillemotleft>player_wins\<guillemotright>"
proof(rule wp_validD)  -- "Work in probabilistic Hoare triples"
  note wp_validI[OF wp_scale, OF wp_hide_prize, simplified]
    -- "Here we apply scaling to get our pre-expectation"
  also note wp_validI[OF wp_make_guess]
  also note wp_validI[OF wp_reveal]
  also note wp_validI[OF wp_switch_guess]
  finally show "\<lbrace>\<lambda>s. 2/3\<rbrace> monty True \<lbrace>\<guillemotleft>player_wins\<guillemotright>\<rbrace>p"
    unfolding monty_def
    by(simp add:healthy_intros sound_intros monty_healthy)
qed

text {* Using the VCG *}

lemmas scaled_hide = wp_scale[OF wp_hide_prize, simplified]
declare scaled_hide[wp] wp_make_guess[wp] wp_reveal[wp] wp_switch_guess[wp]
declare healthy_wp_hide_prize[health] healthy_wp_make_guess[health]
        healthy_wp_reveal[health] healthy_wp_switch_guess[health]

text {* Alternatively, the VCG will get this using the same annotations. *}
lemma wp_monty_switch_vcg:
  "(\<lambda>s. 2/3) \<tturnstile> wp (monty True) \<guillemotleft>player_wins\<guillemotright>"
  unfolding monty_def
  by(simp, pvcg)

end
