(*
 * Copyright (C) 2013 National ICT Australia Ltd ('NICTA')
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Author: David Cock - David.Cock@nicta.com.au *)

header "Complete Sets"

theory CompleteSet imports RComplete begin

section {* Least upper, and greatest lower bounds on complete
           (ordered) sets *}

text {* This is an abstraction of the completeness property of the reals.
        It also applies to any closed subset thereof. *}

class up_complete = order +
  assumes up_complete:
    "\<lbrakk> \<exists>X. X \<in> S; \<exists>Y. isUb UNIV S Y \<rbrakk> \<Longrightarrow> \<exists>!t. isLub UNIV S t"

text {* The reals are up-complete. *}
instantiation real :: up_complete begin
instance by(intro_classes, blast intro:real_isLub_unique reals_complete)
end

class lub = order +
  fixes lub :: "'a set \<Rightarrow> 'a"
  assumes lub_greater:  "\<lbrakk> a \<in> A; \<exists>b. \<forall>a\<in>A. a \<le> b \<rbrakk> \<Longrightarrow> a \<le> lub A"
      and lub_least:    "\<lbrakk> \<exists>a. a \<in> A; \<forall>a\<in>A. a \<le> b \<rbrakk> \<Longrightarrow> lub A \<le> b"

text {* The reals are complete *}
instantiation real :: lub begin

definition
  lub_real :: "real set \<Rightarrow> real"
where
  "lub_real S \<equiv> THE t. isLub UNIV S t"

instance proof
  fix a::real and A
  assume ain: "a \<in> A" and bound: "\<exists>b. \<forall>a\<in>A. a \<le> b"
  thus "a \<le> lub A"
    unfolding lub_real_def
    apply -
    apply(rule ex1E[OF up_complete], force)
     apply(force intro:isUbI setleI)
    apply(rule theI2, assumption, simp)
    apply(force intro:isUbD isLub_isUb)
    done
next
  fix b::real and A
  assume "\<exists>a. a \<in> A" and "\<forall>a\<in>A. a \<le> b"
  thus lub_least: "lub A \<le> b"
    unfolding lub_real_def
    apply -
    apply(rule ex1E[OF up_complete], assumption)
     apply(force intro:isUbI setleI)
    apply(rule theI2, assumption, simp)
    apply(force intro:isLub_le_isUb isUbI setleI)
    done
qed
end

class glb = order +
  fixes glb :: "'a set \<Rightarrow> 'a"
  assumes glb_lower:    "\<lbrakk> a \<in> A; \<exists>b. \<forall>a\<in>A. b \<le> a \<rbrakk> \<Longrightarrow> glb A \<le> a"
      and glb_greatest: "\<lbrakk> \<exists>a. a \<in> A; \<forall>a\<in>A. b \<le> a \<rbrakk> \<Longrightarrow> b \<le> glb A"

instantiation real :: glb begin

text {* The glb is defined in terms of the lub *}
definition
  glb_real :: "real set \<Rightarrow> real"
where
  "glb_real S \<equiv> lub {x. \<forall>s\<in>S. x \<le> s}"

instance proof
  fix a::real and A
  assume "a \<in> A" and "\<exists>b. \<forall>a\<in>A. b \<le> a"
  thus "glb A \<le> a"
    unfolding glb_real_def
    by(force intro:lub_least)

  next
    fix b::real and A
    assume "\<exists>a. a \<in> A" and "\<forall>a\<in>A. b \<le> a"
    thus "b \<le> glb A"
      unfolding glb_real_def
      by(force intro:lub_greater)
  qed
end

class complete = lub + glb

instantiation real :: complete begin
instance proof qed
end

lemma lub_const[simp]:
  "\<exists>x. x \<in> S \<Longrightarrow> lub ((\<lambda>x. c) ` S) = c"
  by(force intro:antisym lub_least lub_greater)

lemma lub_singleton[simp,intro!]:
  "lub {c} = c"
  by(blast intro:lub_greater lub_least antisym)

lemma lub_two_right[simp,intro]:
  "a \<le> b \<Longrightarrow> lub {a,b} = b"
  by(blast intro:antisym lub_least lub_greater)

lemma lub_two_comm[ac_simps]:
  "lub {a,b} = lub {b,a}"
  by(blast intro:arg_cong)

lemma lub_up_closed[simp]:
  "lub {x. x \<le> c} = c"
  by(force intro:antisym lub_least lub_greater)

lemma glb_const[simp]:
  "\<exists>x. x \<in> S \<Longrightarrow> glb ((\<lambda>x. c) ` S) = c"
  by(force intro:antisym glb_lower glb_greatest)

lemma glb_singleton[simp,intro!]:
  "glb {c} = c"
  by(blast intro:glb_lower glb_greatest antisym)

lemma glb_two_left[simp,intro]:
  "a \<le> b \<Longrightarrow> glb {a,b} = a"
  by(blast intro:antisym glb_lower glb_greatest)

lemma glb_two_comm[ac_simps]:
  "glb {a,b} = glb {b,a}"
  by(blast intro:arg_cong)

lemma glb_down_closed[simp]:
  "glb {x. c \<le> x} = c"
  by(force intro:antisym glb_lower glb_greatest)

lemma glb_under_glb:
  assumes ne:"\<exists>x. x\<in>X"
      and find_lower: "\<forall>x\<in>X. \<exists>y\<in>Y. y \<le> x"
      and bounded: "\<exists>b. \<forall>y\<in>Y. b \<le> y"
  shows "glb Y \<le> glb X"
proof(rule glb_greatest)
  from ne show "\<exists>x. x\<in>X" .

  show "\<forall>x\<in>X. glb Y \<le> x"
  proof
    fix x assume xin: "x \<in> X"
    with find_lower obtain y where yin: "y \<in> Y" and yle: "y \<le> x"
      by(blast)
    from yin and bounded have "glb Y \<le> y" by(rule glb_lower)
    also note yle
    finally show "glb Y \<le> x" .
  qed
qed

text {* Functions from any set into a complete set (in particular,
        the reals) form a complete set. *}
instantiation "fun" :: (type,complete) complete begin

text {* Take the upper (lower) bound pointwise over the domain. *}
definition
  lub_fun :: "('a \<Rightarrow> 'b::complete) set \<Rightarrow> 'a \<Rightarrow> 'b"
where
  "lub_fun S \<equiv> \<lambda>s. lub {f s |f. f \<in> S}"

definition
  glb_fun :: "('a \<Rightarrow> 'b::complete) set \<Rightarrow> 'a \<Rightarrow> 'b"
where
  "glb_fun S \<equiv> \<lambda>s. glb {f s |f. f \<in> S}"

instance proof
  fix a::"'a \<Rightarrow> 'b" and A
  assume "a \<in> A" and "\<exists>b. \<forall>a\<in>A. b \<le> a"
  thus "glb A \<le> a"
    unfolding glb_fun_def
    by(blast intro:le_funI glb_lower dest:le_funD)

next
  fix b::"'a \<Rightarrow> 'b" and A
  assume "\<exists>a. a \<in> A" and "\<forall>a\<in>A. b \<le> a"
  thus "b \<le> glb A"
    unfolding glb_fun_def
    by(blast intro:le_funI glb_greatest dest:le_funD)

next
  fix a::"'a \<Rightarrow> 'b" and A
  assume "a \<in> A" and "\<exists>b. \<forall>a\<in>A. a \<le> b"
  thus "a \<le> lub A"
    unfolding lub_fun_def
    by(blast intro:le_funI lub_greater dest:le_funD)

next
  fix b::"'a \<Rightarrow> 'b" and A
  assume "\<exists>a. a \<in> A" and "\<forall>a\<in>A. a \<le> b"
  thus "lub A \<le> b"
    unfolding lub_fun_def
    by(blast intro:le_funI lub_least dest:le_funD)
qed
end

end
