(*
 * Copyright (C) 2013 National ICT Australia Ltd ('NICTA')
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Author: David Cock - David.Cock@nicta.com.au *)

header "Projective Data Refinement"

theory DataRefinement imports pGCL begin

text {* A projective data refinement construction for pGCL.  By
  projective, we mean that the abstract state is always a function
  (\<phi>) of the concrete state.  Refinement may be predicated (G) on
  the state. *}

definition
  drefines :: "('b \<Rightarrow> 'a) \<Rightarrow> ('b \<Rightarrow> bool) \<Rightarrow> 'a prog \<Rightarrow> 'b prog \<Rightarrow> bool"
where
  "drefines \<phi> G A B \<equiv> \<forall>P Q. (sound P \<and> bounded_by 1 P \<and>
                             sound Q \<and> bounded_by 1 Q \<and>
                             (P \<tturnstile> wp A Q)) \<longrightarrow>
                            (\<guillemotleft>G\<guillemotright> && (P o \<phi>) \<tturnstile> wp B (Q o \<phi>))"

lemma drefinesD[dest]:
  "\<lbrakk> drefines \<phi> G A B; sound P; bounded_by 1 P;
     sound Q; bounded_by 1 Q; P \<tturnstile> wp A Q \<rbrakk> \<Longrightarrow>
   \<guillemotleft>G\<guillemotright> && (P o \<phi>) \<tturnstile> wp B (Q o \<phi>)"
  unfolding drefines_def by(blast)

text {* Use G as an assumption *}
lemma drefinesD2:
  assumes dr:  "drefines \<phi> G A B"
      and sP:  "sound P"
      and bP:  "bounded_by 1 P"
      and sQ:  "sound Q"
      and bQ:  "bounded_by 1 Q"
      and wpA: "P \<tturnstile> wp A Q"
      and G:   "G s"
  shows "(P o \<phi>) s \<le> wp B (Q o \<phi>) s"
proof -
  from sP have "0 \<le> (P o \<phi>) s" unfolding o_def by(blast)
  with G have "(P o \<phi>) s = (\<guillemotleft>G\<guillemotright> && (P o \<phi>)) s"
    by(simp add:exp_conj_def)
  also from assms have "... \<le> wp B (Q o \<phi>) s" by(blast)
  finally show "(P o \<phi>) s \<le> ..." .
qed

lemma drefinesD3:
  assumes dr: "drefines \<phi> G a b"
      and G:  "G s"
      and sQ: "sound Q"
      and bQ: "bounded_by 1 Q"
      and ha: "healthy (wp a)"
  shows "wp a Q (\<phi> s) \<le> wp b (Q o \<phi>) s"
proof -
  let "?L s'" = "wp a Q s'"
  from sQ and ha have sL: "sound ?L" by(blast)
  from sQ and bQ and ha have bL: "bounded_by 1 ?L" by(blast)

  have "?L \<tturnstile> ?L" by(simp)
  with sL and bL and assms
  show ?thesis
    by(blast intro:drefinesD2[OF dr, where P="?L", simplified])
qed

lemma drefinesI[intro]:
  "\<lbrakk> \<And>P Q. \<lbrakk> sound P; bounded_by 1 P;
             sound Q; bounded_by 1 Q; P \<tturnstile> wp A Q \<rbrakk> \<Longrightarrow>
           \<guillemotleft>G\<guillemotright> && (P o \<phi>) \<tturnstile> wp B (Q o \<phi>) \<rbrakk> \<Longrightarrow>
   drefines \<phi> G A B"
  unfolding drefines_def by(blast)

text {* Use G as an assumption *}
lemma drefinesI2:
  fixes   A::"'a prog"
    and   B::"'b prog"
    and   \<phi>::"'b \<Rightarrow> 'a"
    and   G::"'b \<Rightarrow> bool"
  assumes fB: "feasible (wp B)"
      and withAs:
        "\<And>P Q s. \<lbrakk> sound P; bounded_by 1 P; sound Q; bounded_by 1 Q;
                 G s; P \<tturnstile> wp A Q \<rbrakk> \<Longrightarrow> (P o \<phi>) s \<le> wp B (Q o \<phi>) s"
  shows "drefines \<phi> G A B"
proof
  fix P and Q
  assume sP:  "sound P"
     and bP:  "bounded_by 1 P"
     and sQ:  "sound Q"
     and bQ:  "bounded_by 1 Q"
     and wpA: "P \<tturnstile> wp A Q"

  hence "\<And>s. G s \<Longrightarrow> (P o \<phi>) s \<le> wp B (Q o \<phi>) s"
    using withAs by(blast)
  moreover
  from sQ have "sound (Q o \<phi>)"
    unfolding o_def by(blast)
  moreover
  from bQ have "bounded_by 1 (Q o \<phi>)"
    unfolding o_def by(blast)
  moreover
  from sP have "sound (P o \<phi>)"
    unfolding o_def by(blast)
  moreover
  from bP have "bounded_by 1 (P o \<phi>)"
    unfolding o_def by(blast)
  ultimately
  show "\<guillemotleft>G\<guillemotright> && (P o \<phi>) \<tturnstile> wp B (Q o \<phi>)"
    using fB by(blast intro:entails_pconj_assumption)
qed

section {* Transitivity of Data Refinement *}

text {* Data refinement is transitive, and composes with
  program refinement. *}

lemma dr_trans[trans]:
  fixes A::"'a prog" and B::"'b prog" and C::"'c prog"
  assumes drAB: "drefines \<phi> G A B"
      and drBC: "drefines \<phi>' G' B C"
      and Gimp: "\<And>s. G' s \<Longrightarrow> G (\<phi>' s)"
  shows "drefines (\<phi> o \<phi>') G' A C"
proof(rule drefinesI)
  fix P::"'a \<Rightarrow> real" and Q::"'a \<Rightarrow> real" and s::'a
  assume sP: "sound P" and sQ: "sound Q"
     and bP: "bounded_by 1 P" and bQ: "bounded_by 1 Q"
     and wpA: "P \<tturnstile> wp A Q"

  have "\<guillemotleft>G'\<guillemotright> && \<guillemotleft>G o \<phi>'\<guillemotright> = \<guillemotleft>G'\<guillemotright>"
  proof(rule ext, unfold exp_conj_def)
    fix x
    show "\<guillemotleft>G'\<guillemotright> x .& \<guillemotleft>G o \<phi>'\<guillemotright> x = \<guillemotleft>G'\<guillemotright> x" (is ?X)
    proof(cases "G' x")
      case False then show ?X by(simp)
    next
      case True
      moreover
      with Gimp have "(G o \<phi>') x" by(simp add:o_def)
      ultimately
      show ?X by(simp)
    qed
  qed

  with sP and bP
  have "\<guillemotleft>G'\<guillemotright> && (P o (\<phi> o \<phi>')) = \<guillemotleft>G'\<guillemotright> && ((\<guillemotleft>G\<guillemotright> && (P o \<phi>)) o \<phi>')"
    by(simp add:exp_conj_assoc o_assoc)

  also {
    from sP and sQ and bP and bQ and wpA and drAB
    have "\<guillemotleft>G\<guillemotright> && (P o \<phi>) \<tturnstile> wp B (Q o \<phi>)"
      by(blast intro:drefinesD)

    with drBC and sP and sQ and bP and bQ
    have "\<guillemotleft>G'\<guillemotright> && ((\<guillemotleft>G\<guillemotright> && (P o \<phi>)) o \<phi>') \<tturnstile> wp C ((Q o \<phi>) o \<phi>')"
      by(blast intro:drefinesD)
  }

  finally
  show "\<guillemotleft>G'\<guillemotright> && (P o (\<phi> o \<phi>')) \<tturnstile> wp C (Q o (\<phi> o \<phi>'))"
    by(simp add:o_assoc)
qed

lemma pr_dr_trans[trans]:
  assumes prAB: "A \<sqsubseteq> B"
      and drBC: "drefines \<phi> G B C"
  shows "drefines \<phi> G A C"
proof(rule drefinesI)
  fix P and Q
  assume sP:  "sound P"
     and bP:  "bounded_by 1 P"
     and sQ:  "sound Q"
     and bQ:  "bounded_by 1 Q"
     and wpA: "P \<tturnstile> wp A Q"

  note wpA
  also from sQ and prAB have "wp A Q \<tturnstile> wp B Q" by(blast)
  finally have "P \<tturnstile> wp B Q" .
  with sP and bP and sQ and bQ and drBC
  show "\<guillemotleft>G\<guillemotright> && (P o \<phi>) \<tturnstile> wp C (Q o \<phi>)" by(blast intro:drefinesD)
qed

lemma dr_pr_trans[trans]:
  assumes drAB: "drefines \<phi> G A B"
  assumes prBC: "B \<sqsubseteq> C"
  shows "drefines \<phi> G A C"
proof(rule drefinesI)
  fix P and Q
  assume sP:  "sound P"
     and bP:  "bounded_by 1 P"
     and sQ:  "sound Q"
     and bQ:  "bounded_by 1 Q"
     and wpA: "P \<tturnstile> wp A Q"

  with drAB have "\<guillemotleft>G\<guillemotright> && (P o \<phi>) \<tturnstile> wp B (Q o \<phi>)" by(blast intro:drefinesD)
  also from sQ and prBC have "... \<tturnstile> wp C (Q o \<phi>)" by(blast)
  finally show "\<guillemotleft>G\<guillemotright> && (P o \<phi>) \<tturnstile> ..." .
qed

section {* Structural Refinement Rules *}

lemma dr_Seq:
  assumes drAB: "drefines \<phi> P A B"
      and drBC: "drefines \<phi> Q C D"
      and wpB:  "\<guillemotleft>P\<guillemotright> \<tturnstile> wp B \<guillemotleft>Q\<guillemotright>"
      and hB:   "healthy (wp B)"
      and hC:   "healthy (wp C)"
      and hD:   "healthy (wp D)"
      and scB:  "sub_conj (wp B)"
  shows "drefines \<phi> P (A;;C) (B;;D)"
proof
  fix R and S
  assume sR: "sound R" and bR: "bounded_by 1 R"
     and sS: "sound S" and bS: "bounded_by 1 S"
     and wpAC: "R \<tturnstile> wp (A;;C) S"

  from  sR and bR
  have "\<guillemotleft>P\<guillemotright> && (R o \<phi>) = \<guillemotleft>P\<guillemotright> && (\<guillemotleft>P\<guillemotright> && (R o \<phi>))"
    by(simp add:exp_conj_assoc)

  also {
    from drAB and hC and sR and bR and sS and bS
     and wpAC[unfolded eval_wp_Seq o_def]
    have "\<guillemotleft>P\<guillemotright> && (R o \<phi>) \<tturnstile> wp B (wp C S o \<phi>)"
      apply -
      by(rule drefinesD,
            (blast dest:healthy_bounded_byD)+)
    with wpB and hC and sS and bS and scB
    have "\<guillemotleft>P\<guillemotright> && (\<guillemotleft>P\<guillemotright> && (R o \<phi>)) \<tturnstile> wp B (\<guillemotleft>Q\<guillemotright> && (wp C S o \<phi>))"
      apply -
      by(rule order_trans, rule wp_combine, blast+)
  }

  also {
    thm drefinesD[OF drBC, where P="wp C S" and Q=S]
    from drBC and sS and bS and hC
    have "\<guillemotleft>Q\<guillemotright> && (wp C S o \<phi>) \<tturnstile> wp D (S o \<phi>)"
      by(blast intro!:drefinesD)
    with hB and hC and hD and sS
    have "wp B (\<guillemotleft>Q\<guillemotright> && (wp C S o \<phi>)) \<tturnstile> wp B (wp D (S o \<phi>))"
      by(blast intro!:mono_transD)
  }

  finally
  show "\<guillemotleft>P\<guillemotright> && (R o \<phi>) \<tturnstile> wp (B;;D) (S o \<phi>)"
    unfolding wp_eval o_def .
qed

lemma dr_Apply:
  fixes \<phi>::"'c \<Rightarrow> 'b"
  assumes commutes: "f o \<phi> = \<phi> o g"
  shows "drefines \<phi> G (Apply f) (Apply g)"
proof(rule drefinesI2, simp_all)
  show "feasible (wp (Apply g))"
    by(simp add:healthy_intros healthy_feasibleD)

  fix P::"'b \<Rightarrow> real" and Q::"'b \<Rightarrow> real" and s::'c

  assume wp: "P \<tturnstile> wp (Apply f) Q"
  hence "P \<tturnstile> (Q o f)" by(simp add:wp_eval)
  hence "P (\<phi> s) \<le> (Q o f) (\<phi> s)" by(blast)
  also have "... = Q ((f o \<phi>) s)" by(simp)
  also with commutes
  have "... = ((Q o \<phi>) o g) s" by(simp)
  also have "... = wp (Apply g) (Q o \<phi>) s"
    by(simp add:wp_eval)
  finally show "P (\<phi> s) \<le> wp (Apply g) (Q o \<phi>) s" .
qed

lemma dr_Skip:
  fixes \<phi>::"'c \<Rightarrow> 'b"
  shows "drefines \<phi> G Skip Skip"
proof(rule drefinesI2)
  show "feasible (wp Skip)" by(simp add:healthy_intros healthy_feasibleD)

  fix P::"'b \<Rightarrow> real" and Q::"'b \<Rightarrow> real" and s::'c
  assume "P \<tturnstile> wp Skip Q"
  hence "(P o \<phi>) s \<le> wp Skip Q (\<phi> s)" by(simp, blast)
  thus "(P o \<phi>) s \<le> wp Skip (Q o \<phi>) s" by(simp add:wp_eval)
qed

section {* Strengthening the Guard *}

lemma dr_weaken_pre:
  fixes A::"'s prog" and B::"'r prog"
  assumes drAB: "drefines \<phi> Q A B"
      and PQ:   "\<And>s. P s \<Longrightarrow> Q s"
  shows "drefines \<phi> P A B"
proof
  fix R::"'s \<Rightarrow> real" and S::"'s \<Rightarrow> real"
  assume sR: "sound R" and bR: "bounded_by 1 R"
     and sS: "sound S" and bS: "bounded_by 1 S"
     and wpA: "R \<tturnstile> wp A S"
  have "\<guillemotleft>P\<guillemotright> && (R o \<phi>) \<tturnstile> \<guillemotleft>Q\<guillemotright> && (R o \<phi>)"
    unfolding exp_conj_def
  proof(rule le_funI, cases)
    fix s assume "P s"
    with PQ show "\<guillemotleft>P\<guillemotright> s .& (R o \<phi>) s \<le> \<guillemotleft>Q\<guillemotright> s .& (R o \<phi>) s"
      by(simp)
  next
    fix s assume "\<not>P s"
    moreover from bR have "R (\<phi> s) \<le> 1" by(blast)
    ultimately
    show "\<guillemotleft>P\<guillemotright> s .& (R o \<phi>) s \<le> \<guillemotleft>Q\<guillemotright> s .& (R o \<phi>) s" by(simp)
  qed
  also from drAB and sR and bR and sS and bS and wpA
  have "... \<tturnstile> wp B (S o \<phi>)" by(rule drefinesD)
  finally show "\<guillemotleft>P\<guillemotright> && (R o \<phi>) \<tturnstile> ..." .
qed

lemma dr_repeat:
  fixes \<phi> :: "'a \<Rightarrow> 'b"
  assumes dr_ab: "drefines \<phi> G a b"
      and Gpr:  "\<guillemotleft>G\<guillemotright> \<tturnstile> wp b \<guillemotleft>G\<guillemotright>"
      and ha:    "healthy (wp a)"
      and hb:    "healthy (wp b)"
      and scb:   "sub_conj (wp b)"
  shows "drefines \<phi> G (repeat n a) (repeat n b)" (is "?X n")
proof(induct n)
  show "?X 0" by(simp add:dr_Skip)

  fix n
  assume IH: "?X n"
  thus "?X (Suc n)"
    apply(simp)
    apply(rule dr_Seq)
    apply(simp_all add:assms healthy_intros)
    apply(rule Gpr)
    done
qed

lemma dr_refl:
  assumes ha: "healthy (wp a)"
      and comm: "\<And>Q. \<lbrakk> sound Q; bounded_by 1 Q \<rbrakk> \<Longrightarrow>
                     wp a Q o \<phi> \<tturnstile> wp a (Q o \<phi>)"
  shows "drefines \<phi> G a a"
proof(rule drefinesI2)
  from ha show "feasible (wp a)" by(blast)

  fix P and Q and s
  assume wp: "P \<tturnstile> wp a Q"
  assume sQ: "sound Q" and bQ: "bounded_by 1 Q"

  have "(P o \<phi>) s = P (\<phi> s)" by(simp)
  also from wp have "... \<le> wp a Q (\<phi> s)" by(blast)
  also {
    from comm and sQ and bQ have "wp a Q o \<phi> \<tturnstile> wp a (Q o \<phi>)" by(blast)
    hence "(wp a Q o \<phi>) s \<le> wp a (Q o \<phi>) s" by(blast)
    hence "wp a Q (\<phi> s) \<le> ..." by(simp)
  }
  finally show  "(P o \<phi>) s \<le> wp a (Q o \<phi>) s" .
qed

end
