(*
 * Copyright (C) 2013 National ICT Australia Ltd ('NICTA')
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Author: David Cock - David.Cock@nicta.com.au *)

header "The Expectation Space of pGCL"

theory Expectations imports Misc CompleteSet begin

text {* Sound expectations are bounded, non-negative functions from
        program state to the the reals. *}

section {* Bounded functions *}

definition bounded_by :: "'b \<Rightarrow> ('a \<Rightarrow> 'b::order) \<Rightarrow> bool"
where     "bounded_by b P \<equiv> \<forall>x. P x \<le> b"

lemma bounded_byI[intro]:
  "\<lbrakk> \<And>x. P x \<le> b \<rbrakk> \<Longrightarrow> bounded_by b P"
  by(simp add:bounded_by_def)

lemma bounded_byI2[intro]:
  "P \<le> (\<lambda>s. b) \<Longrightarrow> bounded_by b P"
  by(blast dest:le_funD)

lemma bounded_byD[dest]:
  "bounded_by b P \<Longrightarrow> P x \<le> b"
  by(simp add:bounded_by_def)

lemma bounded_byD2[dest]:
  "bounded_by b P \<Longrightarrow> P \<le> (\<lambda>s. b)"
  by(blast intro:le_funI)

definition bounded :: "('a \<Rightarrow> 'b::order) \<Rightarrow> bool"
where     "bounded P \<equiv> (\<exists>b. bounded_by b P)"

text {* In an up-complete set, we can extract the bound *}
definition bound_of :: "('a \<Rightarrow> 'b::{up_complete,complete}) \<Rightarrow> 'b"
where     "bound_of P \<equiv> THE b. isLub UNIV (P ` UNIV) b"

lemma bound_of_least[intro]:
  "bounded_by (b::'b::{up_complete,complete}) P \<Longrightarrow> bound_of P \<le> b"
  unfolding bounded_by_def bound_of_def
  apply(rule the1I2)
   apply(blast intro!:up_complete intro:isUbI setleI isLub_isUb)
  apply(blast intro:isLub_le_isUb isUbI setleI)
  done

lemma bounded_by_bound_of[intro!]:
  fixes P::"'a \<Rightarrow> 'b::{complete,up_complete}"
  shows
  "bounded P \<Longrightarrow> bounded_by (bound_of P) P"
  unfolding bounded_by_def bounded_def bound_of_def
  by(rule the1I2,
     (blast dest:isUbD isLub_isUb
      intro:setleI isUbI intro!:up_complete)+)

lemma bound_of_greater[intro]:
  "bounded P \<Longrightarrow> P x \<le> bound_of P"
  by(blast intro:bounded_byD)

lemma bounded_by_mono:
  "\<lbrakk> bounded_by (a::'a::order) P; a \<le> b \<rbrakk> \<Longrightarrow> bounded_by b P"
  unfolding bounded_by_def by(blast intro:order_trans)

lemma bounded_by_imp_bounded[intro]:
  "bounded_by b P \<Longrightarrow> bounded P"
  unfolding bounded_def by(blast)

lemma bounded_by_bound_of_alt:
  "\<lbrakk> bounded P; bound_of P = a \<rbrakk> \<Longrightarrow> bounded_by a P"
  by(blast)

lemma bounded_const[simp]:
  "bounded (\<lambda>x. c)"
  by(blast)

lemma bounded_by_const[intro]:
  "c \<le> b \<Longrightarrow> bounded_by b (\<lambda>x. c)"
  by(blast)

lemma bounded_by_mono_alt[intro]:
  "\<lbrakk> bounded_by b Q; P \<le> Q \<rbrakk> \<Longrightarrow> bounded_by b P"
  by(blast intro:order_trans dest:le_funD)

lemma bound_of_const[simp, intro]:
  "bound_of (\<lambda>x. c) = (c::real)"
  unfolding bound_of_def
proof(rule the1_equality)
  show "isLub UNIV (range (\<lambda>x. c)) c"
    by(blast intro:isLubI2 isUbI setleI setgeI dest:isUbD)
  thus "\<exists>!b. isLub UNIV (range (\<lambda>x. c)) b"
    by(blast intro:real_isLub_unique)
qed

lemma bound_of_leI:
  "\<lbrakk> \<And>x. P x \<le> (c::real) \<rbrakk> \<Longrightarrow> bound_of P \<le> c"
  unfolding bound_of_def
  apply(rule the1I2)
   apply(blast intro:isUbI setleI reals_complete real_isLub_unique)
  apply(blast intro:isLub_le_isUb isUbI setleI)
done

lemma bound_of_mono[intro]:
  "\<lbrakk> P \<le> Q; bounded P; bounded Q \<rbrakk> \<Longrightarrow> bound_of P \<le> bound_of Q"
  by(blast intro:order_trans dest:le_funD)

lemma bounded_by_o[intro,simp]:
  "\<And>b. bounded_by b P \<Longrightarrow> bounded_by b (P o f)"
  unfolding o_def by(blast)

lemma le_bound_of[intro]:
  "\<And>x. bounded f \<Longrightarrow> f x \<le> bound_of f"
  by(blast)

subsection {* Positive semi-definite (non-negative) functions. *}

definition
  nneg :: "('a \<Rightarrow> 'b::{zero,order}) \<Rightarrow> bool"
where
  "nneg P \<longleftrightarrow> (\<forall>x. 0 \<le> P x)"

lemma nnegI[intro]:
  "\<lbrakk> \<And>x. 0 \<le> P x \<rbrakk> \<Longrightarrow> nneg P"
  by(simp add:nneg_def)

lemma nnegI2[intro]:
  "(\<lambda>s. 0) \<le> P \<Longrightarrow> nneg P"
  by(blast dest:le_funD)

lemma nnegD[dest]:
  "nneg P \<Longrightarrow> 0 \<le> P x"
  by(simp add:nneg_def)

lemma nnegD2[dest]:
  "nneg P \<Longrightarrow> (\<lambda>s. 0) \<le> P"
  by(blast intro:le_funI)

lemma nneg_const[iff]:
  "nneg (\<lambda>x. c) \<longleftrightarrow> 0 \<le> c"
  by(simp add:nneg_def)

lemma nneg_o[intro,simp]:
  "nneg P \<Longrightarrow> nneg (P o f)"
  by(force)

lemma nneg_bound_nneg[intro]:
  "\<lbrakk> bounded P; nneg P \<rbrakk> \<Longrightarrow> 0 \<le> bound_of P"
  by(blast intro:order_trans)

lemma nneg_bounded_by_nneg[dest]:
  "\<lbrakk> bounded_by b P; nneg P \<rbrakk> \<Longrightarrow> 0 \<le> (b::'a::{zero,order,complete})"
  by(blast intro:order_trans)

lemma bounded_by_nneg[dest]:
  fixes P::"'s \<Rightarrow> real"
  shows "\<lbrakk> bounded_by b P; nneg P \<rbrakk> \<Longrightarrow> 0 \<le> b"
  by(blast intro:order_trans)

section {* Sound expectations (nneg and bounded) *}

definition sound :: "('s \<Rightarrow> real) \<Rightarrow> bool"
where "sound P \<equiv> bounded P \<and> nneg P"

lemma soundI[intro]:
  "\<lbrakk> bounded P; nneg P \<rbrakk> \<Longrightarrow> sound P"
  by(simp add:sound_def)

lemma sound_bounded[dest]:
  "sound P \<Longrightarrow> bounded P"
  by(simp add:sound_def)

lemma sound_nneg[dest]:
  "sound P \<Longrightarrow> nneg P"
  by(simp add:sound_def)

lemma sound_sum[simp,intro]:
  assumes sP: "sound P" and sQ: "sound Q"
  shows "sound (\<lambda>s. P s + Q s)"
proof
  from sP have "\<And>s. P s \<le> bound_of P" by(blast)
  moreover from sQ have "\<And>s. Q s \<le> bound_of Q" by(blast)
  ultimately have "\<And>s. P s + Q s \<le> bound_of P + bound_of Q"
    by(rule add_mono)
  hence "bounded_by (bound_of P + bound_of Q) (\<lambda>s. P s + Q s)"
    by(blast)
  thus "bounded (\<lambda>s. P s + Q s)" by(blast)

  from sP have "\<And>s. 0 \<le> P s" by(blast)
  moreover from sQ have "\<And>s. 0 \<le> Q s" by(blast)
  ultimately have "\<And>s. 0 \<le> P s + Q s" by(simp add:add_mono)
  thus "nneg (\<lambda>s. P s + Q s)" by(blast)
qed

lemma mult_sound:
  assumes sP: "sound P" and sQ: "sound Q"
  shows "sound (\<lambda>s. P s * Q s)"
proof
  from sP have "\<And>s. P s \<le> bound_of P" by(blast)
  moreover from sQ have "\<And>s. Q s \<le> bound_of Q" by(blast)
  ultimately have "\<And>s. P s * Q s \<le> bound_of P * bound_of Q"
    using sP and sQ by(blast intro:mult_mono)
  thus "bounded (\<lambda>s. P s * Q s)" by(blast)

  from sP and sQ show "nneg (\<lambda>s. P s * Q s)"
    by(blast intro:mult_nonneg_nonneg)
qed

lemma div_sound:
  assumes sP: "sound P" and cpos: "0 < c"
  shows "sound (\<lambda>s. P s / c)"
proof
  from sP and cpos have "\<And>s. P s / c \<le> bound_of P / c"
    by(blast intro:divide_right_mono less_imp_le)
  thus "bounded (\<lambda>s. P s / c)" by(blast)
  from assms show "nneg (\<lambda>s. P s / c)"
    by(blast intro:divide_nonneg_pos)
qed

lemma const_sound:
  "0 \<le> c \<Longrightarrow> sound (\<lambda>s. c)"
  by(blast)

lemma sound_o[intro,simp]:
  "sound P \<Longrightarrow> sound (P o f)"
  unfolding o_def by(blast)

lemma sc_bounded_by[intro,simp]:
  "\<lbrakk> sound P; 0 < c \<rbrakk> \<Longrightarrow> bounded_by (c * bound_of P) (\<lambda>x. c * P x)"
  by(blast intro!:mult_left_mono dest!:less_imp_le)

lemma sc_bounded[intro,simp]:
  "\<lbrakk> sound P; 0 \<le> c \<rbrakk> \<Longrightarrow> bounded (\<lambda>x. c * P x)"
  apply(case_tac "c=0", simp_all)
   apply(subgoal_tac "0 < c") apply(blast)
  apply(simp)
  done

lemma sc_bound[simp]:
  assumes sP: "sound P"
      and cpos: "0 < c"
  shows "c * bound_of P = bound_of (\<lambda>x. c * P x)"
proof(rule antisym)
  from sP and cpos have "bounded (\<lambda>x. c * P x)" by(blast)
  hence "\<And>x. c * P x \<le> bound_of (\<lambda>x. c * P x)"
    by(rule le_bound_of)
  with cpos have "\<And>x. P x \<le> inverse c * bound_of (\<lambda>x. c * P x)"
    by(force intro:mult_div_mono_right)
  hence "bound_of P \<le> inverse c * bound_of (\<lambda>x. c * P x)"
    by(blast)
  with cpos show "c * bound_of P \<le> bound_of (\<lambda>x. c * P x)"
    by(force intro:mult_div_mono_left)
next
  from sP and cpos have "\<And>x. c * P x \<le> c * bound_of P"
    by(blast intro:mult_left_mono less_imp_le)
  thus "bound_of (\<lambda>x. c * P x) \<le> c * bound_of P"
    by(blast)
qed

lemma sc_sound:
  "\<lbrakk> sound P; 0 \<le> c \<rbrakk> \<Longrightarrow> sound (\<lambda>s. c * P s)"
  by(blast intro:sc_bounded mult_nonneg_nonneg)

section {* Fixed points on expectations *}

subsection {* (Internally) Bounded sets *}

definition
  bounded_below :: "'a::order set \<Rightarrow> bool"
where
  "bounded_below S \<equiv> \<exists>!b\<in>S. \<forall>s\<in>S. b \<le> s"

lemma bounded_belowI[intro]:
  "\<lbrakk> b \<in> S; \<And>s. s \<in> S \<Longrightarrow> b \<le> s;
     \<And>z. \<lbrakk> \<And>s. s \<in> S \<Longrightarrow> z \<le> s \<rbrakk> \<Longrightarrow> z \<le> b \<rbrakk> \<Longrightarrow> bounded_below S"
  unfolding bounded_below_def by(force)

lemma bounded_belowD[dest]:
  "bounded_below S \<Longrightarrow> \<exists>!b\<in>S. \<forall>s\<in>S. b \<le> s"
  by(simp add:bounded_below_def)

definition
  bounded_above :: "'a::order set \<Rightarrow> bool"
where
  "bounded_above S \<equiv> \<exists>!t\<in>S. \<forall>s\<in>S. s \<le> t"

lemma bounded_aboveI[intro]:
  "\<lbrakk> t \<in> S; \<And>s. s \<in> S \<Longrightarrow> s \<le> t;
     \<And>z. \<lbrakk> \<And>s. s \<in> S \<Longrightarrow> s \<le> z \<rbrakk> \<Longrightarrow> t \<le> z \<rbrakk> \<Longrightarrow> bounded_above S"
  unfolding bounded_above_def by(force)

lemma bounded_aboveD[dest]:
  "bounded_above S \<Longrightarrow> \<exists>!t\<in>S. \<forall>s\<in>S. s \<le> t"
  by(simp add:bounded_above_def)

text {* Find the bound in a complete set. *}

definition
  lowerbound :: "'a::order set \<Rightarrow> 'a"
where
  "lowerbound S = (THE bot. bot \<in> S \<and> (\<forall>s\<in>S. bot \<le> s))"

lemma lowerbound_lower[intro,simp]:
  "\<lbrakk> bounded_below S; x \<in> S \<rbrakk> \<Longrightarrow> lowerbound S \<le> x"
  unfolding bounded_below_def lowerbound_def
  by(rule the1I2, assumption, blast)

lemma lowerbound_in[intro,simp]:
  "bounded_below S \<Longrightarrow> lowerbound S \<in> S"
  unfolding lowerbound_def
  by(rule the1I2, (blast dest!:bounded_belowD)+)

definition
  upperbound :: "'a::order set \<Rightarrow> 'a"
where
  "upperbound S \<equiv> THE top. top \<in> S \<and> (\<forall>s\<in>S. s \<le> top)"

lemma upperbound_upper[intro,simp]:
  "\<lbrakk> bounded_above S; x \<in> S \<rbrakk> \<Longrightarrow> x \<le> upperbound S"
  unfolding bounded_above_def upperbound_def
  by(rule the1I2, assumption, blast)

lemma upperbound_in[intro,simp]:
  "bounded_above S \<Longrightarrow> upperbound S \<in> S"
  unfolding upperbound_def
  by(rule the1I2, blast+)

definition
  bounded_iterates :: "('a::order \<Rightarrow> 'a) \<Rightarrow> 'a set \<Rightarrow> bool"
where
  "bounded_iterates f S \<longleftrightarrow> (\<exists>b\<in>S. \<forall>s\<in>S. s \<le> f s \<longrightarrow> s \<le> b)"

lemma bounded_iteratesI[intro]:
  "\<lbrakk> b \<in> S; \<And>s. \<lbrakk> s \<in> S;  s \<le> f s \<rbrakk> \<Longrightarrow> s \<le> b \<rbrakk> \<Longrightarrow> bounded_iterates f S"
  unfolding bounded_iterates_def by(blast)

lemma bounded_iteratesD[dest]:
  "bounded_iterates f S \<Longrightarrow> \<exists>b\<in>S. \<forall>s\<in>S. s \<le> f s \<longrightarrow> s \<le> b"
  by(simp add:bounded_iterates_def)

subsection {* Up-complete and down-complete sets *}

definition
  up_complete :: "'a::complete set \<Rightarrow> bool"
where
  "up_complete S \<equiv>
     \<forall>T. (T \<noteq> {} \<and> T \<subseteq> S \<and> (\<exists>t. t \<in> S \<and> (\<forall>x\<in>T. x \<le> t))) \<longrightarrow> lub T \<in> S"

lemma up_completeI:
  assumes X: "\<And>T t. \<lbrakk> T \<noteq> {}; T \<subseteq> S; t \<in> S; \<And>x. x \<in> T \<Longrightarrow> x \<le> t \<rbrakk> \<Longrightarrow> lub T \<in> S"
  shows      "up_complete S"
  unfolding up_complete_def by(blast intro:X)

lemma up_completeD:
  "\<lbrakk> up_complete S; T \<noteq> {}; T \<subseteq> S; t \<in> S; \<And>x. x \<in> T \<Longrightarrow> x \<le> t \<rbrakk> \<Longrightarrow> lub T \<in> S"
  unfolding up_complete_def by(erule_tac x=T in allE, blast)

definition
  down_complete :: "'a::complete set \<Rightarrow> bool"
where
  "down_complete S \<equiv>
     \<forall>T. (T \<noteq> {} \<and> T \<subseteq> S \<and> (\<exists>b. b \<in> S \<and> (\<forall>x\<in>T. b \<le> x))) \<longrightarrow> glb T \<in> S"

lemma down_completeI:
  assumes X: "\<And>T b. \<lbrakk> T \<noteq> {}; T \<subseteq> S; b \<in> S; \<And>x. x \<in> T \<Longrightarrow> b \<le> x \<rbrakk> \<Longrightarrow> glb T \<in> S"
  shows      "down_complete S"
  unfolding down_complete_def by(blast intro:X)

lemma down_completeD:
  "\<lbrakk> down_complete S; T \<noteq> {}; T \<subseteq> S; b \<in> S; \<And>x. x \<in> T \<Longrightarrow> b \<le> x \<rbrakk> \<Longrightarrow> glb T \<in> S"
  unfolding down_complete_def
  by(erule_tac x=T in allE, blast)

subsection {* The set of sound expectations *}

definition
  sound_expectations :: "('s \<Rightarrow> real) set"
where
  "sound_expectations = {P. sound P}"

lemma sound_expectationsI[intro]:
  "sound P \<Longrightarrow> P \<in> sound_expectations"
  by(simp add:sound_expectations_def)

lemma sound_expectations_bounded[dest]:
  "P \<in> sound_expectations \<Longrightarrow> bounded P"
  unfolding sound_expectations_def by(blast)

lemma sound_expectations_nneg[dest]:
  "P \<in> sound_expectations \<Longrightarrow> nneg P"
  unfolding sound_expectations_def by(blast)

lemma sound_bounded_below[simp,intro!]:
  "bounded_below sound_expectations"
  unfolding sound_expectations_def
  by(rule_tac b="\<lambda>s. 0" in bounded_belowI,
     (force intro:le_funI simp:soundI)+)

(* XXX - tidy up these two *)

lemma sound_up_complete[simp,intro!]:
  "up_complete sound_expectations"
  unfolding sound_expectations_def
  apply(rule up_completeI, simp, rule soundI)
   apply(rule_tac b="bound_of t" in bounded_by_imp_bounded)
   apply(rule bounded_byI)
   apply(rule le_funD[OF lub_least], blast)
   apply(clarify, rule le_funI)
   apply(blast intro:order_trans le_funI dest:le_funD)
  apply(rule nnegI)
  apply(drule nonempty_witness, erule exE)
  apply(rule_tac y="xa x" in order_trans)
   apply(blast elim:nnegD)
  apply(blast dest:le_funD dest:lub_greater)
  done

lemma sound_down_complete[simp,intro!]:
  "down_complete sound_expectations"
  unfolding sound_expectations_def
  apply(rule down_completeI, simp, rule soundI)
   apply(erule exE[OF nonempty_witness])
   apply(rule_tac b="bound_of x" in bounded_by_imp_bounded)
   apply(rule bounded_byI)
   apply(rule_tac y="x xa" in order_trans)
    apply(blast intro:le_funD[OF glb_lower])
   apply(blast intro:bounded_byD bounded_by_bound_of)
  apply(rule nnegI)
  apply(erule exE[OF nonempty_witness])
  apply(rule le_funD[OF glb_greatest], blast)
  apply(blast elim:nnegD intro:le_funI)
  done

lemma sound_expectations_sound[dest]:
  "P \<in> sound_expectations \<Longrightarrow> sound P"
  unfolding sound_expectations_def by(simp)

section {* Standard Expectations *}

text {* Standard expectations are the embedding of a boolean predicate, mapping
        False to 0 and True to 1. *}

text {* We write \<guillemotleft>P\<guillemotright> rather than [P] for boolean embedding to avoid
        clashing with list syntax. *}
definition
  embed_bool :: "('s \<Rightarrow> bool) \<Rightarrow> 's \<Rightarrow> real" ("\<guillemotleft> _ \<guillemotright>" 1000)
where
  "\<guillemotleft>P\<guillemotright> \<equiv> (\<lambda>s. if P s then 1 else 0)"

lemma embed_bool_nneg[simp,intro]:
  "nneg \<guillemotleft>P\<guillemotright>"
  unfolding embed_bool_def by(force)

lemma embed_bool_bounded_by_1[simp,intro]:
  "bounded_by 1 \<guillemotleft>P\<guillemotright>"
  unfolding embed_bool_def by(force)

lemma embed_bool_bounded[simp,intro]:
  "bounded \<guillemotleft>P\<guillemotright>"
  by(blast)

lemma embed_bool_idem:
  "\<guillemotleft>P\<guillemotright> s * \<guillemotleft>P\<guillemotright> s = \<guillemotleft>P\<guillemotright> s"
  by(simp add:embed_bool_def)

lemma eval_embed_true[simp]:
  "P s \<Longrightarrow> \<guillemotleft>P\<guillemotright> s = 1"
  by(simp add:embed_bool_def)

lemma eval_embed_false[simp]:
  "\<not>P s \<Longrightarrow> \<guillemotleft>P\<guillemotright> s = 0"
  by(simp add:embed_bool_def)

definition negate :: "('s \<Rightarrow> bool) \<Rightarrow> 's \<Rightarrow> bool" ("\<^boverline>_\<^eoverline>")
where     "negate P = (\<lambda>s. \<not> P s)"

notation (latex output)
  embed_bool ("\\overline{ _ }")

lemma negateI:
  "\<not> P s \<Longrightarrow> \<^boverline>P\<^eoverline> s"
  by(simp add:negate_def)

lemma embed_split:
  "f s = \<guillemotleft>P\<guillemotright> s * f s + \<guillemotleft>\<^boverline>P\<^eoverline>\<guillemotright> s * f s"
  by(simp add:negate_def embed_bool_def)

lemma negate_embed:
  "\<guillemotleft>\<^boverline>G\<^eoverline>\<guillemotright> s = 1 - \<guillemotleft>G\<guillemotright> s"
  by(simp add:embed_bool_def negate_def)

lemma eval_nembed_true[simp]:
  "P s \<Longrightarrow> \<guillemotleft>\<^boverline>P\<^eoverline>\<guillemotright> s = 0"
  by(simp add:embed_bool_def negate_def)

lemma eval_nembed_false[simp]:
  "\<not>P s \<Longrightarrow> \<guillemotleft>\<^boverline>P\<^eoverline>\<guillemotright> s = 1"
  by(simp add:embed_bool_def negate_def)

lemma embed_ge_0[simp,intro]:
  "0 \<le> \<guillemotleft>G\<guillemotright> s"
  by(simp add:embed_bool_def)

lemma embed_le_1[simp,intro]:
  "\<guillemotleft>G\<guillemotright> s \<le> 1"
  by(simp add:embed_bool_def)

lemma embed_le_1_alt[simp,intro]:
  "0 \<le> 1 - \<guillemotleft>G\<guillemotright> s"
  by(subst add_le_cancel_right[where c="\<guillemotleft>G\<guillemotright> s", symmetric], simp)

lemma expect_1_I:
  "P x \<Longrightarrow> 1 \<le> \<guillemotleft>P\<guillemotright> x"
  by(simp)

lemma standard_sound[intro,simp]:
  "sound \<guillemotleft>P\<guillemotright>"
  by(blast)

lemma embed_o[simp]:
  "\<guillemotleft>P\<guillemotright> o f = \<guillemotleft>P o f\<guillemotright>"
  unfolding embed_bool_def o_def by(simp)

section {* Entailment *}

text {* Entailment on expectations is a generalization of that on predicates. *}

abbreviation entails :: "('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real) \<Rightarrow> bool" ("_ \<tturnstile> _" 70)
where "P \<tturnstile> Q \<equiv> P \<le> Q"

lemma entailsI[intro]:
  "\<lbrakk>\<And>s. P s \<le> Q s\<rbrakk> \<Longrightarrow> P \<tturnstile> Q"
  by(simp add:le_funI)

lemma entailsD[dest]:
  "P \<tturnstile> Q \<Longrightarrow> P s \<le> Q s"
  by(simp add:le_funD)

lemma eq_entails[intro]:
  "P = Q \<Longrightarrow> P \<tturnstile> Q"
  by(blast)

lemma entails_trans[trans]:
  "\<lbrakk> P \<tturnstile> Q; Q \<tturnstile> R \<rbrakk> \<Longrightarrow> P \<tturnstile> R"
  by(blast intro:order_trans)

text {* For standard expectations, both notions of entailment coincide. *}

lemma implies_entails:
  "\<lbrakk> \<And>s. P s \<Longrightarrow> Q s \<rbrakk> \<Longrightarrow> \<guillemotleft>P\<guillemotright> \<tturnstile> \<guillemotleft>Q\<guillemotright>"
  by(rule entailsI, case_tac "P s", simp_all)

lemma entails_implies:
  "\<And>s. \<lbrakk> \<guillemotleft>P\<guillemotright> \<tturnstile> \<guillemotleft>Q\<guillemotright>; P s \<rbrakk> \<Longrightarrow> Q s"
  by(rule ccontr, drule_tac s=s in entailsD, simp)

section {* Expectation Conjunction *}

definition
  pconj :: "real \<Rightarrow> real \<Rightarrow>  real" (infixl ".&" 71)
where
  "p .& q \<equiv> p + q \<ominus> 1"

lemma pconj_lzero[intro,simp]:
  "b \<le> 1 \<Longrightarrow> 0 .& b = 0"
  by(simp add:pconj_def tminus_def)

lemma pconj_rzero[intro,simp]:
  "b \<le> 1 \<Longrightarrow> b .& 0 = 0"
  by(simp add:pconj_def tminus_def)

lemma pconj_lone[intro,simp]:
  "0 \<le> b \<Longrightarrow> 1 .& b = b"
  by(simp add:pconj_def tminus_def)

lemma pconj_rone[intro,simp]:
  "0 \<le> b \<Longrightarrow> b .& 1 = b"
  by(simp add:pconj_def tminus_def)

lemma pconj_bconj:
  "\<guillemotleft>a\<guillemotright> s .& \<guillemotleft>b\<guillemotright> s = \<guillemotleft>\<lambda>s. a s \<and> b s\<guillemotright> s"
  unfolding embed_bool_def pconj_def tminus_def by(force)

lemma pconj_comm[ac_simps]:
  "a .& b = b .& a"
  by(simp add:pconj_def ac_simps)

lemma pconj_assoc:
  "\<lbrakk> 0 \<le> a; a \<le> 1; 0 \<le> b; b \<le> 1; 0 \<le> c; c \<le> 1 \<rbrakk> \<Longrightarrow>
   a .& (b .& c) = (a .& b) .& c"
  unfolding pconj_def tminus_def by(simp)

lemma pconj_mono:
  "\<lbrakk> a \<le> b; c \<le> d \<rbrakk> \<Longrightarrow> a .& c \<le> b .& d"
  unfolding pconj_def tminus_def by(simp)

lemma pconj_nneg[intro,simp]:
  "0 \<le> a .& b"
  unfolding pconj_def tminus_def
  by(blast intro:min_max.sup_ge2)

definition exp_conj :: "('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)"
           (infixl "&&" 71)
where "a && b \<equiv> \<lambda>s. (a s .& b s)"

lemma exp_conj_mono_left:
  "P \<tturnstile> Q \<Longrightarrow> P && R \<tturnstile> Q && R"
  unfolding exp_conj_def pconj_def
  apply(rule le_funI)
  apply(rule tminus_left_mono)
  apply(rule add_right_mono)
  apply(blast)
  done

lemma exp_conj_mono_right:
  "Q \<tturnstile> R \<Longrightarrow> P && Q \<tturnstile> P && R"
  apply(simp add:exp_conj_def pconj_def)
  apply(rule le_funI)
  apply(rule tminus_left_mono)
  apply(rule add_left_mono)
  apply(blast)
  done

lemma exp_conj_comm[ac_simps]:
  "a && b = b && a"
  by(simp add:exp_conj_def ac_simps)

lemma exp_conj_bounded_by[intro,simp]:
  assumes bP: "bounded_by 1 P"
      and bQ: "bounded_by 1 Q"
  shows "bounded_by 1 (P && Q)"
proof(rule bounded_byI, unfold exp_conj_def pconj_def)
  fix x
  from bP have "P x \<le> 1" by(blast)
  moreover from bQ have "Q x \<le> 1" by(blast)
  ultimately have "P x + Q x \<le> 2" by(auto)
  thus "P x + Q x \<ominus> 1 \<le> 1"
    unfolding tminus_def by(simp)
qed

lemma exp_conj_o_distrib[simp]:
  "(P && Q) o f = (P o f) && (Q o f)"
  unfolding exp_conj_def o_def by(simp)

lemma exp_conj_assoc:
  assumes "sound P" and "bounded_by 1 P"
      and "sound Q" and "bounded_by 1 Q"
      and "sound R" and "bounded_by 1 R"
  shows "P && (Q && R) = (P && Q) && R"
  unfolding exp_conj_def
proof(rule ext)
  fix s
  from assms have "0 \<le> P s" by(blast)
  moreover from assms have "0 \<le> Q s" by(blast)
  moreover from assms have "0 \<le> R s" by(blast)
  moreover from assms have "P s \<le> 1" by(blast)
  moreover from assms have "Q s \<le> 1" by(blast)
  moreover from assms have "R s \<le> 1" by(blast)
  ultimately
  show "P s .& (Q s .& R s) = (P s .& Q s) .& R s"
    by(simp add:pconj_assoc)
qed

lemma exp_conj_top_left[simp]:
  "sound P \<Longrightarrow> \<guillemotleft>\<lambda>_. True\<guillemotright> && P = P"
  unfolding exp_conj_def by(force)

lemma exp_conj_top_right[simp]:
  "sound P \<Longrightarrow> P && \<guillemotleft>\<lambda>_. True\<guillemotright> = P"
  unfolding exp_conj_def by(force)

lemma exp_conj_idem[simp]:
  "\<guillemotleft>P\<guillemotright> && \<guillemotleft>P\<guillemotright> = \<guillemotleft>P\<guillemotright>"
  unfolding exp_conj_def
  by(rule ext, case_tac "P s", simp_all)

lemma exp_conj_nneg[intro,simp]:
  "(\<lambda>s. 0) \<le> P && Q"
  unfolding exp_conj_def
  by(blast intro:le_funI)

lemma exp_conj_sound[intro,simp]:
  assumes s_P: "sound P"
      and s_Q: "sound Q"
  shows "sound (P && Q)"
  unfolding exp_conj_def
proof(rule soundI)
  from s_P and s_Q have "\<And>s. 0 \<le> P s + Q s" by(blast intro:add_nonneg_nonneg)
  hence "\<And>s. P s .& Q s \<le> P s + Q s"
    unfolding pconj_def by(force intro:tminus_less)
  also from assms have "\<And>s. ... s \<le> bound_of P + bound_of Q"
    by(blast intro:add_mono)
  finally have "bounded_by (bound_of P + bound_of Q) (\<lambda>s. P s .& Q s)"
    by(blast)
  thus "bounded (\<lambda>s. P s .& Q s)" by(blast)

  show "nneg (\<lambda>s. P s .& Q s)"
    unfolding pconj_def tminus_def by(force)
qed

lemma exp_conj_rzero[simp]:
  "bounded_by 1 P \<Longrightarrow> P && (\<lambda>s. 0) = (\<lambda>s. 0)"
  unfolding exp_conj_def by(force)

(* XXX - tidy this *)

lemma exp_conj_1_right[simp]:
  "nneg A \<Longrightarrow> A && (\<lambda>_. 1) = A"
  unfolding exp_conj_def pconj_def tminus_def
  apply(rule ext)
  apply(simp)
  apply(subgoal_tac "0 \<le> A s", simp)
  apply(force)
  done

subsection {* Rules involving entailment and conjunction together *}

text {* Combines entailments using conjunction. *}
lemma entails_frame:
  assumes ePR: "P \<tturnstile> R"
      and eQS: "Q \<tturnstile> S"
  shows "P && Q \<tturnstile> R && S"
proof(rule le_funI)
  fix s
  from ePR have "P s \<le> R s" by(blast)
  moreover from eQS have "Q s \<le> S s" by(blast)
  ultimately have "P s + Q s \<le> R s + S s" by(rule add_mono)
  hence "P s + Q s \<ominus> 1 \<le> R s + S s \<ominus> 1" by(rule tminus_left_mono)
  thus "(P && Q) s \<le> (R && S) s"
    unfolding exp_conj_def pconj_def .
qed

text {* Enables something like a case distinction on the pre-expectation. *}
lemma pentails_cases:
  assumes PQe: "\<And>x. P x \<tturnstile> Q x"
      and exhaust: "\<And>s. \<exists>x. P (x s) s = 1"
      and framed: "\<And>x. P x && R \<tturnstile> Q x && S"
      and sR: "sound R" and sS: "sound S"
      and bQ: "\<And>x. bounded_by 1 (Q x)"
  shows "R \<tturnstile> S"
proof(rule le_funI)
  fix s
  from exhaust obtain x where P_xs: "P x s = 1" by(blast)
  moreover {
    hence "1 = P x s" by(simp)
    also from PQe have "P x s \<le> Q x s" by(blast dest:le_funD)
    finally have "Q x s = 1"
      using bQ by(blast intro:antisym)
  }
  moreover note le_funD[OF framed[where x=x], where x=s]
  moreover from sR have "0 \<le> R s" by(blast)
  moreover from sS have "0 \<le> S s" by(blast)
  ultimately show "R s \<le> S s" by(simp add:exp_conj_def)
qed

end
