(*
 * Copyright (C) 2013 National ICT Australia Ltd ('NICTA')
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Author: David Cock - David.Cock@nicta.com.au *)

header "Induction principles for pGCL"

theory Induction
  imports Expectation_Induction Transformer_Induction
begin

definition
  weakly_bounded_by :: "('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real) set"
where
  "weakly_bounded_by P \<equiv> {Q. sound Q \<and> bound_of Q \<le> bound_of P}"

lemma weakly_bounded_sound:
  "Q \<in> weakly_bounded_by P \<Longrightarrow> sound Q"
  unfolding weakly_bounded_by_def by(simp)

lemma weakly_bounded_bound:
  "Q \<in> weakly_bounded_by P \<Longrightarrow> bound_of Q \<le> bound_of P"
  unfolding weakly_bounded_by_def by(simp)

lemma weakly_bounded_bby:
  "\<lbrakk> Q \<in> weakly_bounded_by P; sound P \<rbrakk> \<Longrightarrow> bounded_by (bound_of P) Q"
  unfolding weakly_bounded_by_def by(simp, blast intro:bounded_by_mono)

lemma weakly_bounded_bot:
  "sound P \<Longrightarrow> (\<lambda>s. 0) \<in> weakly_bounded_by P"
  unfolding weakly_bounded_by_def by(simp, blast)

lemma weakly_bounded_bounded_below:
  "sound P \<Longrightarrow> bounded_below (weakly_bounded_by P)"
  using weakly_bounded_sound by(blast intro:weakly_bounded_bot)

lemma weakly_bounded_top:
  "sound P \<Longrightarrow> (\<lambda>s. bound_of P) \<in> weakly_bounded_by P"
  unfolding weakly_bounded_by_def by(simp, blast)

lemma weakly_bounded_bounded_above:
  "sound P \<Longrightarrow> bounded_above (weakly_bounded_by P)"
  apply(rule bounded_aboveI)
    apply(rule weakly_bounded_top, assumption)
   apply(rule_tac y="\<lambda>x. bound_of s" in order_trans)
    using weakly_bounded_sound apply(blast intro:le_funI)
   using weakly_bounded_bound apply(blast intro:le_funI)
  using weakly_bounded_top apply(blast)
  done

lemma lub_sound:
  assumes ne:    "S \<noteq> {}"
      and Ssnd:  "\<forall>P\<in>S. sound P"
      and bound: "\<forall>P\<in>S. bounded_by b P"
  shows "sound (lub S)"
proof
  show "bounded_by (lub (bound_of ` S)) (lub S)"
    apply(rule bounded_byI)
    apply(rule_tac x=x in le_funD)
    apply(rule lub_least)
     using ne apply(blast)
    apply(safe)
    apply(rule le_funI)
    apply(rule_tac y="bound_of a" in order_trans)
     using Ssnd apply(blast)
    apply(rule lub_greater)
     apply(blast)
    using bound apply(blast)
    done

  from ne obtain P where "P \<in> S" by(blast)
  thus "nneg (lub S)"
    apply -
    apply(rule nnegI)
    apply(rule_tac x=x in le_funD)
    apply(rule_tac y=P in order_trans)
     using Ssnd apply(blast)
    apply(rule lub_greater)
     apply(assumption)
    using bound apply(blast)
    done
qed

lemma glb_sound:
  assumes ne:    "S \<noteq> {}"
      and Ssnd:  "\<forall>P\<in>S. sound P"
      and bound: "\<forall>P\<in>S. bounded_by b P"
  shows "sound (glb S)"
proof
  from ne and Ssnd obtain P where "P \<in> S" and "sound P" by(blast)
  thus "bounded_by b (glb S)"
    apply -
    apply(rule bounded_byI)
    apply(rule_tac y="P x" in order_trans)
     apply(rule le_funD[OF glb_lower])
      apply(assumption)
     using Ssnd apply(blast)
    using bound apply(blast)
    done

  from ne obtain P where "P \<in> S" by(blast)
  thus "nneg (glb S)"
    apply -
    apply(rule nnegI)
    apply(rule le_funD[OF glb_greatest])
     using ne apply(blast)
    using Ssnd apply(blast)
    done
qed

lemma weakly_bounded_up_complete:
  "sound P \<Longrightarrow> up_complete (weakly_bounded_by P)"
  unfolding weakly_bounded_by_def
  apply(rule up_completeI)
  apply(safe)
   apply(rule_tac b="bound_of P" in lub_sound)
     apply(blast)
    apply(blast)
   apply(safe)
   apply(blast intro:bounded_by_mono)
   apply(rule bound_of_leI)
   apply(rule_tac x=xa in le_funD)
   apply(rule lub_least)
    apply(blast)
   apply(safe)
   apply(rule le_funI)
   apply(rule_tac y="bound_of a" in order_trans)
    apply(blast)
   apply(blast)
   done

lemma weakly_bounded_down_complete:
  "down_complete (weakly_bounded_by P)"
  unfolding weakly_bounded_by_def
  apply(rule down_completeI)
  apply(clarsimp, safe)
   apply(rule_tac b="bound_of P" in glb_sound)
     apply(blast)
    apply(blast)
   apply(safe)
   apply(blast)
  apply(rule bound_of_least)
  apply(rule bounded_byI)
  apply(rule_tac y="x xa" in order_trans)
   apply(rule le_funD[OF glb_lower])
    apply(blast)
   apply(blast)
  apply(blast intro:order_trans)
  done

lemma weakly_bounded_mono:
  "\<lbrakk> P \<le> Q; sound P; sound Q \<rbrakk> \<Longrightarrow> weakly_bounded_by P \<subseteq> weakly_bounded_by Q"
  unfolding weakly_bounded_by_def by(blast intro:order_trans)

lemma weakly_bounded_healthy_closed:
  "healthy t \<Longrightarrow> closed_under t (weakly_bounded_by P)"
  unfolding weakly_bounded_by_def
  apply(rule closed_underI)
  apply(simp, safe)
   apply(blast)
  apply(rule bound_of_leI)
  apply(rule_tac y="bound_of x" in order_trans)
   apply(rule_tac x=xa in bounded_byD)
   apply(blast)
  apply(assumption)
  done

lemma weakly_bounded_self:
  "sound P \<Longrightarrow> P \<in> weakly_bounded_by P"
  unfolding weakly_bounded_by_def by(simp)

lemmas weakly_bounded_intros =
  weakly_bounded_bot weakly_bounded_top
  weakly_bounded_bounded_below weakly_bounded_bounded_above
  weakly_bounded_up_complete weakly_bounded_down_complete
  weakly_bounded_mono weakly_bounded_healthy_closed
  weakly_bounded_self

lemma weakly_bounded_byI:
  "\<lbrakk> sound Q; bound_of Q \<le> bound_of P \<rbrakk> \<Longrightarrow> Q \<in> weakly_bounded_by P"
  unfolding weakly_bounded_by_def by(simp)

lemma gfp_pulldown:
  fixes P
  assumes tailcall:  "\<And>u P. T u P = t P (u P)"
      and sound:     "sound P"
      and mono_T:    "mono_ttrans T"
      and weakly_bounded_closed:
                     "\<And>P. sound P \<Longrightarrow> closed_under (t P) (weakly_bounded_by P)"
      and weakly_bounded_mono_on:
                     "\<And>P. sound P \<Longrightarrow> mono_on (t P) (weakly_bounded_by P)"
      and weakly_mono:
                     "\<And>P Q R. \<lbrakk> P \<le> Q; sound P; sound Q;
                                R \<in> weakly_bounded_by Q \<rbrakk> \<Longrightarrow> t P R \<le> t Q R"
      and health_pr: "\<And>t. healthy t \<Longrightarrow> healthy (T t)"
      and semi_sc:    "\<And>P Q s c. \<lbrakk> sound P; sound Q; 0 < c \<rbrakk> \<Longrightarrow>
                                 c * t P Q s = t (\<lambda>s. c * P s) (\<lambda>s. c * Q s) s"
  shows "gfp_trans T P =
         gfp_in (t P) (weakly_bounded_by P)"
         (is "?X P = ?Y P")
  proof -
    have healthy_gfp: "healthy ?Y"
      proof(rule healthyI)
        show feasible: "feasible ?Y"
          proof(rule feasibleI)
            fix b::real and  P::"'a \<Rightarrow> real"
            assume P_bby: "bounded_by b P"
               and P_nneg: "nneg P"
            hence P_sound: "sound P" by(blast)
              
            with P_bby and P_nneg
            have "bounded_by (bound_of P) (?Y P)"
              apply -
              apply(rule bounded_byI)
              apply(rule_tac x=x in le_funD)
              apply(rule gfp_in_least)
                apply(rule weakly_bounded_bounded_below, assumption)
               apply(rule weakly_bounded_top, assumption)
              using weakly_bounded_bby
                and weakly_bounded_closed by(blast intro:le_funI order_trans)

            with P_bby show "bounded_by b (?Y P)"
              by(blast intro:bounded_by_mono)

            from P_bby and P_sound
            show "nneg (?Y P)"
              apply -
              apply(rule nnegI)
              apply(rule_tac x=x in le_funD)
              apply(rule gfp_in_upperbound)
                apply(rule weakly_bounded_bounded_above, assumption)
               apply(rule weakly_bounded_bot, assumption)
              apply(rule le_funI)
              apply(rule_tac x=xa in nnegD)
              apply(rule sound_nneg)
              using weakly_bounded_bot and weakly_bounded_closed
                 and weakly_bounded_sound
              by(blast intro:weakly_bounded_sound)
          qed

          show "mono_trans ?Y"
          apply -
          apply(rule mono_transI)
          apply(rule order_trans)
           apply(rule_tac T="weakly_bounded_by Q" in gfp_in_mono)
           apply(rule weakly_bounded_mono, assumption+)
           apply(rule weakly_bounded_bounded_below, assumption+)
           apply(rule weakly_bounded_bounded_above, assumption+)
           apply(rule weakly_bounded_bounded_above, assumption+)
           apply(rule weakly_bounded_up_complete, assumption+)
           apply(rule weakly_bounded_closed, assumption+)
           apply(rule weakly_bounded_mono_on, assumption+)
          apply(rule gfp_in_least)
            apply(rule weakly_bounded_bounded_below, assumption)
           apply(rule gfp_in_closed)
              apply(rule weakly_bounded_bounded_below, assumption)
             apply(rule weakly_bounded_bounded_above, assumption)
            apply(rule weakly_bounded_up_complete, assumption)
           apply(rule weakly_bounded_closed, assumption)
           apply(rule gfp_in_upperbound)
            apply(rule weakly_bounded_bounded_above, assumption+)
           apply(rule order_trans, assumption)
           apply(rule weakly_mono, assumption+)
           done

        show "scaling ?Y"

          proof(rule scalingI[OF antisym])
            fix P::"'a \<Rightarrow> real" and c::real and s
            assume sound: "sound P" and pos: "0 < c"
            from sound have
              "\<And>s. gfp_in (t P) (weakly_bounded_by P) s
               \<le> t P (gfp_in (t P) (weakly_bounded_by P)) s"
              (is "\<And>s. ?X P s \<le> t P (?X P) s")
              by(blast intro:le_funD gfp_in_lemma2
                             weakly_bounded_intros
                             weakly_bounded_closed
                             weakly_bounded_mono_on)
            with pos have "\<And>s. c * ?X P s \<le> c * t P (?X P) s"
              by(blast intro:mult_left_mono dest:less_imp_le)
            with pos and sound and feasible
            have "\<And>s. c * ?X P s \<le> t (\<lambda>s. c * P s) (\<lambda>s. c * ?X P s) s"
              by(force simp:semi_sc[symmetric] dest:feasible_sound)
            hence "(\<lambda>s. c * ?X P s) \<le> t (\<lambda>s. c * P s) (\<lambda>s. c * ?X P s)"
              by(rule le_funI)
            moreover
            {
              have "bound_of (\<lambda>s. c * gfp_in (t P) (weakly_bounded_by P) s) \<le>
                    bound_of (\<lambda>s. c * P s)"
              using sound and pos
              apply(simp only: sc_bound[symmetric])
              apply(rule bound_of_least)
              apply(rule bounded_byI)
              apply(rule mult_left_mono)
               apply(rule le_funD[OF gfp_in_least])
                 apply(simp_all add:weakly_bounded_intros sound)
               apply(rule le_funI)
               unfolding weakly_bounded_by_def
               apply(rule_tac y="bound_of xa" in order_trans)
                apply(blast)
               apply(blast)
               done
             moreover have
               "sound (gfp_in (t P) (weakly_bounded_by P))"
               using sound and feasible by(blast)
             ultimately
             have "(\<lambda>s. c * gfp_in (t P) (weakly_bounded_by P) s)
                   \<in> weakly_bounded_by (\<lambda>s. c * P s)"
               using pos by(blast intro:weakly_bounded_byI sc_sound less_imp_le)
            }
            ultimately have "(\<lambda>s. c * ?X P s) \<le> ?X (\<lambda>s. c * P s)"
              using sound and pos
              by(blast intro:weakly_bounded_intros sc_sound less_imp_le
                             gfp_in_upperbound)
            thus "c * ?X P s \<le> ?X (\<lambda>s. c * P s) s"
              by(rule le_funD)
            from sound and pos
            have "\<And>s. ?X (\<lambda>s. c * P s) s \<le> t (\<lambda>s. c * P s) (?X (\<lambda>s. c * P s)) s"
              by(blast intro:le_funD gfp_in_lemma2 sc_sound less_imp_le
                             weakly_bounded_intros
                             weakly_bounded_closed
                             weakly_bounded_mono_on)
            also from pos
            have "\<And>s. ... s =
                  t (\<lambda>s. c * P s) (\<lambda>s. c * (inverse c * ?X (\<lambda>s. c * P s) s)) s"
              by(simp add:mult_assoc[symmetric])
            also from sound and pos and feasible
            have "\<And>s. ... s = c * t P (\<lambda>s. inverse c * ?X (\<lambda>s. c * P s) s) s"
              by(force intro!:semi_sc[symmetric] sc_sound
                              feasible_sound[OF _ sc_sound])
            finally
            have "\<And>s. inverse c * ?X (\<lambda>s. c * P s) s
                    \<le> t P (\<lambda>s. inverse c * ?X (\<lambda>s. c * P s) s) s"
              using pos by(force intro!:mult_div_mono_left)
            moreover {
              from sound and pos
              have "\<And>s. ?X (\<lambda>s. c * P s) s \<le> c * bound_of P"
                by(force intro!:le_funD[OF gfp_in_least]
                                weakly_bounded_intros sc_sound
                                weakly_bounded_byI
                          dest!:weakly_bounded_bby)
              with pos have "\<And>s. inverse c * ?X (\<lambda>s. c * P s) s \<le> bound_of P"
                by(force intro:mult_div_mono_left)
              with pos and sound
              have "bound_of (\<lambda>s. inverse c * ?X (\<lambda>s. c * P s) s) \<le> bound_of P"
                by(force simp:sc_bound
                         intro!:bound_of_least bounded_byI)
              moreover from sound and feasible and pos
              have "sound (\<lambda>s. inverse c * ?X (\<lambda>s. c * P s) s)"
                by(force intro!:feasible_sound[OF _ sc_sound] sc_sound)
              ultimately
              have "(\<lambda>s. inverse c * ?X (\<lambda>s. c * P s) s) \<in>
                    weakly_bounded_by P"
                by(blast intro:weakly_bounded_byI)
            }
            ultimately
            have "(\<lambda>s. inverse c * ?X (\<lambda>s. c * P s) s) \<tturnstile> ?X P"
              using sound by(blast intro:gfp_in_upperbound
                                         weakly_bounded_intros)
            with pos show "?X (\<lambda>s. c * P s) s \<le> c * ?X P s"
              by(force intro:mult_div_mono_right)
          qed
        qed

    have "gfp_trans T P = T ?Y P" (is "?A = ?B")
      proof(rule antisym)
        from healthy_gfp show "?A \<le> ?B"
          apply -
          apply(rule le_transD[OF gfp_trans_least])
             prefer 4 apply(rule sound)
            apply(rule health_pr, assumption)
           apply(rule health_pr, assumption)
          apply(rule le_trans_trans, assumption)
          apply(rule_tac T=T in mono_ttransD)
              apply(rule mono_T)
             prefer 2 apply(assumption)
            prefer 2 apply(assumption)
           apply(rule le_trans_trans, assumption)
           apply(rule le_transI)
           apply(simp add:tailcall)
           apply(drule_tac P=P in le_transD, assumption)
           apply(rule gfp_in_upperbound)
             apply(rule weakly_bounded_bounded_above, assumption)
            apply(rule closed_underD[OF weakly_bounded_closed])
             apply(assumption)
            apply(rule_tac f=u in closed_underD)
             apply(rule weakly_bounded_healthy_closed, assumption)
            apply(rule weakly_bounded_self, assumption)
           apply(rule_tac f="t P" in mono_onD)
              apply(rule weakly_bounded_mono_on)
              apply(assumption)
             apply(rule_tac f=u in closed_underD)
              apply(rule weakly_bounded_healthy_closed, assumption)
             apply(rule weakly_bounded_self, assumption)
            apply(rule_tac f="t P" in closed_underD)
             apply(rule weakly_bounded_closed, assumption)
            apply(rule_tac f=u in closed_underD)
             apply(rule weakly_bounded_healthy_closed, assumption)
            apply(rule weakly_bounded_self, assumption)
           apply(assumption)
           done

        from healthy_gfp show "?B \<le> ?A"
          apply(rule_tac P=P in le_transD)
           prefer 2 apply(rule sound)
          apply(rule gfp_trans_upperbound)
             apply(rule health_pr, assumption)
            apply(rule health_pr, assumption)
           apply(blast intro:mono_ttrans_equiv_trans[OF mono_T])
          apply(rule_tac T=T in mono_ttransD)
             apply(rule mono_T)
            prefer 2 apply assumption
           prefer 2 apply(rule health_pr, assumption)
          apply(rule le_transI)
          apply(subst tailcall)
          apply(rule gfp_in_lemma2)
              apply(rule weakly_bounded_bounded_below, assumption)
             apply(rule weakly_bounded_bounded_above, assumption)
            apply(rule weakly_bounded_up_complete, assumption)
           apply(rule weakly_bounded_closed, assumption)
          apply(rule weakly_bounded_mono_on, assumption)
          done
      qed

    also have "... = t P (?Y P)"
      by(simp add:tailcall)

    also
    have "... = ?Y P"
      by(force intro!:gfp_in_unfold[symmetric]
                      weakly_bounded_intros assms)

    finally show ?thesis .
  qed

lemma lfp_in_weakly_mono:
  assumes weakly_mono:
    "\<And>R. R \<in> weakly_bounded_by Q \<Longrightarrow> t R \<tturnstile> u R"
      and sound: "sound Q"
      and cu_t:  "closed_under t (weakly_bounded_by Q)"
  shows "lfp_in t (weakly_bounded_by Q) \<tturnstile> lfp_in u (weakly_bounded_by Q)"
proof(rule lfp_in_greatest)
  from sound show "bounded_above (weakly_bounded_by Q)"
    by(rule weakly_bounded_intros)
  from sound show "lfp_in t (weakly_bounded_by Q) \<in> weakly_bounded_by Q"
    by(blast intro:lfp_in_closed weakly_bounded_bounded_below
                   weakly_bounded_bounded_above weakly_bounded_down_complete cu_t)
  fix R
  assume Rin: "R \<in> weakly_bounded_by Q" and fp_u: "u R \<tturnstile> R"
  from Rin have "t R \<tturnstile> u R" by(rule weakly_mono)
  also note fp_u
  finally have fp_t: "t R \<tturnstile> R" .
  moreover from sound have "bounded_below (weakly_bounded_by Q)"
    by(rule weakly_bounded_intros)
  moreover note Rin
  ultimately show "lfp_in t (weakly_bounded_by Q) \<tturnstile> R"
    by(blast intro: lfp_in_lowerbound)
qed

end
