(*
 * Copyright (C) 2013 National ICT Australia Ltd ('NICTA')
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Author: David Cock - David.Cock@nicta.com.au *)

theory Misc imports Real begin

section {* Miscellaneous mathematics *}

lemma mult_div_mono_left:
  "\<lbrakk> 0 \<le> (c::real); c \<noteq> 0; a \<le> inverse c * b \<rbrakk> \<Longrightarrow> c * a \<le> b"
  apply(subst mult_1_left[where a=b, symmetric])
  apply(subst right_inverse[symmetric], assumption)
  apply(subst mult_assoc)
  apply(force intro:mult_mono)
  done

lemma mult_div_mono_right:
  "\<lbrakk> 0 \<le> (c::real); c \<noteq> 0; inverse c * a \<le> b \<rbrakk> \<Longrightarrow> a \<le> c * b"
  apply(subst mult_1_left[where a=a, symmetric])
  apply(subst right_inverse[symmetric], assumption)
  apply(subst mult_assoc)
  apply(force intro:mult_mono)
  done

lemma flip_mult_left:
  "\<lbrakk> (c::real) \<noteq> 0; a = inverse c * b \<rbrakk> \<Longrightarrow> c * a = b"
  apply(subst mult_1_left[where a=b, symmetric])
  apply(subst right_inverse[symmetric], assumption)
  apply(subst mult_assoc)
  apply(simp add:mult_cancel_left)
  done

lemma min_distrib:
  "0 \<le> (c::real) \<Longrightarrow> c * min a b = min (c * a) (c * b)"
  apply(case_tac "a \<le> b")
   apply(subgoal_tac "c * a \<le> c * b", simp add:min_def)
   apply(force intro: mult_left_mono)
  apply(subgoal_tac "c * b \<le> c * a", simp add:min_def)
  apply(force intro:mult_left_mono)
  done

lemma nonempty_witness:
  "S \<noteq> {} \<Longrightarrow> \<exists>x. x \<in> S"
  by(blast)

section {* Truncated Subtraction *}

definition
  tminus :: "real \<Rightarrow> real \<Rightarrow> real" (infixl "\<ominus>" 60)
where
  "x \<ominus> y = max (x - y) 0"

lemma tminus_cancel_1:
  "0 \<le> a \<Longrightarrow> a + 1 \<ominus> 1 = a"
  unfolding tminus_def by(simp)

lemma tminus_zero_imp_le:
  "x \<ominus> y \<le> 0 \<Longrightarrow> x \<le> y"
  by(simp add:tminus_def)

lemma tminus_zero[simp]:
  "0 \<le> x \<Longrightarrow> x \<ominus> 0 = x"
  by(simp add:tminus_def)

lemma tminus_left_mono:
  "a \<le> b \<Longrightarrow> a \<ominus> c \<le> b \<ominus> c"
  unfolding tminus_def
  by(case_tac "a \<le> c", simp_all)

lemma tminus_less:
  "\<lbrakk> 0 \<le> a; 0 \<le> b \<rbrakk> \<Longrightarrow> a \<ominus> b \<le> a"
  unfolding tminus_def by(force)

lemma tminus_left_distrib:
  "0 \<le> a \<Longrightarrow> a * (b \<ominus> 1) = a * b \<ominus> a"
  unfolding tminus_def
  apply(case_tac "b - 1 \<le> 0", simp_all add:max_absorb1 max_absorb2)
   apply(subgoal_tac "a * b - a \<le> 0", simp)
   apply(subgoal_tac "a * b \<le> a", simp)
   apply(rule mult_left_mono[where b="1::real", simplified], assumption+)
  apply(subgoal_tac "0 \<le> a * b - a", simp add:max_absorb1)
   apply(simp add:right_diff_distrib)
  apply(simp add:sign_simps)
  apply(rule mult_left_mono[where a="1::real", simplified])
   apply(simp)
  apply(assumption)
  done

lemma tminus_le[simp]:
  "b \<le> a \<Longrightarrow> a \<ominus> b = a - b"
  unfolding tminus_def by(simp)

lemma tminus_nle[simp]:
  "\<not>b \<le> a \<Longrightarrow> a \<ominus> b = 0"
  unfolding tminus_def by(simp)

lemma tminus_add_mono:
  "(a+b) \<ominus> (c+d) \<le> (a\<ominus>c) + (b\<ominus>d)"
proof(cases "0 \<le> a - c")
  case True note pac = this
  show ?thesis
  proof(cases "0 \<le> b - d")
    case True note pbd = this
    from pac and pbd have "(c + d) \<le> (a + b)" by(simp)
    with pac and pbd show ?thesis by(simp)
  next
    case False note nbd = this
    with pac and nbd show ?thesis
    proof(cases "c + d \<le> a + b")
      case True with pac and nbd show ?thesis by(simp)
    next
      case False with pac and nbd show ?thesis by(simp)
    qed
  qed
next
  case False note nac = this
  show ?thesis
  proof(cases "0 \<le> b - d")
    case True note pbd = this
    with nac and pbd show ?thesis
    proof(cases "c + d \<le> a + b")
      case True with nac and pbd show ?thesis by(simp)
    next
      case False with nac and pbd show ?thesis by(simp)
    qed
  next
    case False note nbd = this
    with nac have "\<not>(c + d) \<le> (a + b)" by(simp)
    with nac and nbd show ?thesis by(simp)
  qed
qed

lemma tminus_setsum_mono:
  assumes fS: "finite S"
  shows "setsum f S \<ominus> setsum g S \<le> setsum (\<lambda>x. f x \<ominus> g x) S"
        (is "?X S")
proof(rule finite_induct)
  from fS show "finite S" .

  show "?X {}" by(simp)

  fix x and F
  assume fF: "finite F" and xniF: "x \<notin> F"
     and IH: "?X F"
  have "f x + setsum f F \<ominus> g x + setsum g F \<le>
        (f x \<ominus> g x) + (setsum f F \<ominus> setsum g F)"
    by(rule tminus_add_mono)
  also from IH have "... \<le> (f x \<ominus> g x) + (\<Sum>x\<in>F. f x \<ominus> g x)"
    by(rule add_left_mono)
  finally show "?X (insert x F)"
    by(simp add:setsum_insert[OF fF xniF])
qed

lemma tminus_nneg[simp,intro]:
  "0 \<le> a \<ominus> b"
  by(case_tac "b \<le> a", simp_all)

lemma tminus_right_antimono:
  assumes clb: "c \<le> b"
  shows "a \<ominus> b \<le> a \<ominus> c"
proof(cases "b \<le> a")
  case True
  moreover with clb have "c \<le> a" by(auto)
  moreover note clb
  ultimately show ?thesis by(simp)
next
  case False
  then show ?thesis by(simp)
qed

end
