theory Traces imports pGCL begin

text {* A point distribution on states *}
definition
  delta :: "'s \<Rightarrow> 's \<Rightarrow> real" ("\<Delta>")
where
  "delta x = (\<lambda>y. if x = y then 1 else 0)"

text {* The forward interpretation gives the conditional probability
        P(s_final|s_initial). *}
definition
  fw_c :: "'s prog \<Rightarrow> 's \<Rightarrow> 's \<Rightarrow> real"
where
  "fw_c prog s_i s_f = wp prog (\<Delta> s_f) s_i"

text {* Given a prior distribution, we get the joint distribution
        P(s_final,s_initial) *}
definition
  fw_j :: "'s prog \<Rightarrow> ('s \<Rightarrow> real) \<Rightarrow> ('s \<times> 's) \<Rightarrow> real"
where
  "fw_j prog p_i = (\<lambda>(s_i,s_f). p_i s_i * fw_c prog s_i s_f)"

text {* Map a distribution on initial states to one on final states,
        by marginalising the joint distribution. *}
definition
  step :: "'s prog \<Rightarrow> ('s \<Rightarrow> real) \<Rightarrow> 's \<Rightarrow> real"
where
  "step prog p_i s_f = (\<Sum>s_i\<in>UNIV. fw_j prog p_i (s_i,s_f))"

definition
  Pr :: "('a \<Rightarrow> real) \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'b \<Rightarrow> real"
where
  "Pr p f b = (\<Sum>a|f a = b. p a)"

text {* Joint probability: P(b,c) *}
definition
  Pr_j :: "('a \<Rightarrow> real) \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> ('a \<Rightarrow> 'c) \<Rightarrow> ('b \<times> 'c) \<Rightarrow> real"
where
  "Pr_j p f g bc = (\<Sum>a|f a = (fst bc) \<and> g a = (snd bc). p a)"

text {* Conditional probability: P(b|c) *}
definition
  Pr_c :: "('a \<Rightarrow> real) \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> ('a \<Rightarrow> 'c) \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> real"
where
  "Pr_c p f g b c = Pr_j p f g (b,c) / Pr p g c"

definition
  marj_r :: "(('a \<times> 'b) \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real"
where
  "marj_r f a \<equiv> \<Sum>b\<in>UNIV. f (a,b)"

definition
  marj_l :: "(('a \<times> 'b) \<Rightarrow> real) \<Rightarrow> 'b \<Rightarrow> real"
where
  "marj_l f b \<equiv> \<Sum>a\<in>UNIV. f (a,b)"

definition
  indep :: "('a \<Rightarrow> real) \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> ('a \<Rightarrow> 'c) \<Rightarrow> bool"
where
  "indep p f g \<equiv> \<forall>b c. Pr_j p f g (b,c) = Pr p f b * Pr p g c"

text {* If the prior low and high states are uncorrelated, so
        are the posterior low and prior high states. *}
definition
  noleakage :: "(('s \<times> 's) \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> 'h) \<Rightarrow> ('s \<Rightarrow> 'l) \<Rightarrow> bool"
where
  "noleakage p f_h f_l \<equiv>
   indep p (f_h o fst) (f_l o fst) \<longrightarrow> indep p (f_h o fst) (f_l o snd)"

definition
  nl_step :: "'s prog \<Rightarrow> ('s \<Rightarrow> 'h) \<Rightarrow> ('s \<Rightarrow> 'l) \<Rightarrow> bool"
where
  "nl_step prog f_h f_l \<equiv> \<forall>p_i. noleakage (fw_j prog p_i) f_h f_l"

definition
  nocheating :: "(('s \<times> 's) \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> 'h) \<Rightarrow> ('s \<Rightarrow> 'l) \<Rightarrow> bool"
where
  "nocheating p f_h f_l \<equiv>
   (indep p (f_h o fst) (f_l o fst) \<and> indep p (f_h o fst) (f_l o snd)) \<longrightarrow>
   indep p (f_h o snd) (f_l o snd)"

definition
  nc_step :: "'s prog \<Rightarrow> ('s \<Rightarrow> 'h) \<Rightarrow> ('s \<Rightarrow> 'l) \<Rightarrow> bool"
where
  "nc_step prog f_h f_l \<equiv> \<forall>p_i. nocheating (fw_j prog p_i) f_h f_l"

definition
  additive :: "(('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow> bool"
where
  "additive t \<equiv> \<forall>P Q s. t P s + t Q s = t (\<lambda>s. P s + Q s) s"

lemma additive_setsum:
  fixes f::"'a \<Rightarrow> 's \<Rightarrow> real"
    and S::"'a set"
  assumes add: "additive t"
      and fin: "finite S"
      and h_t: "healthy t"
  shows "(\<Sum>a\<in>S. t (f a) s) = t (\<lambda>s. \<Sum>a\<in>S. f a s) s"
proof(rule finite_induct)
  from fin show "finite S" by(assumption)
  from h_t show "(\<Sum>a\<in>{}. t (f a) s) = t (\<lambda>s. \<Sum>a\<in>{}. f a s) s" by(simp)
  fix A and b
  assume finA: "finite A" and ani: "b \<notin> A"
     and ih: "(\<Sum>a\<in>A. t (f a) s) = t (\<lambda>s. \<Sum>a\<in>A. f a s) s"
  have "(\<Sum>a\<in>insert b A. t (f a) s) = t (f b) s + (\<Sum>a\<in>A. t (f a) s)"
    using finA and ani by(simp add:setsum_insert)
  also
  from ih have "... = t (f b) s + t (\<lambda>s. \<Sum>a\<in>A. f a s) s" by(simp)
  also
  from add have "... = t (\<lambda>s. f b s + (\<Sum>a\<in>A. f a s)) s"
    unfolding additive_def by(simp)
  also from finA and ani
  have "... = t (\<lambda>s. \<Sum>a\<in>insert b A. f a s) s"
    by(simp add:setsum_insert)
  finally
  show "(\<Sum>a\<in>insert b A. t (f a) s) = t (\<lambda>s. \<Sum>a\<in>insert b A. f a s) s" .
qed

lemma delta_sound[intro,simp]:
  "sound (\<Delta> x)"
proof
  have "\<And>y. \<Delta> x y \<le> 1"
    unfolding delta_def by(simp)
  thus "bounded_by 1 (\<Delta> x)" by(blast)
  have "\<And>y. 0 \<le> \<Delta> x y"
    unfolding delta_def by(simp)
  thus "nneg (\<Delta> x)" by(blast)
qed

lemma fw_c_Seq:
  fixes a::"'s::finite prog"
  assumes add: "additive (wp a)"
      and h_a: "healthy (wp a)"
      and h_b: "healthy (wp b)"
  shows "fw_c (a ;; b) sa sc = (\<Sum>sb\<in>UNIV. fw_c a sa sb * fw_c b sb sc)"
  unfolding fw_c_def
proof -
  have "\<And>sb. wp b (\<Delta> sc) sb = 0 \<or> 0 < wp b (\<Delta> sc) sb"
  proof(cases)
    fix sb
    assume "wp b (\<Delta> sc) sb = 0"
    thus  "?thesis sb" by(simp)
  next
    assume

  have "UNIV = {sb. wp b (\<Delta> sc) sb = 0} \<union> {sb. 0 < wp b (\<Delta> sc) sb}"

    have

  proof(rule UNIV_eq_I, simp, cases)
    fix s
    assume "wp b (\<Delta> sc) s = 0"
    thus "wp b (\<Delta> sc) s = 0 \<or> 0 < wp b (

    apply(simp)
  proof(rule UNIV_eq_I)
    apply(safe, simp_all)
    using h_b apply(force)
  proof(simp_all)

  have "(\<Sum>sb\<in>UNIV. wp a (\<Delta> sb) sa * wp b (\<Delta> sc) sb) =
        (\<Sum>sb|0 < wp b (\<Delta> sc) sb. wp a (\<Delta> sb) sa * wp b (\<Delta> sc) sb)"

    apply(simp)

  thm scalingD
thm sub_add_def
oops

lemma nl_step_tr:
  assumes nl_a: "nl_step a f_h f_l"
      and nl_b: "nl_step b f_h f_l"
      and nc_a: "nc_step a f_h f_l"
  shows "nl_step (a ;; b) f_h f_l"
  unfolding nl_step_def noleakage_def
proof
  fix p_i
  assume ind: "indep (fw_j (a ;; b) p_i) (f_h o fst) (f_l o fst)"
  hence "indep (fw_j a p_i) (f_h o fst) (f_l o fst)"
    unfolding indep_def
    sorry
  moreover
  with nl_a have "indep (fw_j a p_i) (f_h o fst) (f_l o snd)"
    unfolding nl_step_def noleakage_def by(blast)
  ultimately
  have "indep (fw_j a p_i) (f_h o snd) (f_l o snd)"
    using nc_a unfolding nc_step_def nocheating_def by(blast)
    

  unfolding nl_step_def fw_j_def fw_c_def nc_step_def
  unfolding wp_eval o_def
  unfolding noleakage_def nocheating_def
  unfolding indep_def
  unfolding Pr_j_def Pr_def
  apply(clarify)

