(*
 * Copyright (C) 2013 National ICT Australia Ltd ('NICTA')
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Author: David Cock - David.Cock@nicta.com.au *)

header "Induction on expectation transformers"

(* XXX - Unitary transformers and expectations.  Locale? *)

theory Transformer_Induction
  imports Transformers Misc Hilbert_Choice "~~/src/HOL/Library/Quotient_Set"
begin

section {* Transformers as a Complete Lattice *}

definition htrans_rel :: "(('s \<Rightarrow> real) \<Rightarrow> 's \<Rightarrow> real) \<Rightarrow>
                          (('s \<Rightarrow> real) \<Rightarrow> 's \<Rightarrow> real) \<Rightarrow> bool"
where [simp]: "htrans_rel t u \<longleftrightarrow> healthy t \<and> healthy u \<and> equiv_trans t u"

quotient_type 's trans = "('s \<Rightarrow> real) \<Rightarrow> 's \<Rightarrow> real" / partial : htrans_rel
  by(force intro!:part_equivpI reflpI sympI transpI le_trans_antisym
             simp:ac_simps)

instantiation trans :: (type) complete_lattice begin

lift_definition
  bot_trans :: "'a trans" is "\<lambda>(P::'a \<Rightarrow> real) (s::'a). (0::real)"
  by(force)

lift_definition
  top_trans :: "'a trans" is "\<lambda>P s. bound_of (P::'a \<Rightarrow> real)"
  by(simp, safe intro!:healthyI feasibleI mono_transI scalingI,
     (blast intro:le_funI sc_bound)+)

lift_definition
  less_eq_trans :: "'a trans \<Rightarrow> 'a trans \<Rightarrow> bool" is le_trans
  by(force)

lift_definition
  less_trans :: "'a trans \<Rightarrow> 'a trans \<Rightarrow> bool" is l_trans
  unfolding l_trans_def by(force)

lemma inf_trans_le1[intro]: "le_trans (inf x y) x"
  by(simp add:le_trans_def inf_fun_def le_fun_def)

lemma inf_trans_le2[intro]: "le_trans (inf x y) y"
  by(simp add:le_trans_def inf_fun_def le_fun_def)

lemma le_inf_trans[intro]:
  "\<lbrakk> le_trans x a; le_trans x b \<rbrakk> \<Longrightarrow> le_trans x (inf a b)"
  by(simp add:le_trans_def inf_fun_def le_fun_def)

lemma inf_trans_mono[intro]:
  "\<lbrakk> le_trans a b; le_trans c d \<rbrakk> \<Longrightarrow> le_trans (inf a c) (inf b d)"
  unfolding inf_fun_def
  by(blast intro:le_funI inf_mono[OF le_funD le_funD])+

lemma healthy_inf[intro,simp]:
  assumes h_t: "healthy t"
      and h_u: "healthy u"
  shows "healthy (inf t u)"
proof
  show "feasible (inf t u)" unfolding inf_fun_def
  proof(safe intro!:feasibleI bounded_byI nnegI)
    fix b::real and P::"'b \<Rightarrow> real" and x
    assume bby: "bounded_by b P" and nneg: "nneg P"

    have "inf (t P x) (u P x) \<le> t P x" by(force intro:inf_le1)
    also from bby and nneg and h_t have "... \<le> b" by(blast)
    finally show "inf (t P x) (u P x) \<le> b" .

    from bby and nneg and h_t have "\<And>x. 0 \<le> t P x" by(blast)
    moreover from bby and nneg and h_u have "\<And>x. 0 \<le> u P x" by(blast)
    ultimately show "0 \<le> inf (t P x) (u P x)" by(rule inf_greatest)
  qed

  show "mono_trans (inf t u)" unfolding inf_fun_def
  proof(safe intro!:mono_transI le_funI)
    fix P::"'b \<Rightarrow> real" and Q and x
    assume s_P: "sound P" and s_Q: "sound Q" and le: "P \<le> Q"
    from h_t and s_P and s_Q and le have "t P x \<le> t Q x"
      by(blast dest:le_funD)
    moreover
    from h_u and s_P and s_Q and le have "u P x \<le> u Q x"
      by(blast dest:le_funD)
    ultimately show "inf (t P x) (u P x) \<le> inf (t Q x) (u Q x)"
      by(rule inf_mono)
  qed

  show "scaling (inf t u)" unfolding inf_fun_def
  proof
    fix P::"'b \<Rightarrow> real" and c::real and x
    assume s_P: "sound P" and pos: "0 < c"
    hence "c * inf (t P x) (u P x) = inf (c * t P x) (c * u P x)" (is ?X)
    proof(cases "t P x \<le> u P x")
      case True then show ?X
        by(simp add:mult_left_mono less_imp_le pos inf_absorb1)
    next
      case False then show ?X
        by(simp add:mult_left_mono less_imp_le pos inf_absorb2)
    qed
    with s_P and pos and h_t and h_u show "c * inf (t P x) (u P x) =
      inf (t (\<lambda>x. c * P x) x) (u (\<lambda>x. c * P x) x)"
      by(simp, blast intro:arg_cong2)
  qed
qed

lift_definition
  inf_trans :: "'a trans \<Rightarrow> 'a trans \<Rightarrow> 'a trans" is inf
  by(force intro!:le_trans_antisym inf_trans_mono simp:ac_simps)

lemma sup_trans_ge1[intro]: "le_trans x (sup x y)"
  by(simp add:le_trans_def sup_fun_def le_fun_def)

lemma sup_trans_ge2[intro]: "le_trans y (sup x y)"
  by(simp add:le_trans_def sup_fun_def le_fun_def)

lemma ge_sup_trans[intro]:
  "\<lbrakk> le_trans a x; le_trans b x \<rbrakk> \<Longrightarrow> le_trans (sup a b) x"
  by(simp add:le_trans_def sup_fun_def le_fun_def)

lemma sup_trans_mono[intro]:
  "\<lbrakk> le_trans a b; le_trans c d \<rbrakk> \<Longrightarrow> le_trans (sup a c) (sup b d)"
  unfolding sup_fun_def
  by(blast intro:le_funI sup_mono[OF le_funD le_funD])+

lemma healthy_sup[intro,simp]:
  assumes h_t: "healthy t"
      and h_u: "healthy u"
  shows "healthy (sup t u)"
proof
  show "feasible (sup t u)" unfolding sup_fun_def
  proof(safe intro!:feasibleI bounded_byI nnegI)
    fix b::real and P::"'b \<Rightarrow> real" and x
    assume bby: "bounded_by b P" and nneg: "nneg P"

    from bby and nneg and h_t have "0 \<le> t P x" by(blast)
    also have "t P x \<le> sup (t P x) (u P x)" by(rule sup_ge1)
    finally show "0 \<le> ..." .

    from bby and nneg and h_t have "\<And>x. t P x \<le> b" by(blast)
    moreover from bby and nneg and h_u have "\<And>x. u P x \<le> b" by(blast)
    ultimately show "sup (t P x) (u P x) \<le> b" by(rule sup_least)
  qed

  show "mono_trans (sup t u)" unfolding sup_fun_def
  proof(safe intro!:mono_transI le_funI)
    fix P::"'b \<Rightarrow> real" and Q and x
    assume s_P: "sound P" and s_Q: "sound Q" and le: "P \<le> Q"
    from h_t and s_P and s_Q and le have "t P x \<le> t Q x"
      by(blast dest:le_funD)
    moreover
    from h_u and s_P and s_Q and le have "u P x \<le> u Q x"
      by(blast dest:le_funD)
    ultimately show "sup (t P x) (u P x) \<le> sup (t Q x) (u Q x)"
      by(rule sup_mono)
  qed

  show "scaling (sup t u)" unfolding sup_fun_def
  proof
    fix P::"'b \<Rightarrow> real" and c::real and x
    assume s_P: "sound P" and pos: "0 < c"
    hence "c * sup (t P x) (u P x) = sup (c * t P x) (c * u P x)" (is ?X)
    proof(cases "t P x \<le> u P x")
      case True then show ?X
        by(simp add:mult_left_mono less_imp_le pos sup_absorb2)
    next
      case False then show ?X
        by(simp add:mult_left_mono less_imp_le pos sup_absorb1)
    qed
    with s_P and pos and h_t and h_u show "c * sup (t P x) (u P x) =
      sup (t (\<lambda>x. c * P x) x) (u (\<lambda>x. c * P x) x)"
      by(simp, blast intro:arg_cong2)
  qed
qed

lift_definition
  sup_trans :: "'a trans \<Rightarrow> 'a trans \<Rightarrow> 'a trans" is sup
  by(force intro!:le_trans_antisym sup_trans_mono simp:ac_simps)

lemma healthy_top[simp,intro!]:
  "healthy (\<lambda>P s. bound_of P)"
  by(blast intro:le_funI sc_bound)

lemma healthy_bot[simp,intro!]:
  "healthy (\<lambda>P s. 0)"
  by(force)

lemma set_rel_empty_right[dest]:
  "\<And>R S. set_rel R S {} \<Longrightarrow> S = {}"
  unfolding set_rel_def by(blast)

lemma set_rel_empty_left[dest]:
  "\<And>R S. set_rel R {} S \<Longrightarrow> S = {}"
  unfolding set_rel_def by(blast)

definition Inf_trans' ::
  "(('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real) set \<Rightarrow> (('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real)"
where
  "Inf_trans' S \<equiv> if S = {}
                  then (\<lambda>P s. bound_of P)
                  else (\<lambda>P s. glb {t P s|t. t \<in> S})"

lemma glb_healthy:
  assumes ne: "S \<noteq> {}"
      and h:  "\<forall>t\<in>S. healthy t"
  shows "healthy (\<lambda>P s. glb {t P s |t. t \<in> S})"
  (is "healthy (\<lambda>P s. glb (?S P s))")
proof(safe intro!:healthyI feasibleI mono_transI scalingI bounded_byI
                  nnegI le_funI)
  fix b::real and P::"'b \<Rightarrow> real" and s
  assume bby: "bounded_by b P" and nneg: "nneg P"

  from ne obtain t where tin: "t \<in> S" by(blast)

  show "glb (?S P s) \<le> b"
  proof -
    from h and bby and nneg have "\<forall>x \<in> ?S P s. 0 \<le> x" by(blast)
    with tin have "glb (?S P s) \<le> t P s" by(blast intro:glb_lower)
    also with h and tin and bby and nneg have "t P s \<le> b" by(blast)
    finally show ?thesis .
  qed

  from h and bby and nneg and tin show "0 \<le> glb (?S P s)"
    by(blast intro:glb_greatest)

next
  fix P::"'b \<Rightarrow> real" and Q and s
  assume sP: "sound P" and sQ: "sound Q" and le: "P \<le> Q"

  show "glb (?S P s) \<le> glb (?S Q s)"
  proof(rule glb_under_glb)
    from ne show "\<exists>x. x \<in> ?S Q s" by(blast)

    from h and sP have "\<forall>y \<in> ?S P s. 0 \<le> y" by(blast)
    thus "\<exists>b. \<forall>y \<in> ?S P s. b \<le> y" by(blast)

    show "\<forall>x \<in> ?S Q s. \<exists>y \<in> ?S P s. y \<le> x"
    proof
      fix x assume xin: "x \<in> ?S Q s"
      from this obtain t where tin: "t \<in> S" and xeq: "t Q s = x" by(blast)
      with h have "healthy t" by(blast)
      with sP and sQ and le have "t P \<le> t Q" by(blast)
      hence "t P s \<le> t Q s" by(blast)
      with xeq have "t P s \<le> x" by(blast)
      moreover from tin have "t P s \<in> ?S P s" by(blast)
      ultimately show "\<exists>y \<in> ?S P s. y \<le> x" by(blast)
    qed
  qed

next
  fix P::"'b \<Rightarrow> real" and c::real and s
  assume sP: "sound P" and pos: "0 < c"

  from ne obtain t where tin: "t \<in> S" by(blast)

  show "c * glb (?S P s) = glb (?S (\<lambda>x. c * P x) s)"
  proof -
    from h and sP have bot1: "\<forall>x \<in> ?S P s. 0 \<le> x" by(blast)
    from h and sP and pos
    have bot2: "\<forall>x \<in> {c * t P s |t. t \<in> S}. 0 \<le> x"
      by(blast intro:mult_nonneg_nonneg less_imp_le)
    from pos have nneg: "0 \<le> c" by(rule less_imp_le)
    from pos have nz:   "c \<noteq> 0" by(blast)
    from nneg have inneg: "0 \<le> inverse c" by(simp)
    from nz have inz: "inverse c \<noteq> 0" by(simp)
    note c_facts = pos nneg nz inneg inz
    have A: "c * glb (?S P s) =
             glb {c * t P s |t. t \<in> S}" (is "c * glb ?X = glb ?Y")
    proof -
      from c_facts have "\<And>t. inverse c * (c * t P s) = t P s" by(simp)
      with c_facts and bot1 have "\<And>x. x \<in> ?Y \<Longrightarrow> c * glb ?X \<le> x"
        by(blast intro:mult_div_mono_left glb_lower)
      with tin have L: "c * glb ?X \<le> glb ?Y" by(blast intro:glb_greatest)

      from bot2 have "\<And>t. t \<in> S \<Longrightarrow> glb ?Y \<le> c * t P s" by(blast intro:glb_lower)
      with c_facts have "\<And>t. t \<in> S \<Longrightarrow> inverse c * glb ?Y \<le> t P s"
        by(simp add:mult_div_mono_left)
      with c_facts tin have "glb ?Y \<le> c * glb ?X"
        by(blast intro:mult_div_mono_right glb_greatest)

      with L show ?thesis by(rule antisym)
    qed
          
    from sP and pos and h
    have "{c * t P s | t. t \<in> S} = ?S (\<lambda>x. c * P x) s" by(blast intro:sym)
             
    hence B: "glb {c * t P s |t. t \<in> S} = glb (?S (\<lambda>x. c * P x) s)"
      by(simp)

    from A and B show ?thesis by(simp)
  qed
qed

lemma htrans_rel_glb:
  assumes neS:   "S \<noteq> {}"
      and neT:   "T \<noteq> {}"
      and sr:    "set_rel htrans_rel S T"
      and sound: "sound P"
  shows "glb {t P s |t. t \<in> S} = glb {t P s |t. t \<in> T}" (is "glb ?X = glb ?Y")
proof(rule antisym)
  from sr have hS: "\<And>t. t \<in> S \<Longrightarrow> healthy t" by(force simp:set_rel_def)
  from sr have hT: "\<And>t. t \<in> T \<Longrightarrow> healthy t" by(force simp:set_rel_def)

  from sound and hS have botX: "\<And>x. x \<in> ?X \<Longrightarrow> 0 \<le> x" by(blast)
  from sound and hT have botY: "\<And>x. x \<in> ?Y \<Longrightarrow> 0 \<le> x" by(blast)

  have "\<And>t. t \<in> T \<Longrightarrow> glb ?X \<le> t P s" proof -
    fix t assume tin: "t \<in> T"
    from tin and sr obtain t' where sin: "t' \<in> S" and rst: "htrans_rel t' t"
      unfolding set_rel_def by(blast intro:bexE)
    hence est: "equiv_trans t' t" by(simp)
    from sin and botX have "glb ?X \<le> t' P s" by(blast intro:glb_lower)
    also with sound and sin and tin and sound and hS and hT and est
    have "... \<le> t P s"
      by(blast dest!:le_funD[OF le_transD, OF equiv_imp_le])
    finally show "glb ?X \<le> ..." .
  qed

  with neT show "glb ?X \<le> glb ?Y" by(blast intro:glb_greatest)

  have "\<And>t. t \<in> S \<Longrightarrow> glb ?Y \<le> t P s" proof -
    fix t' assume sin: "t' \<in> S"
    from sin and sr obtain t where tin: "t \<in> T" and rst: "htrans_rel t' t"
      unfolding set_rel_def by(blast intro:bexE)
    hence est: "equiv_trans t t'" by(simp add:ac_simps)
    from tin and botY have "glb ?Y \<le> t P s" by(blast intro:glb_lower)
    also with sound and sin and tin and sound and hS and hT and est
    have "... \<le> t' P s"
      by(blast dest!:le_funD[OF le_transD, OF equiv_imp_le])
    finally show "glb ?Y \<le> ..." .
  qed

  with neS show "glb ?Y \<le> glb ?X" by(blast intro:glb_greatest)
qed

lift_definition
  Inf_trans :: "'a trans set \<Rightarrow> 'a trans" is Inf_trans'
proof -
  fix set1::"(('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real) set" and set2
  assume sr: "set_rel htrans_rel set1 set2"

  show "htrans_rel (local.Inf_trans' set1) (local.Inf_trans' set2)"
  proof(cases "set1 = {}")
    case True
    moreover hence "set2 = {}" using sr by(blast)
    ultimately show ?thesis by(simp add:Inf_trans'_def)
  next
    case False
    note ne1 = this
    moreover from ne1 obtain t1 where "t1 \<in> set1" by(blast)
    moreover from ne1 have ne2: "set2 \<noteq> {}" using sr by(blast)
    moreover from ne2 obtain t2 where "t2 \<in> set2" by(blast)
    moreover from sr have "\<forall>t\<in>set1. healthy t" unfolding set_rel_def by(force)
    moreover from sr have "\<forall>t\<in>set2. healthy t" unfolding set_rel_def by(force)
    moreover note sr
    ultimately show ?thesis
      by(force intro!:htrans_rel_glb ext equiv_transI
               simp:Inf_trans'_def glb_healthy)
  qed
qed

definition Sup_trans' ::
  "(('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real) set \<Rightarrow> (('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real)"
where
  "Sup_trans' S \<equiv> if S = {} then (\<lambda>P s. 0) else (\<lambda>P s. lub {t P s |t. t \<in> S})"

lemma lub_healthy:
  assumes ne: "S \<noteq> {}"
      and h:  "\<forall>t\<in>S. healthy t"
  shows "healthy (\<lambda>P s. lub {t P s |t. t \<in> S})"
  (is "healthy (\<lambda>P s. lub (?S P s))")
proof(safe intro!:healthyI feasibleI mono_transI scalingI bounded_byI
                  nnegI le_funI)
  fix b::real and P::"'b \<Rightarrow> real" and s
  assume bby: "bounded_by b P" and nneg: "nneg P"

  from ne obtain t where tin: "t \<in> S" by(blast)

  from h and bby and nneg have top: "\<forall>x \<in> ?S P s. x \<le> b" by(blast)
  with tin show "lub (?S P s) \<le> b" by(blast intro:lub_least)

  show "0 \<le> lub (?S P s)"
  proof -
    from h and bby and nneg and tin have "0 \<le> t P s" by(blast)
    also with tin and top have "... \<le> lub (?S P s)"
      by(blast intro:lub_greater)
    finally show ?thesis .
  qed

next
  fix P::"'b \<Rightarrow> real" and Q and s
  assume sP: "sound P" and sQ: "sound Q" and le: "P \<le> Q"

  from ne obtain t where tin: "t \<in> S" by(blast)

  from sQ and h have top: "\<forall>a\<in>{t Q s |t. t \<in> S}. a \<le> bound_of Q" by(blast)

  show "lub (?S P s) \<le> lub (?S Q s)"
  proof(rule lub_least, blast intro:tin, safe)
    fix t' assume sin: "t' \<in> S"

    from sP and sQ and h and le and sin have "t' P s \<le> t' Q s"
      by(blast intro:le_funD)
    also from top and sin have "... \<le> lub {t' Q s |t'. t' \<in> S}"
      by(blast intro:lub_greater)
    finally show "t' P s \<le> ..." .
  qed

next
  fix P::"'b \<Rightarrow> real" and c::real and s
  assume sP: "sound P" and pos: "0 < c"

  from ne obtain t where tin: "t \<in> S" by(blast)

  show "c * lub (?S P s) = lub (?S (\<lambda>x. c * P x) s)"
  proof -
    from h and sP have top1: "\<forall>x \<in> ?S P s. x \<le> bound_of P" by(blast)
    from h and sP and pos
    have top2: "\<forall>x \<in> {c * t P s |t. t \<in> S}. x \<le> c * bound_of P"
      by(blast intro!:mult_left_mono dest:less_imp_le)
    from pos have nneg: "0 \<le> c" by(rule less_imp_le)
    from pos have nz:   "c \<noteq> 0" by(blast)
    from nneg have inneg: "0 \<le> inverse c" by(simp)
    from nz have inz: "inverse c \<noteq> 0" by(simp)
    note c_facts = pos nneg nz inneg inz
    have A: "c * lub (?S P s) =
             lub {c * t P s |t. t \<in> S}" (is "c * lub ?X = lub ?Y")
    proof -
      from c_facts have "\<And>t. inverse c * (c * t P s) = t P s" by(simp)
      from top2 have "\<And>t. t \<in> S \<Longrightarrow> c * t P s \<le> lub ?Y"
        by(blast intro:lub_greater)
      hence "\<And>x. x \<in> ?X \<Longrightarrow> c * x \<le> lub ?Y" by(blast)
      with c_facts have "\<And>x. x \<in> ?X \<Longrightarrow> x \<le> inverse c * lub ?Y"
        by(simp add:mult_div_mono_right)
      with tin have "lub ?X \<le> inverse c * lub ?Y"
        by(blast intro:lub_least)
      with c_facts have L:"c * lub ?X \<le> lub ?Y" by(simp add:mult_div_mono_left)

      from tin and c_facts and top1 have "lub ?Y \<le> c * lub ?X"
        by(blast intro:lub_least mult_left_mono lub_greater)

      with L show ?thesis by(rule antisym)
    qed
          
    from sP and pos and h
    have "{c * t P s | t. t \<in> S} = ?S (\<lambda>x. c * P x) s" by(blast intro:sym)
             
    hence B: "lub {c * t P s |t. t \<in> S} = lub (?S (\<lambda>x. c * P x) s)"
      by(simp)

    from A and B show ?thesis by(simp)
  qed
qed

lemma htrans_rel_lub:
  assumes neS:   "S \<noteq> {}"
      and neT:   "T \<noteq> {}"
      and sr:    "set_rel htrans_rel S T"
      and sound: "sound P"
  shows "lub {t P s |t. t \<in> S} = lub {t P s |t. t \<in> T}" (is "lub ?X = lub ?Y")
proof(rule antisym)
  from sr have hS: "\<And>t. t \<in> S \<Longrightarrow> healthy t" by(force simp:set_rel_def)
  from sr have hT: "\<And>t. t \<in> T \<Longrightarrow> healthy t" by(force simp:set_rel_def)

  from sound and hS have topX: "\<And>x. x \<in> ?X \<Longrightarrow> x \<le> bound_of P" by(blast)
  from sound and hT have topY: "\<And>x. x \<in> ?Y \<Longrightarrow> x \<le> bound_of P" by(blast)

  have "\<And>t. t \<in> S \<Longrightarrow> t P s \<le> lub ?Y"
  proof -
    fix t' assume sin: "t' \<in> S"
    from sin and sr obtain t where tin: "t \<in> T" and rst: "htrans_rel t' t"
      unfolding set_rel_def by(blast intro:bexE)
    hence est: "equiv_trans t' t" by(simp add:ac_simps)
    with sound and sin and tin and sound and hS and hT and est
    have "t' P s \<le> t P s"
      by(blast dest!:le_funD[OF le_transD, OF equiv_imp_le])
    also from tin and topY have "... \<le> lub ?Y" by(blast intro:lub_greater)
    finally show "t' P s \<le> ..." .
  qed

  with neS show "lub ?X \<le> lub ?Y" by(blast intro:lub_least)

  have "\<And>t. t \<in> T \<Longrightarrow> t P s \<le> lub ?X" proof -
    fix t assume tin: "t \<in> T"
    from tin and sr obtain t' where sin: "t' \<in> S" and rst: "htrans_rel t' t"
      unfolding set_rel_def by(blast intro:bexE)
    hence est: "equiv_trans t t'" by(simp add:ac_simps)
    with sound and sin and tin and sound and hS and hT and est
    have "t P s \<le> t' P s"
      by(blast dest!:le_funD[OF le_transD, OF equiv_imp_le])
    also from sin and topX have "... \<le> lub ?X" by(blast intro:lub_greater)
    finally show "t P s \<le> ..." .
  qed

  with neT show "lub ?Y \<le> lub ?X" by(blast intro:lub_least)
qed

lift_definition
  Sup_trans :: "'a trans set \<Rightarrow> 'a trans" is Sup_trans'
proof -
  fix set1::"(('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real) set" and set2
  assume sr: "set_rel htrans_rel set1 set2"

  show "htrans_rel (local.Sup_trans' set1) (local.Sup_trans' set2)"
  proof(cases "set1 = {}")
    case True
    moreover hence "set2 = {}" using sr by(blast)
    ultimately show ?thesis by(simp add:Sup_trans'_def)
  next
    case False
    note ne1 = this
    moreover from ne1 obtain t1 where "t1 \<in> set1" by(blast)
    moreover from ne1 have ne2: "set2 \<noteq> {}" using sr by(blast)
    moreover from ne2 obtain t2 where "t2 \<in> set2" by(blast)
    moreover from sr have "\<forall>t\<in>set1. healthy t" unfolding set_rel_def by(force)
    moreover from sr have "\<forall>t\<in>set2. healthy t" unfolding set_rel_def by(force)
    moreover note sr
    ultimately show ?thesis
      by(force intro!:htrans_rel_lub ext equiv_transI
               simp:Sup_trans'_def lub_healthy)
  qed
qed

lemma healthy_rep_abs:
  "healthy t \<Longrightarrow> equiv_trans (rep_trans (abs_trans t)) t"
  using Quotient_rep_abs[OF Quotient_trans]
  unfolding htrans_rel_def by(blast)

lemma healthy_rep:
  "healthy (rep_trans t)"
  using Quotient_rep_reflp[OF Quotient_trans]
  unfolding htrans_rel_def by(blast)

(* XXX - a generic version of this should be in the transfer package *)
lemma tr_set_trans[transfer_rule]:
 "(cr_trans ===> set_rel cr_trans ===> op =)
  (\<lambda>x X. \<exists>x'\<in>X. htrans_rel x x') op \<in>"
proof((rule fun_relI)+, rule iffI, safe)
  fix x and x' and y and X and Y
  assume cr: "cr_trans x y" and sr: "set_rel cr_trans X Y"
     and x2in: "x' \<in> X" and hr: "htrans_rel x x'"

  hence abs_eq: "abs_trans x = abs_trans x'"
    by(blast intro:Quotient_rel_abs[OF Quotient_trans])
  with cr and hr have "cr_trans x' y"
    unfolding cr_trans_def by(simp)
  with x2in and sr obtain y' where y2in: "y' \<in> Y" and cr2: "cr_trans x' y'"
    unfolding set_rel_def by(blast)
  from cr2 and cr and abs_eq have "y' = y"
    unfolding cr_trans_def by(simp)
  with y2in show "y \<in> Y" by(simp)
next
  fix x and y and X and Y
  assume cr: "cr_trans x y" and sr: "set_rel cr_trans X Y" and yin: "y \<in> Y"
  from yin and sr obtain x' where x2in: "x' \<in> X" and cr2: "cr_trans x' y"
    unfolding set_rel_def by(blast)
  have "htrans_rel x x'"
    apply(rule iffD1[OF Quotient_rel[OF Quotient_trans]])
    using cr and cr2 apply(simp add:cr_trans_def htrans_rel_def)
    done
  with x2in show "\<exists>x'\<in>X. htrans_rel x x'" by(blast)
qed

instance proof
  fix x::"'a trans" and y
  show "(x < y) = (x \<le> y \<and> \<not> y \<le> x)"
    by(transfer, unfold l_trans_def, simp)
next
  fix x::"'a trans" show "x \<le> x"
    by(transfer, simp)
next
  fix x::"'a trans" and y and z
  assume "x \<le> y" and "y \<le> z" thus "x \<le> z"
    by(transfer, blast intro:le_trans_trans)
next
  fix x::"'a trans" and y
  assume "x \<le> y" and "y \<le> x" thus "x = y"
    by(transfer, simp, blast intro!:le_trans_antisym)
next
  fix x::"'a trans" and y
  show "inf x y \<le> x"
      by(transfer, simp add:inf_fun_def, blast intro:inf_le1)
  show "inf x y \<le> y"
      by(transfer, simp add:inf_fun_def, blast intro:inf_le2)
  fix z assume "x \<le> y" and "x \<le> z" thus "x \<le> inf y z"
    by(transfer, simp add:inf_fun_def,
       blast intro!:le_transI[OF le_funI, OF le_infI] dest!:le_transD)
next
  fix x::"'a trans" and y
  show "x \<le> sup x y"
    by(transfer, simp add:sup_fun_def, blast intro:sup_ge1)
  show "y \<le> sup x y"
    by(transfer, simp add:sup_fun_def, blast intro:sup_ge2)
  fix z assume "y \<le> x" and "z \<le> x" thus "sup y z \<le> x"
    by(transfer, simp add:sup_fun_def,
       blast intro!:le_transI[OF le_funI, OF le_supI] dest!:le_transD)
next
  fix a::"'a trans"
  show "bot \<le> a" by(transfer, simp)
  show "a \<le> top" by(transfer, simp)
next
  fix x::"'a trans" and A assume xin: "x \<in> A"
  hence ne: "A \<noteq> {}" by(blast)
  from xin show "Inf A \<le> x"
  proof(transfer, clarsimp)
    fix x::"('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real" and x' and X and Y
    assume x2in: "x' \<in> X" and eq: "equiv_trans x x'"
       and hx2: "healthy x'"
       and sr: "set_rel cr_trans X Y"
    from x2in have neA: "X \<noteq> {}" by(blast)
    from sr have hX: "\<forall>x\<in>X. healthy x"
      by(simp add:cr_trans_def set_rel_def)
    from x2in and hx2 show "le_trans (Inf_trans' X) x'"
      apply -
      apply(rule le_transI)
      apply(simp add:Inf_trans'_def neA)
      apply(rule le_funI)
      apply(rule glb_lower)
       apply(blast)
      apply(rule_tac x=0 in exI)
      apply(clarify)
      using hX apply(force)
      done
  qed
  from xin show "x \<le> Sup A"
  proof(transfer, clarsimp)
    fix x::"('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real" and x' and X and Y
    assume x2in: "x' \<in> X" and eq: "equiv_trans x x'"
       and hx2: "healthy x'"
       and sr: "set_rel cr_trans X Y"
    from x2in have neA: "X \<noteq> {}" by(blast)
    from sr have hX: "\<forall>x\<in>X. healthy x"
      by(simp add:cr_trans_def set_rel_def)
    from x2in and hx2 show "le_trans x' (Sup_trans' X)"
      apply -
      apply(rule le_transI)
      apply(simp add:Sup_trans'_def neA)
      apply(rule le_funI)
      apply(rule lub_greater)
       apply(blast)
      apply(rule_tac x="bound_of P" in exI)
      apply(clarify)
      apply(rule_tac x=x in bounded_byD)
      apply(subgoal_tac "healthy t")
       apply(blast)
      using hX apply(simp)
      done
  qed
next
  fix z::"'a trans" and A
  assume "\<And>x. x \<in> A \<Longrightarrow> z \<le> x"
  thus "z \<le> Inf A"
  proof(transfer, clarsimp)
    fix z::"('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real" and X and Y
    assume hz: "healthy z"
      and IH: "\<And>x. \<lbrakk> healthy x; \<exists>x'\<in>X. healthy x' \<and> equiv_trans x x' \<rbrakk> \<Longrightarrow>
                   le_trans z x"
      and sr: "set_rel cr_trans X Y"
    show "le_trans z (Inf_trans' X)"
    proof(cases "X = {}")
      case True
      with hz show ?thesis by(simp add:Inf_trans'_def)
    next
      case False note ne=this

      from sr have hX: "\<forall>x\<in>X. healthy x"
        by(simp add:cr_trans_def set_rel_def)

      have "\<And>P s. sound P \<Longrightarrow> z P s \<le> glb {t P s |t. t \<in> X}"
      proof(rule glb_greatest)
        fix P::"'a \<Rightarrow> real" and s::'a
        assume sP: "sound P"
        from ne show "\<exists>y. y \<in> {t P s |t. t \<in> X}" by(blast)
        have "\<And>t. t \<in> X \<Longrightarrow> z P s \<le> t P s"
        proof -
          fix t assume tin: "t \<in> X"
          with hX have ht: "healthy t" by(simp)
          moreover {
            from ht have "equiv_trans t t"
              by(simp add:equiv_trans_def)
            with ht and tin have "\<exists>t' \<in> X. healthy t' \<and> equiv_trans t t'"
              by(blast)
          }
          ultimately have "le_trans z t" by(rule IH)
          with sP show "z P s \<le> t P s" by(blast)
        qed
        thus "\<forall>a\<in>{t P s |t. t \<in> X}. z P s \<le> a" by(blast)
      qed

      hence "le_trans z (\<lambda>P s. glb {t P s |t. t \<in> X})"
        by(blast)

      with ne show ?thesis
        unfolding Inf_trans'_def by(simp)
    qed
  qed
next
  fix z::"'a trans" and A
  assume "\<And>x. x \<in> A \<Longrightarrow> x \<le> z"
  thus "Sup A \<le> z"
  proof(transfer, clarsimp)
    fix z::"('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real" and X and Y
    assume hz: "healthy z"
      and IH: "\<And>x. \<lbrakk> healthy x; \<exists>x'\<in>X. healthy x' \<and> equiv_trans x x' \<rbrakk> \<Longrightarrow>
                   le_trans x z"
      and sr: "set_rel cr_trans X Y"
    show "le_trans (Sup_trans' X) z"
    proof(cases "X = {}")
      case True
      with hz show ?thesis by(simp add:Sup_trans'_def)
    next
      case False note ne=this

      from sr have hX: "\<forall>x\<in>X. healthy x"
        by(simp add:cr_trans_def set_rel_def)

      have "\<And>P s. sound P \<Longrightarrow> lub {t P s |t. t \<in> X} \<le> z P s"
      proof(rule lub_least)
        fix P::"'a \<Rightarrow> real" and s::'a
        assume sP: "sound P"
        from ne show "\<exists>y. y \<in> {t P s |t. t \<in> X}" by(blast)
        have "\<And>t. t \<in> X \<Longrightarrow> t P s \<le> z P s"
        proof -
          fix t assume tin: "t \<in> X"
          with hX have ht: "healthy t" by(simp)
          moreover {
            from ht have "equiv_trans t t"
              by(simp add:equiv_trans_def)
            with ht and tin have "\<exists>t' \<in> X. healthy t' \<and> equiv_trans t t'"
              by(blast)
          }
          ultimately have "le_trans t z" by(rule IH)
          with sP show "t P s \<le> z P s" by(blast)
        qed
        thus "\<forall>a\<in>{t P s |t. t \<in> X}. a \<le> z P s" by(blast)
      qed

      hence "le_trans (\<lambda>P s. lub {t P s |t. t \<in> X}) z"
        by(blast)

      with ne show ?thesis
        unfolding Sup_trans'_def by(simp)
    qed
  qed
qed
end

subsection {* Lifting and Lowering Transformers *}

lemma healthy_rep_trans[simp,intro!]:
  "healthy (rep_trans t)"
  using Quotient_rep_reflp[OF Quotient_trans] by(force)

lemma rep_abs_trans[simp]:
  "healthy t \<Longrightarrow> equiv_trans (rep_trans (abs_trans t)) t"
  using Quotient_rep_abs[OF Quotient_trans] by(force)

lemma abs_rep_trans[simp]:
  "abs_trans (rep_trans t) = t"
  using Quotient_abs_rep[OF Quotient_trans] by(force)

lemma abs_trans_equiv:
  "\<lbrakk> equiv_trans t u; healthy t; healthy u \<rbrakk> \<Longrightarrow> abs_trans t = abs_trans u"
  by(blast intro: Quotient_rel_abs[OF Quotient_trans, simplified])

lemma equiv_trans_trans[trans]:
  assumes xy: "equiv_trans x y"
      and yz: "equiv_trans y z"
  shows "equiv_trans x z"
proof(rule le_trans_antisym)
  from xy have "le_trans x y" by(blast)
  also from yz have "le_trans y z" by(blast)
  finally show "le_trans x z" .
  from yz have "le_trans z y" by(force simp:ac_simps)
  also from xy have "le_trans y x" by(force simp:ac_simps)
  finally show "le_trans z x" .
qed

subsection {* Lifted infima and suprema for transformers *}

definition
  lift_Inf :: "(('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) set \<Rightarrow> ('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)"
where
  "lift_Inf S = rep_trans (Inf (abs_trans ` S))"

definition
  lift_Sup :: "(('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) set \<Rightarrow> ('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)"
where
  "lift_Sup S = rep_trans (Sup (abs_trans ` S))"

section {* Least and Greatest Fixed Points for Transformers *}

text {* Now that we have a complete lattice, we pull back the existing
        fixed point definitions, re-establishing the standard results
        (unfolding, induction) under healthiness assumptions. *}

subsubsection {* Monotonicity for Transformer-Transformers *}

definition
  mono_ttrans :: "((('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow>
                  (('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real))) \<Rightarrow> bool"
where
  "mono_ttrans T \<longleftrightarrow> (\<forall>t u. (healthy t \<and> healthy u \<and> le_trans t u) \<longrightarrow>
                          le_trans (T t) (T u))"

lemma mono_ttransI:
  "\<lbrakk> \<And>t u. \<lbrakk> healthy t; healthy u;
             le_trans t u \<rbrakk> \<Longrightarrow> le_trans (T t) (T u) \<rbrakk> \<Longrightarrow> mono_ttrans T"
  by(simp add:mono_ttrans_def)

lemma mono_ttransD:
  "\<lbrakk> mono_ttrans T; le_trans t u; healthy t; healthy u \<rbrakk> \<Longrightarrow>
   le_trans (T t) (T u)"
  by(simp add:mono_ttrans_def)

lemma mono_ttrans_equiv_trans:
  "\<lbrakk> mono_ttrans T; equiv_trans t u; healthy t; healthy u \<rbrakk> \<Longrightarrow>
   equiv_trans (T t) (T u)"
  by(blast dest:mono_ttransD[OF _ equiv_imp_le]
                mono_ttransD[OF _ equiv_imp_le_alt]
                intro!:le_trans_antisym)

subsubsection {* Lifted Definitions *}
 
definition
  lfp_trans :: "((('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow>
                 (('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real))) \<Rightarrow>
                (('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real))"
where
  "lfp_trans T = rep_trans (lfp (abs_trans o T o rep_trans))"

lemma lfp_trans_healthy[intro!,simp]:
  "healthy (lfp_trans f)"
  by(simp add:lfp_trans_def)

text {* These are the equivalents of the standard lemmas on fixed points
        (from HOL/Inductive.thy) *}

(* XXX move *)
lemma abs_trans_leI:
  "\<lbrakk> le_trans t u; healthy t; healthy u \<rbrakk> \<Longrightarrow> abs_trans t \<le> abs_trans u"
  unfolding less_eq_trans.rep_eq
  by(simp add:le_trans_equiv_rsp[OF rep_abs_trans]
              le_trans_equiv_rsp_right[OF rep_abs_trans])

lemma rep_trans_leI:
  "t \<le> u \<Longrightarrow> le_trans (rep_trans t) (rep_trans u)"
  unfolding less_eq_trans.rep_eq by(assumption)

lemma lfp_trans_lowerbound:
  assumes healthy:    "healthy t"
  assumes healthy_pr: "\<And>t. healthy t \<Longrightarrow> healthy (T t)"
  assumes equiv_pr:   "\<And>t u. \<lbrakk> healthy t; healthy u; equiv_trans t u \<rbrakk> \<Longrightarrow>
                             equiv_trans (T t) (T u)"
  assumes fp:         "le_trans (T t) t"
  shows   "le_trans (lfp_trans T) t"
  unfolding lfp_trans_def
  proof -
    from equiv_pr and healthy
    have "equiv_trans (T (rep_trans (abs_trans t))) (T t)" by(simp)

    with healthy and healthy_pr and fp have
      "(abs_trans o T o rep_trans) (abs_trans t) \<le> abs_trans t"
      unfolding o_def by(simp add:abs_trans_leI)

    hence "lfp (abs_trans o T o rep_trans) \<le> abs_trans t"
      by(rule lfp_lowerbound)

    hence "le_trans (rep_trans (lfp (abs_trans o T o rep_trans)))
                    (rep_trans (abs_trans t))"
      by(rule rep_trans_leI)

    with healthy
    show "le_trans (rep_trans (lfp (abs_trans o T o rep_trans))) t"
      by(simp add:le_trans_equiv_rsp_right[OF rep_abs_trans])
qed

lemma lfp_trans_greatest:
  assumes healthy:    "healthy t"
  assumes healthy_pr: "\<And>t. healthy t \<Longrightarrow> healthy (T t)"
  assumes lowerbound: "\<And>u. \<lbrakk> healthy u; le_trans (T u) u \<rbrakk> \<Longrightarrow> le_trans t u"
  shows "le_trans t (lfp_trans T)"
  unfolding lfp_trans_def
proof -
  have "\<And>u. (abs_trans o T o rep_trans) u \<le> u \<Longrightarrow> abs_trans t \<le> u"
    unfolding o_def
  proof -
    fix u assume "(abs_trans (T (rep_trans u))) \<le> u"
    hence "le_trans ((rep_trans (abs_trans (T (rep_trans u))))) (rep_trans u)"
      by(rule rep_trans_leI)
    with healthy_pr have "le_trans (T (rep_trans u)) (rep_trans u)"
      by(simp add:le_trans_equiv_rsp[OF rep_abs_trans])
    with lowerbound have "le_trans t (rep_trans u)" by(simp)
    with healthy have "abs_trans t \<le> abs_trans (rep_trans u)"
      by(blast intro:abs_trans_leI)
    thus "abs_trans t \<le> u"
      by(simp)
  qed

  hence "abs_trans t \<le> lfp (abs_trans o T o rep_trans)"
    by(rule lfp_greatest)

  hence "le_trans (rep_trans (abs_trans t))
                  (rep_trans (lfp (abs_trans o T o rep_trans)))"
    by(rule rep_trans_leI)

  with healthy
  show "le_trans t (rep_trans (lfp (abs_trans o T o rep_trans)))"
    by(simp add:le_trans_equiv_rsp[OF rep_abs_trans])
qed

lemma lfp_trans_lemma2:
  "\<lbrakk> mono_ttrans T; \<And>t. healthy t \<Longrightarrow> healthy (T t) \<rbrakk> \<Longrightarrow>
   le_trans (T (lfp_trans T)) (lfp_trans T)"
  apply(rule lfp_trans_greatest, simp_all)[]
  apply(rule_tac y="T u" in le_trans_trans, simp_all)[]
  apply(rule_tac T=T in mono_ttransD, simp_all)[]
  apply(rule lfp_trans_lowerbound, simp_all)[]
  apply(rule_tac T=T in mono_ttrans_equiv_trans, simp_all)[]
done

lemma lfp_trans_lemma3:
  "\<lbrakk> mono_ttrans T; \<And>t. healthy t \<Longrightarrow> healthy (T t) \<rbrakk> \<Longrightarrow>
   le_trans (lfp_trans T) (T (lfp_trans T))"
  apply(rule lfp_trans_lowerbound, simp_all)[]
   apply(rule_tac T=T in mono_ttrans_equiv_trans, simp_all)[]
  apply(rule_tac T=T in mono_ttransD, simp_all)[]
  apply(rule lfp_trans_lemma2, simp_all)
done

lemma lfp_trans_unfold:
  "\<lbrakk> mono_ttrans T; \<And>t. healthy t \<Longrightarrow> healthy (T t) \<rbrakk> \<Longrightarrow>
   equiv_trans (lfp_trans T) (T (lfp_trans T))"
  by (iprover intro: le_trans_antisym lfp_trans_lemma2 lfp_trans_lemma3)

text {* Properties of the lifted Supremum operator *}

(* XXX - tidy *)

lemma lift_Sup_least:
  "\<lbrakk> healthy z; \<And>x. x \<in> A \<Longrightarrow> healthy x; \<And>x. x \<in> A \<Longrightarrow> le_trans x z \<rbrakk> \<Longrightarrow>
   le_trans (lift_Sup A) z"
  unfolding lift_Sup_def
  apply(rule_tac y="rep_trans (abs_trans z)" in le_trans_trans)
   apply(rule rep_trans_leI)
   apply(rule Sup_least)
   apply(rule imageE, assumption)
   apply(simp add:abs_trans_leI)
  apply(force intro: equiv_imp_le)
done

lemma lift_Sup_upper:
  "\<lbrakk> healthy x; x \<in> A \<rbrakk> \<Longrightarrow> le_trans x (lift_Sup A)"
  unfolding lift_Sup_def
  apply(rule_tac y="rep_trans (abs_trans x)" in le_trans_trans)
   apply(rule equiv_imp_le)
   apply(subst equiv_trans_comm, simp)
  apply(rule rep_trans_leI)
  apply(rule Sup_upper)
  apply(force intro:imageE)
done

lemma lift_Sup_healthy[simp]:
  "healthy (lift_Sup S)"
  unfolding lift_Sup_def by(simp)
 
text {* The (relevant) induction principle *}

lemma lfp_trans_ordinal_induct:
  assumes mono: "mono_ttrans f"
  and healthy_pr: "\<And>t. healthy t \<Longrightarrow> healthy (f t)"
  and P_f: "\<And>S. \<lbrakk> healthy S; P S \<rbrakk> \<Longrightarrow> P (f S)"
  and P_Union: "\<And>M. \<lbrakk> \<forall>S\<in>M. P S; \<forall>S\<in>M. healthy S \<rbrakk> \<Longrightarrow> P (lift_Sup M)"
  and P_equiv: "\<And>t u. \<lbrakk> equiv_trans t u; P t \<rbrakk> \<Longrightarrow> P u"
  shows "P (lfp_trans f)"
proof -
  let ?M = "{S. le_trans S (lfp_trans f) \<and> P S \<and> healthy S}"
  have P_sup: "P (lift_Sup ?M)" using P_Union by simp
  also have "equiv_trans (lift_Sup ?M) (lfp_trans f)"
  proof (rule le_trans_antisym)
    have M_healthy: "healthy (lift_Sup ?M)" by(simp add:lift_Sup_healthy)
    thus "le_trans (lift_Sup ?M) (lfp_trans f)" by (force intro: lift_Sup_least)
    with mono have "le_trans (f (lift_Sup ?M)) (f (lfp_trans f))"
      by(force dest!:mono_ttransD[OF mono])
    hence "le_trans (f (lift_Sup ?M)) (lfp_trans f)"     
      using healthy_pr and mono[THEN lfp_trans_unfold,
                                THEN le_trans_equiv_rsp_right] by simp
    hence "f (lift_Sup ?M) \<in> ?M" using P_f P_Union M_healthy healthy_pr by simp
    hence "le_trans (f (lift_Sup ?M)) (lift_Sup ?M)" using M_healthy healthy_pr
      by (simp add:lift_Sup_upper)
    thus "le_trans (lfp_trans f) (lift_Sup ?M)"
      using M_healthy healthy_pr mono
      by(simp add:lfp_trans_lowerbound mono_ttrans_equiv_trans)
  qed
  from this and P_sup show ?thesis by (rule P_equiv)
qed

definition
  gfp_trans :: "((('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow>
                (('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real))) \<Rightarrow>
               (('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real))"
where
  "gfp_trans T = rep_trans (gfp (abs_trans o T o rep_trans))"

lemma gfp_trans_healthy[simp]:
  "healthy (gfp_trans S)"
  by(simp add:gfp_trans_def)

lemma gfp_trans_upperbound:
  assumes healthy:    "healthy t"
  assumes healthy_pr: "\<And>t. healthy t \<Longrightarrow> healthy (T t)"
  assumes equiv_pr:   "\<And>t u. \<lbrakk> healthy t; healthy u; equiv_trans t u \<rbrakk> \<Longrightarrow>
                             equiv_trans (T t) (T u)"
  assumes fp:         "le_trans t (T t)"
  shows   "le_trans t (gfp_trans T)"
  unfolding gfp_trans_def
  proof -
    from equiv_pr and healthy
    have "equiv_trans (T (rep_trans (abs_trans t))) (T t)"
      by(simp)

    from this and healthy and healthy_pr and fp have
      "abs_trans t \<le> (abs_trans o T o rep_trans) (abs_trans t)"
      unfolding o_def by(simp add:abs_trans_leI)

    hence "abs_trans t \<le> gfp (abs_trans o T o rep_trans)"
      by(rule gfp_upperbound)

    hence "le_trans (rep_trans (abs_trans t))
                    (rep_trans (gfp (abs_trans o T o rep_trans)))"
      by(rule rep_trans_leI)

    with healthy
    show "le_trans t (rep_trans (gfp (abs_trans o T o rep_trans)))"
      by(simp add:le_trans_equiv_rsp[OF rep_abs_trans])
qed

lemma gfp_trans_least:
  assumes healthy:     "healthy t"
  assumes healthy_pr: "\<And>t. healthy t \<Longrightarrow> healthy (T t)"
  assumes lowerbound: "\<And>u. \<lbrakk> healthy u; le_trans u (T u) \<rbrakk> \<Longrightarrow> le_trans u t"
  shows "le_trans (gfp_trans T) t"
  unfolding gfp_trans_def
  proof -
    have "\<And>u. u \<le> (abs_trans o T o rep_trans) u \<Longrightarrow> u \<le> abs_trans t"
      unfolding o_def proof -
        fix u assume "u \<le> abs_trans( T (rep_trans u))"
        hence "le_trans (rep_trans u)
                        (rep_trans (abs_trans (T (rep_trans u))))"
          by(rule rep_trans_leI)
        with healthy_pr have "le_trans (rep_trans u) (T (rep_trans u))"
          by(simp add:le_trans_equiv_rsp_right[OF rep_abs_trans])
        with lowerbound have "le_trans (rep_trans u) t" by(simp)
        with healthy and healthy_rep_trans
        have "abs_trans (rep_trans u) \<le> abs_trans t"
          by(blast intro: abs_trans_leI)
        thus "u \<le> abs_trans t" by(simp)
      qed

    hence "gfp (abs_trans o T o rep_trans) \<le> (abs_trans t)"
      by(rule gfp_least)

    with healthy
    show "le_trans (rep_trans (gfp (abs_trans o T o rep_trans))) t"
      by(force dest:rep_trans_leI
               simp:le_trans_equiv_rsp_right[OF rep_abs_trans])
  qed

lemma gfp_trans_lemma2:
  "\<lbrakk> mono_ttrans T; \<And>t. healthy t \<Longrightarrow> healthy (T t) \<rbrakk> \<Longrightarrow>
  le_trans (gfp_trans T) (T (gfp_trans T))"
  apply(rule gfp_trans_least, simp_all)
  apply(rule_tac y="T u" in le_trans_trans, assumption)
  apply(rule_tac T=T in mono_ttransD, simp_all)
  apply(rule gfp_trans_upperbound, simp_all)
  apply(rule_tac T=T in mono_ttrans_equiv_trans, simp_all)
  done

lemma gfp_trans_lemma3:
  "\<lbrakk> mono_ttrans T; \<And>t. healthy t \<Longrightarrow> healthy (T t) \<rbrakk> \<Longrightarrow>
   le_trans (T (gfp_trans T)) (gfp_trans T)"
  apply(rule gfp_trans_upperbound, simp_all)[]
   apply(rule_tac T=T in mono_ttrans_equiv_trans, simp_all)[]
  apply(rule_tac T=T in mono_ttransD, simp_all)[]
  apply(iprover intro:gfp_trans_lemma2)
  done

lemma gfp_trans_unfold:
  "\<lbrakk> mono_ttrans T; \<And>t. healthy t \<Longrightarrow>  healthy (T t) \<rbrakk> \<Longrightarrow>
   equiv_trans (gfp_trans T) (T (gfp_trans T))"
  by (iprover intro: le_trans_antisym gfp_trans_lemma2 gfp_trans_lemma3)

text {* Properties of the lifted Infimum operator *}

lemma lift_Inf_greatest:
  "\<lbrakk> healthy z; \<And>x. x \<in> A \<Longrightarrow> healthy x; \<And>x. x \<in> A \<Longrightarrow> le_trans z x \<rbrakk> \<Longrightarrow>
   le_trans z (lift_Inf A)"
  unfolding lift_Inf_def
  apply(subst le_trans_equiv_rsp[OF rep_abs_trans, symmetric], assumption)
  apply(rule rep_trans_leI)
  apply(rule Inf_greatest)
  apply(rule imageE, assumption)
  apply(simp add:abs_trans_leI)
done

lemma lift_Inf_lower:
  "\<lbrakk> healthy x; x \<in> A \<rbrakk> \<Longrightarrow> le_trans (lift_Inf A) x"
  unfolding lift_Inf_def
  apply(subst le_trans_equiv_rsp_right[OF rep_abs_trans, symmetric], assumption)
  apply(rule rep_trans_leI)
  apply(rule Inf_lower)
  apply(force intro:imageE)
  done

lemma lift_Inf_healthy[simp]:
  "healthy (lift_Inf S)"
  unfolding lift_Inf_def by(simp)

lemma gfp_trans_ordinal_induct:
  assumes mono: "mono_ttrans f"
  and healthy_pr: "\<And>t. healthy t \<Longrightarrow> healthy (f t)"
  and P_f: "\<And>S. \<lbrakk> P S; healthy S \<rbrakk> \<Longrightarrow> P (f S)"
  and P_Union: "\<And>M. \<lbrakk> \<forall>S\<in>M. P S; \<forall>S\<in>M. healthy S \<rbrakk> \<Longrightarrow> P (lift_Inf M)"
  and P_equiv: "\<And>t u. \<lbrakk> equiv_trans t u; P t \<rbrakk> \<Longrightarrow> P u"
  shows "P (gfp_trans f)"
proof -
  let ?M = "{S. le_trans (gfp_trans f) S \<and> P S \<and> healthy S}"
  have P_inf: "P (lift_Inf ?M)" using P_Union by simp
  also have "equiv_trans (lift_Inf ?M) (gfp_trans f)"
  proof (rule le_trans_antisym)
    have M_healthy: "healthy (lift_Inf ?M)" by(simp)
    thus "le_trans (gfp_trans f) (lift_Inf ?M)"
      by (force intro: lift_Inf_greatest)
    with mono have "le_trans (f (gfp_trans f)) (f (lift_Inf ?M))"
      by(force dest!:mono_ttransD[OF mono])
    hence "le_trans (gfp_trans f) (f (lift_Inf ?M))"
      using healthy_pr and mono [THEN gfp_trans_unfold,
                                 THEN le_trans_equiv_rsp] by simp
    hence "f (lift_Inf ?M) \<in> ?M" using P_f P_Union M_healthy healthy_pr by simp
    hence "le_trans (lift_Inf ?M) (f (lift_Inf ?M))" using M_healthy healthy_pr
      by (simp add:lift_Inf_lower)
    thus "le_trans (lift_Inf ?M) (gfp_trans f)"
      using M_healthy healthy_pr mono
      by(simp add:gfp_trans_upperbound mono_ttrans_equiv_trans)
  qed
  from this and P_inf show ?thesis by (rule P_equiv)
qed

end
