(*
 * Copyright (C) 2013 National ICT Australia Ltd ('NICTA')
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Author: David Cock - David.Cock@nicta.com.au *)

header "Healthy Expectation Transformers for pGCL"

theory Transformers imports Expectations begin

section {* Ordering and equality for transformers *}

text {* Transformers are compared only on bounded, non-negative (sound)
        expectations. We define a pre-order on transformers pointwise by
        their action on these expectations.  Equivalence is induced,
        giving a partial order. *}

definition
  le_trans :: "(('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow>
               (('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow> bool"
where
  "le_trans t u \<equiv> \<forall>P. sound P \<longrightarrow> t P \<le> u P"

lemma le_transI[intro]:
  "\<lbrakk> \<And>P. sound P \<Longrightarrow> t P \<le> u P \<rbrakk> \<Longrightarrow> le_trans t u"
  by(simp add:le_trans_def)

lemma  le_transD[dest]:
  "\<lbrakk> le_trans t u; sound P \<rbrakk> \<Longrightarrow> t P \<le> u P"
  by(simp add:le_trans_def)

lemma le_trans_trans[trans]:
  "\<lbrakk> le_trans x y; le_trans y z \<rbrakk> \<Longrightarrow> le_trans x z"
  unfolding le_trans_def by(blast intro:order_trans)

lemma le_trans_refl[intro!,simp]:
  "le_trans x x"
  by(simp add:le_trans_def)

definition
  l_trans :: "(('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow>
              (('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow> bool"
where
  "l_trans t u \<longleftrightarrow> le_trans t u \<and> \<not> le_trans u t"

text {* Transformer equivalence is induced by comparison *}

definition
  equiv_trans :: "(('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow>
                  (('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow> bool"
where
  "equiv_trans t u \<longleftrightarrow> le_trans t u \<and> le_trans u t"

lemma equiv_transI[intro]:
  "\<lbrakk> \<And>P. sound P \<Longrightarrow> t P = u P \<rbrakk> \<Longrightarrow> equiv_trans t u"
  unfolding equiv_trans_def by(force)

lemma equiv_transD[dest]:
  "\<lbrakk> equiv_trans t u; sound P \<rbrakk> \<Longrightarrow> t P = u P"
  unfolding equiv_trans_def by(blast intro:antisym)

lemma equiv_trans_refl[intro!,simp]:
  "equiv_trans t t"
  by(blast)

lemma le_trans_antisym:
  "\<lbrakk> le_trans x y; le_trans y x \<rbrakk> \<Longrightarrow> equiv_trans x y"
  unfolding equiv_trans_def by(simp)

lemma equiv_trans_comm[ac_simps]:
  "equiv_trans t u \<longleftrightarrow> equiv_trans u t"
  unfolding equiv_trans_def by(blast)

lemma equiv_imp_le[intro]:
  "equiv_trans t u \<Longrightarrow> le_trans t u"
  unfolding equiv_trans_def by(clarify)

lemma equiv_imp_le_alt:
  "equiv_trans t u \<Longrightarrow> le_trans u t"
  by(force simp:ac_simps)

lemma le_trans_equiv_rsp[simp]:
  "equiv_trans t u \<Longrightarrow> le_trans t v \<longleftrightarrow> le_trans u v"
  unfolding equiv_trans_def by(blast intro:le_trans_trans)

lemma le_trans_equiv_rsp_right[simp]:
  "equiv_trans t u \<Longrightarrow> le_trans v t \<longleftrightarrow> le_trans v u"
  unfolding equiv_trans_def by(blast intro:le_trans_trans)

section {* Healthiness conditions for expectation transformers. *}

text {* Feasible transformers preserve bounds, and non-negativity. *}

definition feasible :: "(('a \<Rightarrow> ('b::{zero,order})) \<Rightarrow> ('a \<Rightarrow> 'b)) \<Rightarrow> bool"
where     "feasible t \<longleftrightarrow> (\<forall>P b. bounded_by b P \<and> nneg P \<longrightarrow>
                               bounded_by b (t P) \<and> nneg (t P))"

lemma feasibleI[intro]:
  "\<lbrakk> \<And>b P. \<lbrakk> bounded_by b P; nneg P \<rbrakk> \<Longrightarrow> bounded_by b (t P);
     \<And>b P. \<lbrakk> bounded_by b P; nneg P \<rbrakk> \<Longrightarrow> nneg (t P) \<rbrakk> \<Longrightarrow> feasible t"
  by(force simp:feasible_def)

lemma feasible_boundedD[dest]:
  "\<lbrakk> feasible t; bounded_by b P; nneg P \<rbrakk> \<Longrightarrow> bounded_by b (t P)"
  by(simp add:feasible_def)

lemma feasible_nnegD[dest]:
  "\<lbrakk> feasible t; bounded_by b P; nneg P \<rbrakk> \<Longrightarrow> nneg (t P)"
  by(simp add:feasible_def)

lemma feasible_sound[dest]:
  "\<lbrakk> feasible t; sound P \<rbrakk> \<Longrightarrow> sound (t P)"
  apply(rule soundI)
  unfolding sound_def
  apply(blast)+
  done

(* XXX - tidy *)

lemma feasible_pr_0[simp]:
  "feasible t \<Longrightarrow> t (\<lambda>x. 0) = (\<lambda>x. 0)"
  apply(rule ext, rule antisym)
   apply(subgoal_tac "bounded_by (0::'b) (\<lambda>x. 0)", blast+)[]
  apply(rule_tac P="t (\<lambda>x. 0)" in nnegD)
  apply(rule_tac t=t  and b=0 in feasible_nnegD, blast+)
done

lemma feasible_id:
  "feasible (\<lambda>x. x)"
  unfolding feasible_def by(blast)

text {* Monotonicity is restricted to sound predicates *}

definition
  mono_trans :: "(('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow> bool"
where
  "mono_trans t \<equiv> \<forall>P Q. (sound P \<and> sound Q \<and> P \<le> Q) \<longrightarrow> t P \<le> t Q"

lemma mono_transI[intro]:
  "\<lbrakk> \<And>P Q. \<lbrakk> sound P; sound Q; P \<le> Q \<rbrakk> \<Longrightarrow>  t P \<le> t Q \<rbrakk> \<Longrightarrow> mono_trans t"
  by(simp add:mono_trans_def)

lemma mono_transD[dest]:
  "\<lbrakk> mono_trans t; sound P; sound Q; P \<le> Q \<rbrakk> \<Longrightarrow> t P \<le> t Q"
  by(simp add:mono_trans_def)

text {* Respect scaling by c > 0.  Holds with c=0 for wp, but not wlp. *}

definition
  scaling :: "(('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow> bool"
where
  "scaling t \<equiv> \<forall>P c x. sound P \<and> 0 < c \<longrightarrow> c * t P x = t (\<lambda>x. c * P x) x"

lemma scalingI[intro]:
  "\<lbrakk> \<And>P c x. \<lbrakk> sound P; 0 < c \<rbrakk> \<Longrightarrow> c * t P x = t (\<lambda>x. c * P x) x \<rbrakk> \<Longrightarrow> scaling t"
  by(simp add:scaling_def)

lemma scalingD[dest]:
  "\<lbrakk> scaling t; sound P; 0 < c \<rbrakk>  \<Longrightarrow> c * t P x = t (\<lambda>x. c * P x) x"
  by(simp add:scaling_def)

text {* Healthy transformers are feasible and monotonic, and respect scaling *}

definition
  healthy :: "(('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow> bool"
where
  "healthy t \<longleftrightarrow> feasible t \<and> mono_trans t \<and> scaling t"

lemma healthyI[intro]:
  "\<lbrakk> feasible t; mono_trans t; scaling t \<rbrakk> \<Longrightarrow> healthy t"
  by(simp add:healthy_def)

lemma healthy_monoD[dest]:
  "healthy t \<Longrightarrow> mono_trans t"
  by(simp add:healthy_def)

lemmas healthy_monoD2 = mono_transD[OF healthy_monoD]

lemma healthy_feasibleD[dest]:
  "healthy t \<Longrightarrow> feasible t"
  by(simp add:healthy_def)

lemma healthy_scalingD[dest]:
  "healthy t \<Longrightarrow> scaling t"
  by(simp add:healthy_def)

lemma healthy_bounded_byD[intro]:
  "\<lbrakk> healthy t; bounded_by b P; nneg P \<rbrakk> \<Longrightarrow> bounded_by b (t P)"
   by(blast)

(* XXX - should be the default rule *)
lemma healthy_bounded_byD2:
  "\<lbrakk> healthy t; bounded_by b P; sound P \<rbrakk> \<Longrightarrow> bounded_by b (t P)"
  by(blast)

lemma healthy_boundedD[dest,simp]:
  "\<lbrakk> healthy t; sound P \<rbrakk> \<Longrightarrow> bounded (t P)"
  by(blast)

lemma healthy_nnegD[dest,simp]:
  "\<lbrakk> healthy t; sound P \<rbrakk> \<Longrightarrow> nneg (t P)"
  by(blast intro!:feasible_nnegD)

lemma healthy_nnegD2[dest,simp]:
  "\<lbrakk> healthy t; bounded_by b P; nneg P \<rbrakk> \<Longrightarrow> nneg (t P)"
  by(blast)

lemma healthy_sound[intro]:
  "\<lbrakk> healthy t; sound P \<rbrakk> \<Longrightarrow> sound (t P)"
  by(rule soundI, blast, blast intro:feasible_nnegD)

lemma healthy_id[simp,intro!]:
  "healthy id"
  by(simp add:healthyI feasibleI mono_transI scalingI)

text {* Some additional results on @term{le_trans} *}

lemma le_trans_bot[intro,simp]:
  "healthy t \<Longrightarrow> le_trans (\<lambda>P s. 0) t"
  by(blast intro:le_funI)

lemma le_trans_top[intro,simp]:
  "healthy t \<Longrightarrow> le_trans t (\<lambda>P s. bound_of P)"
  by(blast intro!:le_transI[OF le_funI])

lemma soundI2[intro]:
  "\<lbrakk> bounded_by b P; nneg P \<rbrakk> \<Longrightarrow> sound P"
  by(blast)

(* XXX - tidy *)
lemma healthy_equivI:
  assumes equiv:   "equiv_trans t u"
      and healthy: "healthy t"
  shows "healthy u"
proof
  have le_t_u: "le_trans t u" by(blast intro:equiv)
  have le_u_t: "le_trans u t" by(simp add:equiv_imp_le ac_simps equiv)
  from equiv have eq_u_t: "equiv_trans u t" by(simp add:ac_simps)

  show "feasible u"
    apply -
    apply(rule feasibleI)
     apply(rule bounded_byI)
     apply(rule_tac y="t P x" in order_trans)
      apply(blast intro:le_funD[OF le_transD, OF le_u_t])
     apply(blast intro:healthy bounded_byD)
    apply(rule nnegI)
    apply(rule_tac y="t P x" in order_trans)
     apply(rule_tac x=x in nnegD, rule healthy_nnegD[OF healthy], blast)
    apply(blast intro:le_funD[OF le_transD, OF le_t_u])
    done

  show "mono_trans u"
    apply(rule mono_transI)
    apply(rule_tac y="t P" in order_trans)
     apply(blast intro:le_u_t le_transD)
    apply(rule_tac y="t Q" in order_trans)
     apply(blast intro:mono_transD healthy_monoD[OF healthy])
    apply(blast intro:le_transD le_t_u)
    done

  show "scaling u"
  proof
    fix P::"'a \<Rightarrow> real" and c::real and x::'a
    assume sound: "sound P"
       and pos:   "0 < c"

    hence "bounded_by (c * bound_of P) (\<lambda>x. c * P x)"
      by(blast intro!:mult_left_mono dest!:less_imp_le)
    hence sc_bounded: "bounded (\<lambda>x. c * P x)"
      by(blast)
    moreover from sound and pos have sc_nneg: "nneg (\<lambda>x. c * P x)"
      by(blast intro:mult_nonneg_nonneg less_imp_le)
    ultimately have sc_sound: "sound (\<lambda>x. c * P x)" ..
        
    show "c * u P x = u (\<lambda>x. c * P x) x"
    proof -
      have nzero: "0 \<noteq> c" by(rule less_imp_neq[OF pos])
      hence "c * u P x = c * t P x"
        apply(simp add:pos)
        using eq_u_t and sound by(blast intro:fun_cong)

      also have "... = t (\<lambda>x. c * P x) x"
        using healthy and sound and pos
        by(blast intro: scalingD)

      also from sc_sound and equiv have "... = u (\<lambda>x. c * P x) x"
        by(blast intro:fun_cong)

      finally show ?thesis .
    qed
  qed
qed

lemma healthy_equiv:
  "equiv_trans t u \<Longrightarrow> healthy t \<longleftrightarrow> healthy u"
  apply(rule iffI, rule healthy_equivI, assumption+)
  apply(simp add:healthy_equivI ac_simps)
  done

section {* Sublinearity *}

text {* We generally only insist on the above healthiness predicate,
  which is easily established, and show the various sublinearity
  properties on demand. *}

definition sublinear ::
  "(('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow> bool"
where
  "sublinear t \<longleftrightarrow> (\<forall>a b c P Q s. (sound P \<and> sound Q) \<longrightarrow>
                  a * t P s + b * t Q s \<ominus> c
                  \<le> t (\<lambda>s'. a * P s' + b * Q s' \<ominus> c) s)"

lemma sublinearI[intro]:
  "\<lbrakk> \<And>a b c P Q s. \<lbrakk> sound P; sound Q \<rbrakk> \<Longrightarrow>
     a * t P s + b * t Q s \<ominus> c \<le>
     t (\<lambda>s'. a * P s' + b * Q s' \<ominus> c) s \<rbrakk> \<Longrightarrow> sublinear t"
  by(simp add:sublinear_def)

lemma sublinearD[dest]:
  "\<lbrakk> sublinear t; sound P; sound Q \<rbrakk> \<Longrightarrow>
   a * t P s + b * t Q s \<ominus> c \<le>
   t (\<lambda>s'. a * P s' + b * Q s' \<ominus> c) s"
  by(simp add:sublinear_def)

definition bounded_cts ::
  "(('a \<Rightarrow> 'b::complete_lattice) \<Rightarrow> ('a \<Rightarrow> 'b)) \<Rightarrow> bool"
where
  "bounded_cts t \<longleftrightarrow> (\<forall>A. t (Sup A) = Sup (t ` A))"

lemma bounded_ctsI[intro]:
  "\<lbrakk> \<And>A. t (Sup A) = Sup (t ` A) \<rbrakk> \<Longrightarrow> bounded_cts t"
  by(simp add:bounded_cts_def)

lemma bounded_ctsD[dest]:
  "bounded_cts t \<Longrightarrow> t (Sup A) = Sup (t ` A)"
  by(simp add:bounded_cts_def)

definition sub_add ::
  "(('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow> bool"
where
  "sub_add t \<longleftrightarrow> (\<forall>P Q s. (sound P \<and> sound Q) \<longrightarrow>
                t P s + t Q s \<le> t (\<lambda>s'. P s' + Q s') s)"

lemma sub_addI[intro]:
  "\<lbrakk> \<And>P Q s. \<lbrakk> sound P; sound Q \<rbrakk> \<Longrightarrow>
             t P s + t Q s \<le> t (\<lambda>s'. P s' + Q s') s \<rbrakk> \<Longrightarrow> sub_add t"
  by(simp add:sub_add_def)

lemma sub_addD[dest]:
  "\<lbrakk> sub_add t; sound P; sound Q \<rbrakk> \<Longrightarrow> t P s + t Q s \<le> t (\<lambda>s'. P s' + Q s') s"
  by(simp add:sub_add_def)

definition sub_distrib ::
  "(('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)) \<Rightarrow> bool"
where
  "sub_distrib t \<longleftrightarrow> (\<forall>P s. sound P \<longrightarrow> t P s \<ominus> 1 \<le> t (\<lambda>s'. P s' \<ominus> 1) s)"

lemma sub_distribI[intro]:
  "\<lbrakk> \<And>P s. sound P \<Longrightarrow> t P s \<ominus> 1 \<le> t (\<lambda>s'. P s' \<ominus> 1) s \<rbrakk> \<Longrightarrow> sub_distrib t"
  by(simp add:sub_distrib_def)
  
lemma sub_distribD[dest]:
  "\<lbrakk> sub_distrib t; sound P \<rbrakk> \<Longrightarrow> t P s \<ominus> 1 \<le> t (\<lambda>s'. P s' \<ominus> 1) s"
  by(simp add:sub_distrib_def)

definition
  sub_conj :: "(('s \<Rightarrow> real) \<Rightarrow> 's \<Rightarrow> real) \<Rightarrow> bool"
where
  "sub_conj t \<equiv> \<forall>P Q. (sound P \<and> sound Q \<and>
                       bounded_by 1 P \<and> bounded_by 1 Q) \<longrightarrow>
                       t P && t Q \<tturnstile> t (P && Q)"

lemma sub_conjI[intro]:
  "\<lbrakk> \<And>P Q. \<lbrakk> sound P; sound Q; bounded_by 1 P; bounded_by 1 Q \<rbrakk> \<Longrightarrow>
           t P && t Q \<tturnstile> t (P && Q) \<rbrakk> \<Longrightarrow> sub_conj t"
  unfolding sub_conj_def by(simp)

lemma sub_conjD[dest]:
  "\<lbrakk> sub_conj t; sound P; sound Q; bounded_by 1 P; bounded_by 1 Q \<rbrakk> \<Longrightarrow>
   t P && t Q \<tturnstile> t (P && Q)"
  unfolding sub_conj_def by(simp)

lemma sub_conj_wp_twice:
  "\<lbrakk> \<forall>s. sub_conj (f s) \<rbrakk> \<Longrightarrow> sub_conj (\<lambda>P s. f s P s)"
  apply(rule sub_conjI)
  apply(rule le_funI)
  apply(erule_tac x=x in allE)
  apply(drule_tac P=P and Q=Q in sub_conjD, assumption+)
  apply(drule_tac x=x in le_funD)
  apply(simp add:exp_conj_def)
  done

text {* Determinism *}

text {* Full additivity is not generally satisfied.  It holds for (sub-)probabilistic
  transformers however, such as \<oplus>. *}
definition
  additive :: "(('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real) \<Rightarrow> bool"
where
  "additive t \<equiv> \<forall>P Q. (sound P \<and> sound Q) \<longrightarrow>
                      t (\<lambda>s. P s + Q s) = (\<lambda>s. t P s + t Q s)"

lemma additiveD:
  "\<lbrakk> additive t; sound P; sound Q \<rbrakk> \<Longrightarrow> t (\<lambda>s. P s + Q s) = (\<lambda>s. t P s + t Q s)"
  by(simp add:additive_def)

lemma additiveI[intro]:
  "\<lbrakk> \<And>P Q s. \<lbrakk> sound P; sound Q \<rbrakk> \<Longrightarrow> t (\<lambda>s. P s + Q s) s = t P s + t Q s \<rbrakk> \<Longrightarrow>
   additive t"
  unfolding additive_def by(blast)

text {* It is strictly stronger than sub-additivity. *}
lemma additive_sub_add:
  "additive t \<Longrightarrow> sub_add t"
  by(simp add:sub_addI additiveD)

definition
  maximal :: "(('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real) \<Rightarrow> bool"
where
  "maximal t \<equiv> \<forall>c. t (\<lambda>_. c) = (\<lambda>_. c)"

lemma maximalI[intro]:
  "\<lbrakk> \<And>c. t (\<lambda>_. c) = (\<lambda>_. c) \<rbrakk> \<Longrightarrow> maximal t"
  by(simp add:maximal_def)

lemma maximalD[dest]:
  "maximal t \<Longrightarrow> t (\<lambda>_. c) = (\<lambda>_. c)"
  by(simp add:maximal_def)

text {* A transformer that is both additive and maximal is deterministic, and
  thus maximal in the refinement order. *}
definition determ :: "(('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real) \<Rightarrow> bool"
where
  "determ t \<equiv> additive t \<and> maximal t"

lemma determI[intro]:
  "\<lbrakk> additive t; maximal t \<rbrakk> \<Longrightarrow> determ t"
  by(simp add:determ_def)

lemma determ_additiveD[intro]:
  "determ t \<Longrightarrow> additive t"
  by(simp add:determ_def)

lemma determ_maximalD[intro]:
  "determ t \<Longrightarrow> maximal t"
  by(simp add:determ_def)

end
