(*
 * Copyright (C) 2013 National ICT Australia Ltd ('NICTA')
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Author: David Cock - David.Cock@nicta.com.au *)

header "Weakest-precondition and Weakest-liberal-precondition semantics
        for pGCL"

theory pGCL imports Misc Induction begin

section {* Core Primitives and Syntax *}

text {* We make a shallow embedding of pGCL in HOL.  A program is
        embedded directly as its wp (or wlp) transformer.  This is
        achieved with an additional parameter, specifying whether
        wp or wlp semantics should be obeyed. *}

type_synonym 's rtrans = "('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)"
type_synonym 's prog = "bool \<Rightarrow> ('s \<Rightarrow> real) \<Rightarrow> ('s \<Rightarrow> real)"

text {* Abort is either top or bottom *}
definition Abort :: "'s prog"
where     "Abort \<equiv> \<lambda>ab P s. if ab then 0 else bound_of P"

definition Skip :: "'s prog"
where     "Skip \<equiv> \<lambda>ab P. P"

definition Apply :: "('s \<Rightarrow> 's) \<Rightarrow> 's prog"
where     "Apply f \<equiv> \<lambda>ab P s. P (f s)"

definition Seq :: "'s prog \<Rightarrow> 's prog \<Rightarrow> 's prog"
                 (infixl ";;" 59)
where     "Seq a b \<equiv> (\<lambda>ab. a ab o b ab)"

definition PC :: "'s prog \<Rightarrow> ('s \<Rightarrow> real) \<Rightarrow> 's prog \<Rightarrow> 's prog"
                 ("_ \<^bsub>_\<^esub>\<oplus> _" [58,57,57] 57)
where     "PC a P b \<equiv> \<lambda>ab Q s. P s * a ab Q s + (1 - P s) * b ab Q s" 

definition DC :: "'s prog \<Rightarrow> 's prog \<Rightarrow> 's prog" ("_ \<Sqinter> _" [58,57] 57)
where     "DC a b \<equiv> \<lambda>ab Q s. min (a ab Q s) (b ab Q s)"

text {* The Embed primitive takes a healthy transformer, not defined
        structurally, and applies it. *}
definition Embed :: "'s rtrans \<Rightarrow> 's prog"
where     "Embed t = (\<lambda>ab. t)"

text {* Recursion is either either least or greatest fixed point. *}
definition Mu :: "('s prog \<Rightarrow> 's prog) \<Rightarrow> 's prog" (binder "\<mu>" 50)
where     "Mu(T) \<equiv> (\<lambda>ab. if ab then lfp_trans (\<lambda>t. T (Embed t) ab)
                               else gfp_trans (\<lambda>t. T (Embed t) ab))"
text {* Finite repetition *}
primrec
  repeat :: "nat \<Rightarrow> 'a prog \<Rightarrow> 'a prog"
where
  "repeat 0 p = Skip" |
  "repeat (Suc n) p = p ;; repeat n p"

text {* Demonic choice on a set *}
definition SetDC :: "('a \<Rightarrow> 's prog) \<Rightarrow> ('s \<Rightarrow> 'a set) \<Rightarrow> 's prog"
  where "SetDC f S \<equiv> \<lambda>ab P s. glb ((\<lambda>a. f a ab P s) ` S s)"

syntax "_SetDC" :: "pttrn => ('s => 'a set) => 's prog => 's prog"
                   ("\<Sqinter>_\<in>_./ _" 100)
translations "\<Sqinter>x\<in>S. p" == "CONST SetDC (%x. p) S"

definition
  SetPC :: "('a \<Rightarrow> 's prog) \<Rightarrow> ('s \<Rightarrow> 'a set) \<Rightarrow> ('s \<Rightarrow> 'a \<Rightarrow> real) \<Rightarrow> 's prog"
where
  "SetPC f S p \<equiv> \<lambda>ab P s. \<Sum>a\<in>(S s).  p s a * f a ab P s"

text {* Bind a name, deterministically *}

definition
  Bind :: "('s \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 's prog) \<Rightarrow> 's prog"
where
  "Bind g f ab \<equiv> \<lambda>P s. let a = g s in f a ab P s"

syntax "_Bind" :: "pttrn => ('s => 'a) => 's prog => 's prog"
       ("_ is _ in _" [55,55,55]55)
translations "x is f in a" => "CONST Bind f (%x. a)"

definition
  K :: "'b \<Rightarrow> 'a \<Rightarrow> 'b"
where
  "K b \<equiv> \<lambda>_. b"

lemma apply_K[simp]:
  "K c x = c" unfolding K_def by(rule refl)

text {* Bind a name, probabilistically or demonically from a set *}

syntax "_PBind" :: "pttrn => ('s => 'a set) \<Rightarrow> real => 's prog => 's prog"
                   ("_ from _ at _ in _" [55,55,55]55)
translations "x from S at p in a" => "CONST SetPC (%x. a) S (CONST K (%x. p))"

syntax "_DBind" :: "pttrn => ('s => 'a set) \<Rightarrow> 's prog => 's prog"
                   ("_ from _ in _" [55,55,55]55)
translations "x from S in a" => "CONST SetDC (%x. a) S"

text {* Syntax for updating a record-type state *}
syntax
  "_assign" :: "ident => 'a => 's prog" ("_ := _" [1000,900]900)

ML {*
  fun assign_tr [Const (name,_), arg] =
      Const ("pGCL.Apply", dummyT) $
      Abs ("s", dummyT,
           Syntax.const (suffix Record.updateN name) $
           Abs (Name.uu_, dummyT, arg $ Bound 1) $ Bound 0)
    | assign_tr ts = raise TERM ("assign_tr", ts)
*}

parse_translation {* [(@{syntax_const "_assign"}, assign_tr)] *}

syntax
  "_SetPC" :: "ident => ('s => 'a set) => ('s => 'a => real) => 's prog"
              ("_ :\<in> _ at _" [66,66,66]66)

ML {*
  fun set_pc_tr [Const (f,_), S, P] =
      Const ("SetPC", dummyT) $
      Abs ("v", dummyT,
           (Const ("pGCL.Apply", dummyT) $
            Abs ("s", dummyT,
                 Syntax.const (suffix Record.updateN f) $
                 Abs (Name.uu_, dummyT, Bound 2) $ Bound 0))) $
      S $ P
    | set_pc_tr ts = raise TERM ("set_pc_tr", ts)

*}

parse_translation {* [(@{syntax_const "_SetPC"}, set_pc_tr)] *}

syntax
  "_set_dc" :: "ident => ('s => 'a set) => 's prog" ("_ :\<in> _" [66,66]66)

ML {*
  fun set_dc_tr [Const (f,_), S] =
      Const ("SetDC", dummyT) $
      Abs ("v", dummyT,
           (Const ("pGCL.Apply", dummyT) $
            Abs ("s", dummyT,
                 Syntax.const (suffix Record.updateN f) $
                 Abs (Name.uu_, dummyT, Bound 2) $ Bound 0))) $
      S
    | set_dc_tr ts = raise TERM ("set_dc_tr", ts)

*}

parse_translation {* [(@{syntax_const "_set_dc"}, set_dc_tr)] *}

text {* These definitions instantiate the embedding as either
        weakest precondition (True) or weakest liberal precondition
        (False). *}

definition
  wp :: "'s prog \<Rightarrow> 's rtrans"
where
  "wp pr \<equiv> pr True"

definition
  wlp :: "'s prog \<Rightarrow> 's rtrans"
where
  "wlp pr \<equiv> pr False"

text {* If-Then-Else as a degenerate probabilistic choice. *}
abbreviation(input)
  if_then_else :: "['s \<Rightarrow> bool, 's prog, 's prog] \<Rightarrow> 's prog"
      ("If _ Then _ Else _" 58)
where
  "If P Then a Else b == a \<^bsub>\<guillemotleft>P\<guillemotright>\<^esub>\<oplus> b"

text {* Syntax for loops *}
abbreviation
  do_while :: "['s \<Rightarrow> bool, 's prog] \<Rightarrow> 's prog"
              ("do _ \<longrightarrow>// (4 _) //od")
where
  "do_while P a \<equiv> \<mu> x. If P Then a ;; x Else Skip"

section {* Proved wp and wlp rules for non-recursive primitives *}

lemma eval_wp_Abort:
  "wp Abort P = (\<lambda>s. 0)"
  unfolding wp_def Abort_def by(simp)

lemma eval_wlp_Abort:
  "wlp Abort P = (\<lambda>s. bound_of P)"
  unfolding wlp_def Abort_def by(simp)

lemma eval_wp_Skip:
  "wp Skip P = P"
  unfolding wp_def Skip_def by(simp)

lemma eval_wlp_Skip:
  "wlp Skip P = P"
  unfolding wlp_def Skip_def by(simp)

lemma eval_wp_Apply:
  "wp (Apply f) P = P o f"
  unfolding wp_def Apply_def by(simp add:o_def)

lemma eval_wlp_Apply:
  "wlp (Apply f) P = P o f"
  unfolding wlp_def Apply_def by(simp add:o_def)

lemma eval_wp_Seq:
  "wp (a ;; b) P = (wp a o wp b) P"
  unfolding wp_def Seq_def by(simp)

lemma eval_wlp_Seq:
  "wlp (a ;; b) P = (wlp a o wlp b) P"
  unfolding wlp_def Seq_def by(simp)

lemma eval_wp_PC:
  "wp (a \<^bsub>Q\<^esub>\<oplus> b) P = (\<lambda>s. Q s * wp a P s + (1 - Q s) * wp b P s)"
  unfolding wp_def PC_def by(simp)

lemma eval_wlp_PC:
  "wlp (a \<^bsub>Q\<^esub>\<oplus> b) P = (\<lambda>s. Q s * wlp a P s + (1 - Q s) * wlp b P s)"
  unfolding wlp_def PC_def by(simp)

lemma eval_wp_DC:
  "wp (a \<Sqinter> b) P = (\<lambda>s. min (wp a P s) (wp b P s))"
  unfolding wp_def DC_def by(simp)

lemma eval_wlp_DC:
  "wlp (a \<Sqinter> b) P = (\<lambda>s. min (wlp a P s) (wlp b P s))"
  unfolding wlp_def DC_def by(simp)

lemma eval_wp_Embed:
  "wp (Embed t) = t"
  unfolding wp_def Embed_def by(simp)

lemma eval_wlp_Embed:
  "wlp (Embed t) = t"
  unfolding wlp_def Embed_def by(simp)

lemma eval_wp_SetDC:
  "wp (SetDC p S) R s = glb ((\<lambda>a. wp (p a) R s) ` S s)"
  unfolding wp_def SetDC_def by(simp)

lemma eval_wlp_SetDC:
  "wlp (SetDC p S) R s = glb ((\<lambda>a. wlp (p a) R s) ` S s)"
  unfolding wlp_def SetDC_def by(simp)

lemma eval_wp_SetPC:
  "wp (SetPC f S p) P = (\<lambda>s. \<Sum>a\<in>(S s). p s a * wp (f a) P s)"
  unfolding wp_def SetPC_def by(simp)

lemma eval_wlp_SetPC:
  "wlp (SetPC f S p) P = (\<lambda>s. \<Sum>a\<in>(S s). p s a * wlp (f a) P s)"
  unfolding wlp_def SetPC_def by(simp)

lemma eval_wp_Mu:
  "wp (\<mu> t. T t) = lfp_trans (\<lambda>t. wp (T (Embed t)))"
  unfolding wp_def Mu_def by(simp)

lemma eval_wlp_Mu:
  "wlp (\<mu> t. T t) = gfp_trans (\<lambda>t. wlp (T (Embed t)))"
  unfolding wlp_def Mu_def by(simp)

lemma eval_wp_Bind:
  "wp (Bind g f) = (\<lambda>P s. wp (f (g s)) P s)"
  unfolding Bind_def wp_def Let_def by(simp)

lemma eval_wlp_Bind:
  "wlp (Bind g f) = (\<lambda>P s. wlp (f (g s)) P s)"
  unfolding Bind_def wlp_def Let_def by(simp)

text {* Use simp add:wp_eval to fully unfold a program fragment *}
lemmas wp_eval = eval_wp_Abort eval_wlp_Abort eval_wp_Skip eval_wlp_Skip
                 eval_wp_Apply eval_wlp_Apply eval_wp_Seq eval_wlp_Seq
                 eval_wp_PC eval_wlp_PC eval_wp_DC eval_wlp_DC
                 eval_wp_Embed eval_wlp_Embed eval_wp_SetDC eval_wlp_SetDC
                 eval_wp_SetPC eval_wlp_SetPC eval_wp_Mu eval_wlp_Mu
                 eval_wp_Bind eval_wlp_Bind

lemma Skip_Seq:
  "Skip ;; A = A"
  unfolding Skip_def Seq_def o_def by(rule refl)

lemma Seq_Skip:
  "A ;; Skip = A"
  unfolding Skip_def Seq_def o_def by(rule refl)

text {* Use these as simp rules to clear out Skips *}
lemmas skip_simps = Skip_Seq Seq_Skip

section {* Healthiness of wp and wlp *}

text {* Healthiness is derived by structural induction using
        the simplifier.  Abort and Skip form base cases. *}

lemma healthy_wp_Abort:
  "healthy (wp Abort)"
  apply(rule healthyI)
    apply(rule feasibleI, simp_all add:wp_eval)[]
     apply(rule_tac a=0 in bounded_by_mono)
      apply(force intro:bounded_byI)
     apply(force intro:bounded_by_nneg)
    apply(rule mono_transI, simp add:wp_eval)
   apply(rule scalingI, simp add:wp_eval)
   done

lemma healthy_wlp_Abort:
  "healthy (wlp Abort)"
  apply(rule healthyI)
    apply(rule feasibleI, simp_all add:wp_eval)[]
     apply(force intro:bounded_byI)
    apply(force intro:bounded_by_nneg)
   apply(rule mono_transI, simp add:wp_eval)
   apply(blast)
  apply(rule scalingI, simp add:wp_eval)
  done

lemma healthy_wp_Skip:
  "healthy (wp Skip)"
  by(force intro!:healthyI feasibleI mono_transI scalingI simp:wp_eval)

lemma healthy_wlp_Skip:
  "healthy (wlp Skip)"
  by(force intro!:healthyI feasibleI mono_transI scalingI simp:wp_eval)

lemma healthy_Seq:
  "\<lbrakk> healthy (a ab); healthy (b ab) \<rbrakk> \<Longrightarrow> healthy ((a ;; b) ab)"
  apply(safe intro!:healthyI feasibleI mono_transI scalingI,
        simp_all add:Seq_def)
     apply(blast)
    apply(rule_tac t="a ab" in feasible_nnegD, blast)
     apply(rule_tac t="b ab" in feasible_boundedD, blast+)[]
    apply(rule_tac t="b ab" in feasible_nnegD, blast+)[]

   apply(rule_tac t="a ab" in mono_transD, blast+)[]

   apply(subst scalingD[where t="a ab"], blast, blast, blast)
   apply(rule_tac x=x in fun_cong)
   apply(rule_tac f="a ab" in arg_cong)
   apply(blast intro!:ext)
   done

lemma healthy_wp_Seq:
  "\<lbrakk> healthy (wp a); healthy (wp b) \<rbrakk> \<Longrightarrow> healthy (wp (a ;; b))"
  by(simp add:wp_def healthy_Seq)

lemma healthy_wlp_Seq:
  "\<lbrakk> healthy (wlp a); healthy (wlp b) \<rbrakk> \<Longrightarrow> healthy (wlp (a ;; b))"
  by(simp add:wlp_def healthy_Seq)

lemma healthy_PC:
  "\<lbrakk> healthy (a ab); healthy (b ab); nneg P; bounded_by 1 P \<rbrakk> \<Longrightarrow>
   healthy ((a \<^bsub>P\<^esub>\<oplus> b) ab)"
  apply(subgoal_tac "\<And>x. 0 \<le> 1 - P x") prefer 2
   apply(subgoal_tac "P x \<le> 1", simp)
   apply(blast intro:bounded_byD)
  apply(safe intro!:healthyI feasibleI mono_transI scalingI,
        simp_all add:PC_def)
(* feasible *)
     apply(rule bounded_byI)
     apply(rule order_trans)
      apply(rule add_mono)
       apply(rule mult_left_mono)
       apply(rule_tac x=x in bounded_byD)
       apply(rule_tac t="a ab" in feasible_boundedD, force+)[]
      apply(rule_tac P=P in nnegD, assumption)
     apply(rule mult_left_mono)[]
      apply(rule_tac x=x in bounded_byD)
      apply(rule_tac t="b ab" in feasible_boundedD, force+)[]
     apply(simp)
    apply(simp add:algebra_simps)
   apply(rule nnegI)
   apply(rule add_nonneg_nonneg)
    apply(rule mult_nonneg_nonneg)
     apply(rule_tac P=P in nnegD, assumption)
    apply(rule_tac x=x in nnegD)
    apply(rule_tac t="a ab" in feasible_nnegD, force+)[]
   apply(rule mult_nonneg_nonneg, simp)
   apply(rule_tac P="b ab Pa" in nnegD)
   apply(rule_tac t="b ab" in feasible_nnegD, force+)[]
(* mono *)
   apply(rule le_funI)
   apply(rule add_mono)
    apply(rule mult_left_mono)
     apply(rule_tac t1="a ab" in le_funD[OF mono_transD], force+)[]
    apply(rule_tac x=x in nnegD, assumption)
   apply(rule mult_left_mono)
    apply(rule_tac t1="b ab" in le_funD[OF mono_transD], force+)[]
   apply(simp)
(* scaling *)
  apply(subst distrib_left)
  apply(rule_tac f="op +" in arg_cong2)
   apply(force intro!:scalingD)
  apply(force intro!:scalingD)
  done

lemma healthy_wp_PC:
  "\<lbrakk> healthy (wp a); healthy (wp b); nneg P; bounded_by 1 P \<rbrakk> \<Longrightarrow>
   healthy (wp (a \<^bsub>P\<^esub>\<oplus> b))"
  by(simp add:wp_def healthy_PC)

lemma healthy_wlp_PC:
  "\<lbrakk> healthy (wlp a); healthy (wlp b); nneg P; bounded_by 1 P \<rbrakk> \<Longrightarrow>
   healthy (wlp (a \<^bsub>P\<^esub>\<oplus> b))"
  by(simp add:wlp_def healthy_PC)

lemma healthy_DC:
  "\<lbrakk> healthy (a ab); healthy (b ab) \<rbrakk> \<Longrightarrow> healthy ((a \<Sqinter> b) ab)"
  apply(safe intro!:healthyI feasibleI mono_transI scalingI,
        simp_all add:DC_def)
     apply(rule bounded_byI)
     apply(rule min_max.le_infI1)
     apply(rule_tac x=x in bounded_byD)
     apply(simp add:healthy_bounded_byD)
    apply(rule nnegI)
    apply(rule min_max.le_infI)
     apply(rule_tac x=x in nnegD, simp)
    apply(rule_tac x=x in nnegD, simp)
   apply(rule le_funI)
   apply(rule min_max.inf_mono)
    apply(rule_tac t1="a ab" in le_funD[OF mono_transD], force+)[]
   apply(rule_tac t1="b ab" in le_funD[OF mono_transD], force+)[]
  apply(simp add:min_distrib)
  apply(rule_tac f=min in arg_cong2, (force intro!:scalingD)+)[]
  done

lemma healthy_wp_DC:
  "\<lbrakk> healthy (wp a); healthy (wp b) \<rbrakk> \<Longrightarrow> healthy (wp (a \<Sqinter> b))"
  by(simp add:wp_def healthy_DC)

lemma healthy_wlp_DC:
  "\<lbrakk> healthy (wlp a); healthy (wlp b) \<rbrakk> \<Longrightarrow> healthy (wlp (a \<Sqinter> b))"
  by(simp add:wlp_def healthy_DC)

lemma healthy_wp_Mu:
  "healthy (wp (\<mu> x. P x))"
  unfolding wp_def Mu_def by(simp)

lemma healthy_wlp_Mu:
  "healthy (wlp (\<mu> x. P x))"
  unfolding wlp_def Mu_def by(simp)

lemma healthy_wp_Embed:
  "healthy t \<Longrightarrow> healthy (wp (Embed t))"
  unfolding wp_def Embed_def by(simp)

lemma healthy_wlp_Embed:
  "healthy t \<Longrightarrow> healthy (wlp (Embed t))"
  unfolding wlp_def Embed_def by(simp)

lemma healthy_wp_repeat:
  assumes h_a: "healthy (wp a)"
  shows "healthy (wp (repeat n a))" (is "?X n")
proof(induct n)
  show "?X 0" by(simp add:healthy_wp_Skip)
next
  fix n assume IH: "?X n"
  thus "?X (Suc n)" by(simp add:healthy_wp_Seq h_a)
qed

lemma healthy_SetDC:
  fixes prog::"'b \<Rightarrow> 'a prog" and S::"'a \<Rightarrow> 'b set"
  assumes healthy:  "\<And>x s. x \<in> S s \<Longrightarrow> healthy ((prog x) ab)"
      and nonempty: "\<And>s. \<exists>x. x \<in> S s"
  shows "healthy ((SetDC prog S) ab)" (is "healthy ?T")
proof
  show "feasible ?T"
  proof
    fix b::real and P::"'a \<Rightarrow> real"
    assume bby: "bounded_by b P" and nneg: "nneg P"
    show "bounded_by b (?T P)"
    proof(rule bounded_byI)
      fix s::'a
      from nonempty obtain x where xin: "x \<in> S s" by(blast)
      with healthy and bby and nneg
      have "SetDC prog S ab P s \<le> prog x ab P s"
        unfolding SetDC_def by(blast intro!:glb_lower)
      also from healthy and bby and nneg and xin have "... \<le> b" by(blast)
      finally show "SetDC prog S ab P s \<le> b" .
    qed
    from healthy and nneg and bby
    have "\<And>x s. x \<in> S s \<Longrightarrow> 0 \<le> prog x ab P s" by(blast)
    have "\<And>x. 0 \<le> SetDC prog S ab P x"
      unfolding SetDC_def
      apply -
      apply(rule glb_greatest)
      using nonempty apply(blast)
      using healthy and nneg and bby apply(blast)
      done
    thus "nneg (SetDC prog S ab P)" by(blast)
  qed

  show "mono_trans ?T"
  proof
    fix P::"'a \<Rightarrow> real" and Q
    assume sP: "sound P" and sQ: "sound Q" and le: "P \<tturnstile> Q"
    have "\<And>x s. x \<in> S s \<Longrightarrow> glb ((\<lambda>x. prog x ab P s) ` S s) \<le> prog x ab Q s"
    proof -
      fix x and s
      assume xin: "x \<in> S s"
      from healthy and sP and xin
      have "glb ((\<lambda>x. prog x ab P s) ` S s) \<le> prog x ab P s"
        by(blast intro:glb_lower)
      also from healthy and sP and sQ and le and xin
      have "... \<le> prog x ab Q s" by(blast)
      finally show "glb ((\<lambda>x. prog x ab P s) ` S s) \<le> prog x ab Q s" .
    qed
    hence "\<And>x. glb ((\<lambda>a. prog a ab P x) ` S x) \<le>
               glb ((\<lambda>a. prog a ab Q x) ` S x)"
      using nonempty by(blast intro!:glb_greatest)
    thus "?T P \<tturnstile> ?T Q" by(unfold SetDC_def, blast)
  qed

  show "scaling ?T"
  proof
    fix P::"'a \<Rightarrow> real" and c::real and s
    assume sound: "sound P" and pos: "0 < c"
    from healthy have scaling: "\<And>x. x \<in> S s \<Longrightarrow> scaling (prog x ab)"
      by(blast)
    show "c * ?T P s = ?T (\<lambda>s. c * P s) s" (is "?X = ?Y") proof(rule antisym)
      show "?X \<le> ?Y"
        unfolding SetDC_def
        apply(rule glb_greatest)
         using nonempty apply(blast)
        apply(safe)
        apply(simp add:scalingD[OF scaling sound pos, symmetric])
        apply(rule mult_left_mono)
         prefer 2 using pos apply(rule less_imp_le)
        using healthy and sound apply(blast intro!:glb_lower)
        done

      show "?Y \<le> ?X"
        unfolding SetDC_def
        apply(rule mult_div_mono_right)
          using pos apply(simp)
         using pos apply(simp)
        apply(rule glb_greatest)
         using nonempty apply(blast)
        apply(safe)
        apply(rule mult_div_mono_left)
          using pos apply(simp)
         using pos apply(simp)
        apply(simp)
        apply(simp add:scalingD[OF scaling sound pos])
        apply(rule glb_lower)
         apply(blast)
        apply(rule_tac x=0 in exI, safe)
        apply(rule_tac x=s in nnegD)
        apply(rule_tac t="prog aba ab" in healthy_nnegD)
         apply(blast intro: healthy)
        using pos and sound apply(blast intro:less_imp_le sc_sound)
        done
    qed
  qed
qed

lemma healthy_wp_SetDC:
  "\<lbrakk> \<And>x s. x \<in> S s \<Longrightarrow> healthy (wp (prog x)); \<And>s. \<exists>x. x \<in> S s \<rbrakk> \<Longrightarrow>
   healthy (wp (SetDC prog S))"
  unfolding wp_def
  by(rule healthy_SetDC)

lemma healthy_wlp_SetDC:
  "\<lbrakk> \<And>x s. x \<in> S s \<Longrightarrow> healthy (wlp (prog x)); \<And>s. \<exists>x. x \<in> S s \<rbrakk> \<Longrightarrow>
   healthy (wlp (SetDC prog S))"
  unfolding wlp_def
  by(rule healthy_SetDC)

lemma healthy_SetPC:
  fixes p::"'s \<Rightarrow> 'a \<Rightarrow> real"
  and   f::"'a \<Rightarrow> 's prog"
  and   S::"'s \<Rightarrow> 'a set"
  assumes healthy: "\<And>a s. a \<in> S s \<Longrightarrow> healthy (f a ab)"
      and sound: "\<And>s. sound (p s)"
      and sub_dist: "\<And>s. (\<Sum>a\<in>(S s). p s a) \<le> 1"
  shows "healthy (SetPC f S p ab)" (is "healthy ?X")
proof
  show "feasible ?X"
    unfolding SetPC_def
  proof(safe intro!:feasibleI bounded_byI nnegI)
    fix b::real and P::"'s \<Rightarrow> real" and s::'s
    assume bby: "bounded_by b P" and nneg: "nneg P"

    from sound have "\<And>a. 0 \<le> p s a"
      by(blast)
    with healthy and sound and bby and nneg
    show "0 \<le> (\<Sum>a\<in>S s. p s a * f a ab P s)"
      apply -
      apply(rule setsum_nonneg)
      apply(clarify)
      apply(rule mult_nonneg_nonneg)
       apply(force)
      by(blast intro!:setsum_nonneg mult_nonneg_nonneg)

    show "(\<Sum>a\<in>S s. p s a * f a ab P s) \<le> b"
      apply(rule_tac y="(\<Sum>a\<in>S s. p s a * b)" in order_trans)
       apply(rule setsum_mono)
       apply(rule mult_left_mono)
        using bby and nneg and sound and healthy apply(blast)
       using sound apply(blast)
      apply(simp add:setsum_left_distrib[symmetric])
      apply(rule order_trans)
       apply(rule mult_right_mono)
        apply(rule sub_dist)
       using bby and nneg apply(blast)
      apply(simp)
      done
  qed

  show "mono_trans ?X"
    unfolding SetPC_def
  proof(rule mono_transI, rule le_funI)
    fix P::"'s \<Rightarrow> real" and Q::"'s \<Rightarrow> real" and s
    assume s_P: "sound P" and s_Q: "sound Q" and ent: "P \<tturnstile> Q"
    with healthy have "\<And>a. a \<in> S s \<Longrightarrow> f a ab P s \<le> f a ab Q s"
      by(blast)
    with sound show "(\<Sum>a\<in>S s. p s a * f a ab P s) \<le> (\<Sum>a\<in>S s. p s a * f a ab Q s)"
      by(blast intro:setsum_mono mult_left_mono)
  qed

  show "scaling ?X"
    unfolding SetPC_def
  proof
    fix P::"'s \<Rightarrow> real" and c::real and s::'s
    assume sound: "sound P" and pos: "0 < c"
    have "c * (\<Sum>a\<in>S s. p s a * f a ab P s) = (\<Sum>a\<in>S s. p s a * (c * f a ab P s))"
      (is "?A = ?B")
      by(simp add:setsum_right_distrib ac_simps)
    also from sound and pos and healthy
    have "... = (\<Sum>a\<in>S s. p s a * f a ab (\<lambda>s. c * P s) s)"
      by(force simp:scalingD[OF healthy_scalingD])
    finally show "?A = ..." .
  qed
qed

lemma healthy_wlp_SetPC:
  "\<lbrakk> \<And>a s. a \<in> S s \<Longrightarrow> healthy (wlp (f a)); \<And>s. sound (p s);
    \<And>s. (\<Sum>a\<in>(S s). p s a) \<le> 1 \<rbrakk> \<Longrightarrow>
   healthy (wlp (SetPC f S p))"
  unfolding wlp_def
  by(blast intro: healthy_SetPC)

lemma healthy_wp_SetPC:
  "\<lbrakk> \<And>a s. a \<in> S s \<Longrightarrow> healthy (wp (f a)); \<And>s. sound (p s);
     \<And>s. (\<Sum>a\<in>(S s). p s a) \<le> 1 \<rbrakk> \<Longrightarrow>
   healthy (wp (SetPC f S p))"
  unfolding wp_def
  by(blast intro: healthy_SetPC)

lemma healthy_Apply:
  "healthy ((Apply f) ab)"
  unfolding Apply_def by(blast)

lemma healthy_wp_Apply:
  "healthy (wp (Apply f))"
  unfolding wp_def by(rule healthy_Apply)

lemma healthy_wlp_Apply:
  "healthy (wlp (Apply f))"
  unfolding wlp_def by(rule healthy_Apply)

lemma healthy_Bind:
  assumes hsub: "\<And>s. healthy (p (f s) ab)"
  shows "healthy (Bind f p ab)"
  unfolding Bind_def Let_def
proof
  let "?X P s" = "p (f s) ab P s"
  show "feasible ?X"
  proof
    fix b and P::"'a \<Rightarrow> real"
    assume bP: "bounded_by b P" and nP: "nneg P"
    show "bounded_by b (?X P)"
      using bP and nP and hsub by(blast intro!:bounded_byI)
    show "nneg (?X P)"
      using hsub and nP and bP by(blast)
  qed
  show "mono_trans ?X"
  proof
    fix P::"'a \<Rightarrow> real" and Q
    assume sP: "sound P" and sQ: "sound Q" and ent: "P \<tturnstile> Q"
    show "?X P \<tturnstile> ?X Q"
    proof(rule le_funI)
      fix s
      from hsub have "mono_trans (p (f s) ab)" by(blast)
      thm mono_transD
      with sP and sQ and ent
      have "p (f s) ab P \<tturnstile> p (f s) ab Q" by(blast)
      thus "?X P s \<le> ?X Q s" by(rule le_funD)
    qed
  qed
  show "scaling ?X"
  proof
    fix P::"'a \<Rightarrow> real" and c::real and s
    assume sP: "sound P" and pos: "0 < c"
    from hsub have "scaling (p (f s) ab)" by(blast)
    with sP and pos show "c * ?X P s = ?X (\<lambda>s. c * P s) s" by(blast)
  qed
qed

lemma healthy_wp_Bind:
  "\<lbrakk> \<And>s. healthy (wp (p (f s))) \<rbrakk> \<Longrightarrow> healthy (wp (Bind f p))"
  unfolding wp_def by(simp add:healthy_Bind)

lemma healthy_wlp_Bind:
  "\<lbrakk> \<And>s. healthy (wp (p (f s))) \<rbrakk> \<Longrightarrow> healthy (wp (Bind f p))"
  unfolding wp_def by(simp add:healthy_Bind)

(* XXX - complete these *)
text {* Use 'simp add:healthy_intros' or 'blast intro:healthy_intros' as
        appropriate to discharge healthiness side-contitions for primitive
        programs automatically. *}
lemmas healthy_intros =
  healthy_wp_Abort healthy_wlp_Abort healthy_wp_Skip healthy_wlp_Skip
  healthy_wp_Seq healthy_wlp_Seq healthy_wp_PC healthy_wlp_PC
  healthy_wp_DC healthy_wlp_DC healthy_wp_Mu healthy_wlp_Mu
  healthy_wp_Embed healthy_wlp_Embed healthy_wp_Apply healthy_wlp_Apply
  healthy_wp_SetDC healthy_wlp_SetDC healthy_wp_SetPC healthy_wlp_SetPC
  healthy_wp_Bind healthy_wlp_Bind healthy_wp_repeat

text {* These soundness lemmas are useful in conjunction with the above,
        to discharge soundness for common forms *}
lemmas sound_intros = standard_sound mult_sound div_sound const_sound sound_sum

subsection {* Transformer-transformer monotonicity *}

text {* This is required to invoke the fixed-point lemmas *}

lemma mono_ttrans_const:
  "mono_ttrans (\<lambda>t. c)"
  by(force intro:mono_ttransI)

lemma mono_ttrans_Seq:
  assumes monoT: "mono_ttrans (\<lambda>t. T t ab)"
      and monoU: "mono_ttrans (\<lambda>t. U t ab)"
      and hprT:  "\<And>t. healthy t \<Longrightarrow> healthy (T t ab)"
      and hprU:  "\<And>t. healthy t \<Longrightarrow> healthy (U t ab)"
  shows "mono_ttrans (\<lambda>t. ((T t) ;; (U t)) ab)" (is "mono_ttrans ?X")
proof(rule mono_ttransI)
  fix t::"('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real" and u
  assume ht: "healthy t" and hu: "healthy u" and le: "le_trans t u"
  from ht and hprU have X: "\<And>P. sound P \<Longrightarrow> sound (U t ab P)" by(blast)
  from hu and hprU have Y: "\<And>P. sound P \<Longrightarrow> sound (U u ab P)" by(blast)
  from ht and hu and le and monoT
  have "le_trans ((T t ;; U t) ab) ((T u ;; U t) ab)"
    unfolding Seq_def o_def
    by(blast dest!:mono_ttransD intro!:X le_transI)
  also have "le_trans ... ((T u ;; U u) ab)"
    unfolding Seq_def o_def
    apply -
    apply(rule le_transI)
    apply(rule_tac t="T u ab" in mono_transD)
       using hu and hprT apply(blast)
      using X apply(blast)
     using Y apply(blast)
    using ht and hu and le
    apply(blast intro:mono_ttransD[OF monoU] le_transD)
    done
  finally show "le_trans (?X t) (?X u)" .
qed

lemma mono_ttrans_wp_Seq:
  "\<lbrakk> mono_ttrans (\<lambda>t. wp (T t)); mono_ttrans (\<lambda>t. wp (U t));
     \<And>t. healthy t \<Longrightarrow> healthy (wp (T t)); \<And>t. healthy t \<Longrightarrow> healthy (wp (U t)) \<rbrakk> \<Longrightarrow>
   mono_ttrans (\<lambda>t. wp ((T t) ;; (U t)))"
  unfolding wp_def by(simp add:mono_ttrans_Seq)

lemma mono_ttrans_wlp_Seq:
  "\<lbrakk> mono_ttrans (\<lambda>t. wlp (T t)); mono_ttrans (\<lambda>t. wlp (U t));
     \<And>t. healthy t \<Longrightarrow> healthy (wlp (T t));
     \<And>t. healthy t \<Longrightarrow> healthy (wlp (U t)) \<rbrakk> \<Longrightarrow>
   mono_ttrans (\<lambda>t. wlp ((T t) ;; (U t)))"
  unfolding wlp_def by(simp add:mono_ttrans_Seq)

lemma mono_ttrans_PC:
  "\<lbrakk> mono_ttrans (\<lambda>t. T t ab); mono_ttrans (\<lambda>t. U t ab);
     \<And>t. healthy t \<Longrightarrow> healthy (T t ab); \<And>t. healthy t \<Longrightarrow> healthy (U t ab);
     bounded_by 1 Q; nneg Q \<rbrakk> \<Longrightarrow>
   mono_ttrans (\<lambda>t. ((T t) \<^bsub>Q\<^esub>\<oplus> (U t)) ab)"
  apply(rule mono_ttransI)
  apply(rule le_transI)
  apply(simp add:PC_def)
  apply(rule le_funI)
  apply(rule add_mono)
   apply(rule mult_left_mono, simp_all add:nnegD)
   apply(rule_tac x=x in le_funD)
   apply(rule_tac P=P in le_transD, simp_all)[]
   apply(rule_tac T="\<lambda>t. T t ab" in mono_ttransD, simp_all)[]
  apply(rule mult_left_mono)
  apply(rule_tac x=x in le_funD)
  apply(rule_tac P=P in le_transD, simp_all)[]
  apply(rule_tac T="\<lambda>t. U t ab" in mono_ttransD, simp_all)[]
  apply(subgoal_tac "Q x \<le> 1", simp)
  apply(blast intro:bounded_byD)
  done

lemma mono_ttrans_wp_PC:
  "\<lbrakk> mono_ttrans (\<lambda>t. wp (T t)); mono_ttrans (\<lambda>t. wp (U t));
     \<And>t. healthy t \<Longrightarrow> healthy (wp (T t)); \<And>t. healthy t \<Longrightarrow> healthy (wp (U t));
     bounded_by 1 Q; nneg Q \<rbrakk> \<Longrightarrow>
   mono_ttrans (\<lambda>t. wp ((T t) \<^bsub>Q\<^esub>\<oplus> (U t)))"
  unfolding wp_def by(simp add:mono_ttrans_PC)

lemma mono_ttrans_wlp_PC:
  "\<lbrakk> mono_ttrans (\<lambda>t. wlp (T t)); mono_ttrans (\<lambda>t. wlp (U t));
     \<And>t. healthy t \<Longrightarrow> healthy (wlp (T t)); \<And>t. healthy t \<Longrightarrow> healthy (wlp (U t));
     bounded_by 1 Q; nneg Q \<rbrakk> \<Longrightarrow>
   mono_ttrans (\<lambda>t. wlp ((T t) \<^bsub>Q\<^esub>\<oplus> (U t)))"
  unfolding wlp_def by(simp add:mono_ttrans_PC)

lemma mono_ttrans_DC:
  "\<lbrakk> mono_ttrans (\<lambda>t. T t ab); mono_ttrans (\<lambda>t. U t ab);
     \<And>t. healthy t \<Longrightarrow> healthy (T t ab); \<And>t. healthy t \<Longrightarrow> healthy (U t ab);
     bounded_by 1 Q; nneg Q \<rbrakk> \<Longrightarrow>
   mono_ttrans (\<lambda>t. ((T t) \<Sqinter> (U t)) ab)"
  apply(rule mono_ttransI, rule le_transI)
  apply(simp add:DC_def)
  apply(rule le_funI)
  apply(rule min_max.inf_mono)
   apply(rule_tac x=x in le_funD)
   apply(rule_tac P=P in le_transD, simp_all)[]
   apply(rule_tac T="\<lambda>t. T t ab" in mono_ttransD, simp_all)[]
  apply(rule_tac x=x in le_funD)
  apply(rule_tac P=P in le_transD, simp_all)[]
  apply(rule_tac T="\<lambda>t. U t ab" in mono_ttransD, simp_all)[]
  done

lemma mono_ttrans_wp_DC:
  "\<lbrakk> mono_ttrans (\<lambda>t. wp (T t)); mono_ttrans (\<lambda>t. wp (U t));
     \<And>t. healthy t \<Longrightarrow> healthy (wp (T t)); \<And>t. healthy t \<Longrightarrow> healthy (wp (U t));
     bounded_by 1 Q; nneg Q \<rbrakk> \<Longrightarrow>
   mono_ttrans (\<lambda>t. wp ((T t) \<Sqinter> (U t)))"
  unfolding wp_def by(simp add:mono_ttrans_DC)

lemma mono_ttrans_wlp_DC:
  "\<lbrakk> mono_ttrans (\<lambda>t. wlp (T t)); mono_ttrans (\<lambda>t. wlp (U t));
     \<And>t. healthy t \<Longrightarrow> healthy (wlp (T t)); \<And>t. healthy t \<Longrightarrow> healthy (wlp (U t));
     bounded_by 1 Q; nneg Q \<rbrakk> \<Longrightarrow>
   mono_ttrans (\<lambda>t. wlp ((T t) \<Sqinter> (U t)))"
  unfolding wlp_def by(simp add:mono_ttrans_DC)

lemma mono_ttrans_wp_Mu:
  "\<lbrakk> \<And>u. mono_ttrans (\<lambda>t. wp (T (Embed t) u));
     \<And>t. mono_ttrans (\<lambda>u. wp (T (Embed t) (Embed u)));
     \<And>t u. healthy t \<Longrightarrow> healthy (wp (T (Embed t) u)) \<rbrakk> \<Longrightarrow>
   mono_ttrans (\<lambda>t. wp (\<mu> u. T (Embed t) u))"
  unfolding Mu_def wp_def
  apply(simp)
  apply(rule mono_ttransI)
  apply(rule le_transI)
  apply(rule le_transD[OF lfp_trans_lowerbound], simp_all)[]
   apply(rule_tac T="\<lambda>u. T (Embed t) (Embed u) True" in mono_ttrans_equiv_trans,
         simp_all)[]
  apply(rule_tac
    y="T (Embed u) (Embed (lfp_trans (\<lambda>t. T (Embed u) (Embed t) True))) True"
    in le_trans_trans)
  apply(rule_tac
    T="\<lambda>t. T (Embed t)
             (Embed (lfp_trans (\<lambda>t. T (Embed u)
                                     (Embed t) True))) True"
    in mono_ttransD, simp_all)[]
  apply(rule lfp_trans_lemma2, simp_all)
  done

lemma mono_ttrans_wlp_Mu:
  "\<lbrakk> \<And>u. mono_ttrans (\<lambda>t. wlp (T (Embed t) u));
     \<And>t. mono_ttrans (\<lambda>u. wlp (T (Embed t) (Embed u)));
     \<And>t u. healthy t \<Longrightarrow> healthy (wlp (T (Embed t) u)) \<rbrakk> \<Longrightarrow>
   mono_ttrans (\<lambda>t. wlp (\<mu> u. T (Embed t) u))"
  unfolding Mu_def wlp_def
  apply(simp)
  apply(rule mono_ttransI)
  apply(rule le_transI)
  apply(rule le_transD[OF gfp_trans_upperbound], simp_all)[]
   apply(rule_tac T="\<lambda>t. T (Embed u) (Embed t) False" in mono_ttrans_equiv_trans,
         simp_all)[]
  apply(rule_tac
    y="T (Embed t) (Embed (gfp_trans (\<lambda>u. T (Embed t) (Embed u) False))) False"
    in le_trans_trans)
   apply(rule gfp_trans_lemma2, simp_all)[]
  apply(rule_tac
    T="\<lambda>t'. T (Embed t')
             (Embed (gfp_trans (\<lambda>u. T (Embed t)
                                     (Embed u) False))) False"
    in mono_ttransD, simp_all)[]
  done

lemma mono_ttrans_wp_Embed:
  "mono_ttrans (\<lambda>t. wp (Embed t))"
  unfolding wp_def Embed_def
  by(simp add:mono_ttransI)

lemma mono_ttrans_wlp_Embed:
  "mono_ttrans (\<lambda>t. wlp (Embed t))"
  unfolding wlp_def Embed_def
  by(simp add:mono_ttransI)

lemmas mono_ttrans_intros = mono_ttrans_const
  mono_ttrans_wp_Seq mono_ttrans_wlp_Seq
  mono_ttrans_wp_PC mono_ttrans_wlp_PC
  mono_ttrans_wp_DC mono_ttrans_wlp_DC
  mono_ttrans_wp_Mu mono_ttrans_wlp_Mu
  mono_ttrans_wp_Embed mono_ttrans_wlp_Embed

section {* Loops *}

text {* This gives the equivalence with the alternative definition for loops
        (footnote 23, section 7, page 198 of ARPPS) *}

lemma wlp_Loop1:
  assumes sound:   "sound P"
      and healthy: "healthy (wlp body)"
  shows "wlp (do G \<longrightarrow> body od) P =
   gfp_in (\<lambda>Q s. \<guillemotleft>G\<guillemotright> s * wlp body Q s + \<guillemotleft>\<^boverline>G\<^eoverline>\<guillemotright> s * P s)
   (weakly_bounded_by P)"
proof -
  have loop_biscaling:
    "\<And>P Q c s. \<lbrakk> sound P; sound Q; 0 < c \<rbrakk> \<Longrightarrow>
        c * (\<guillemotleft> G \<guillemotright> s * wlp body Q s + \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> s * P s) =
        \<guillemotleft> G \<guillemotright> s * wlp body (\<lambda>a. c * Q a) s + \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> s * (c * P s)"
    by(simp add:healthy healthy_scalingD scalingD[symmetric] algebra_simps)

  have loop_bounded_by:
    "\<And>P Q. \<lbrakk> sound P; sound Q; bound_of Q \<le> bound_of P \<rbrakk> \<Longrightarrow>
             bounded_by (bound_of P)
                        (\<lambda>b. \<guillemotleft> G \<guillemotright> b * wlp body Q b + \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> b * P b)"
    apply(rule bounded_byI)
     apply(rule order_trans)
      apply(rule add_mono)
       apply(rule_tac b="bound_of P" in mult_left_mono)
        apply(rule order_trans) prefer 2 apply(assumption)
        apply(rule_tac x=x in bounded_byD)
        using healthy apply(blast)
       apply(blast intro:nnegD)
      apply(rule_tac b="bound_of P" in mult_left_mono)
       apply(blast)
      apply(blast intro:nnegD)
     apply(subst embed_split[symmetric], simp)
     done

   have loop_sound:
     "\<And>P Q. \<lbrakk>sound P; sound Q; bound_of Q \<le> bound_of P\<rbrakk> \<Longrightarrow>
             sound (\<lambda>b. \<guillemotleft> G \<guillemotright> b * wlp body Q b + \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> b * P b)"
     apply(rule soundI)
      apply(blast intro:loop_bounded_by)
     apply(rule nnegI)
     apply(rule add_nonneg_nonneg)
      apply(rule mult_nonneg_nonneg)
       apply(blast intro:nnegD)
      using healthy apply(blast)
     apply(rule mult_nonneg_nonneg)
      apply(blast intro:nnegD)
     apply(blast)
     done

   have loop_weak_closed:
     "\<And>P. sound P \<Longrightarrow>
      closed_under (\<lambda>a b. \<guillemotleft> G \<guillemotright> b * wlp body a b + \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> b * P b)
                   (weakly_bounded_by P)"
     apply(rule closed_underI)
     unfolding weakly_bounded_by_def
     apply(simp, safe)
      apply(rule loop_sound, assumption+)
     apply(blast intro:loop_bounded_by)
     done

   have loop_mono:
     "\<And>P. sound P \<Longrightarrow>
      mono_on (\<lambda>a b. \<guillemotleft> G \<guillemotright> b * wlp body a b + \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> b * P b)
              (weakly_bounded_by P)"
     apply(rule mono_onI)
     apply(rule le_funI)
     apply(rule add_right_mono)
     apply(rule mult_left_mono)
      apply(rule_tac x=xa in le_funD)
      apply(rule_tac t="wlp body" in mono_transD)
         using healthy apply(blast)
        using weakly_bounded_sound apply(blast)
       using weakly_bounded_sound apply(blast)
      apply(assumption)
     apply(blast intro:nnegD)
     done

   from sound and healthy show ?thesis
     apply(simp add:wp_eval)
      apply(rule gfp_pulldown, simp_all add:healthy_intros mono_ttrans_intros)
       apply(simp add:wp_eval negate_embed)
      apply(rule loop_weak_closed, assumption)
     apply(rule loop_mono, assumption)
    apply(rule le_funI)
    apply(rule add_left_mono)
    apply(rule mult_left_mono)
     apply(blast)
    apply(blast intro:nnegD)
   apply(simp add: scalingD[OF healthy_scalingD, OF healthy, symmetric]
                   algebra_simps)
   done
qed

subsection {* Liberal and strict invariants. *}

definition
  wp_inv :: "('s \<Rightarrow> bool) \<Rightarrow> 's prog \<Rightarrow> ('s \<Rightarrow> real) \<Rightarrow> bool"
where
  "wp_inv G body I \<longleftrightarrow> (\<forall>s. \<guillemotleft>G\<guillemotright> s * I s \<le> wp body I s)"

definition
  wlp_inv :: "('s \<Rightarrow> bool) \<Rightarrow> 's prog \<Rightarrow> ('s \<Rightarrow> real) \<Rightarrow> bool"
where
  "wlp_inv G body I \<longleftrightarrow> (\<forall>s. \<guillemotleft>G\<guillemotright> s * I s \<le> wlp body I s)"

lemma wlp_invD:
  "wlp_inv G body I \<Longrightarrow> \<guillemotleft>G\<guillemotright> s * I s \<le> wlp body I s"
  by(simp add:wlp_inv_def)

text {* Partial correctness for loops (Lemma 7.2.2, page 185) *}
lemma wlp_Loop:
  "\<lbrakk> healthy (wlp body); sound I; I \<in> weakly_bounded_by (\<lambda>s. \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> s * I s) \<rbrakk> \<Longrightarrow>
   wlp_inv G body I \<Longrightarrow> I \<le> wlp do G \<longrightarrow> body od (\<lambda>s. \<guillemotleft>\<^boverline>G\<^eoverline>\<guillemotright> s * I s)"
  apply(subst wlp_Loop1) prefer 3
  apply(rule gfp_in_upperbound) prefer 3
  apply(rule le_funI)
  apply(rule order_trans) prefer 2
   apply(rule add_right_mono)
   apply(rule mult_left_mono)
    apply(rule wlp_invD, assumption)
   apply(blast)
  apply(simp only:mult_assoc[symmetric] embed_bool_idem)
  apply(simp add:embed_split[symmetric])
     apply(blast intro:weakly_bounded_intros sound_intros)
    apply(blast intro:weakly_bounded_byI weakly_bounded_bound)
   apply(blast intro:sound_intros)
  apply(assumption)
  done

lemma nentails_weakly_bounded:
  fixes G and I
  assumes nentails: "\<not>(\<forall>s. I s \<longrightarrow> G s)"
  shows "\<guillemotleft>I\<guillemotright> \<in> weakly_bounded_by (\<lambda>s. \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> s * \<guillemotleft>I\<guillemotright> s)"
proof -
  have le1: "\<And>s. \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> s * \<guillemotleft>I\<guillemotright> s \<le> 1" (is "\<And>s. ?X s")
  proof -
    fix s
    have Ile: "\<guillemotleft>I\<guillemotright> s \<le> 1" by(blast)
    have "\<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> s \<le> 1" by(blast)
    hence "\<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> s * \<guillemotleft>I\<guillemotright> s \<le> 1 * \<guillemotleft>I\<guillemotright> s" by(rule mult_mono, simp_all)
    also with Ile have "... \<le> 1" by(simp)
    finally show "?X s" .
  qed
  hence bounded: "bounded (\<lambda>s. \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> s * \<guillemotleft>I\<guillemotright> s)" by(blast)
  from le1 have upper: "bound_of (\<lambda>s. \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> s * \<guillemotleft>I\<guillemotright> s) \<le> 1"
    by(rule bound_of_leI)
  from nentails obtain s where "I s \<and> \<not>G s" by(blast)
  hence "1 \<le> \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> s * \<guillemotleft>I\<guillemotright> s" by(simp)
  also with bounded have "... \<le> bound_of (\<lambda>s. \<guillemotleft>\<^boverline>G\<^eoverline>\<guillemotright> s * \<guillemotleft>I\<guillemotright> s)"
    by(blast dest:bounded_by_bound_of)
  finally have lower: "1 \<le> ..." .
  from upper and lower have "... = 1" by(rule antisym)
  thus ?thesis
    unfolding weakly_bounded_by_def by(force)
qed

section "Sub-conjunctivity"

lemma sub_conj_wp_Bind:
  "\<lbrakk> \<And>s x. x = f s \<Longrightarrow> sub_conj (wp (a x)) \<rbrakk> \<Longrightarrow>
  sub_conj (wp (Bind f a))"
  apply(simp add:wp_eval)
  apply(rule_tac f="\<lambda>s. wp (a (f s))" in sub_conj_wp_twice)
  apply(simp)
  done

lemma sub_conj_wp_Apply:
  "sub_conj (wp (Apply f))"
  by(rule sub_conjI, simp add:wp_eval)

lemma sub_conj_wp_Abort:
  "sub_conj (wp Abort)"
  apply(rule sub_conjI)
  unfolding wp_eval exp_conj_def by(simp)

lemma sub_conj_wp_Seq:
  "\<lbrakk> sub_conj (wp a); sub_conj (wp b);
     healthy (wp a); healthy (wp b) \<rbrakk> \<Longrightarrow> sub_conj (wp (a ;; b))"
  apply(rule sub_conjI)
  apply(simp add:wp_eval)
  apply(rule_tac y="wp a (wp b P && wp b Q)" in order_trans)
   apply(rule sub_conjD, assumption)
      apply(blast, blast)
    apply(rule_tac t="wp b" in healthy_bounded_byD, assumption+)
    apply(rule sound_nneg, assumption)
   apply(rule_tac t="wp b" in healthy_bounded_byD, assumption+)
   apply(rule sound_nneg, assumption)
  apply(rule_tac t="wp a" in mono_transD)
     apply(blast)
    apply(blast)
   apply(blast)
  apply(rule sub_conjD, assumption+)
  done

lemma sub_conj_wp_SetPC:
  "\<lbrakk> \<And>x s. x \<in> S s \<Longrightarrow> sub_conj (wp (a x)); \<And>s. finite (S s);
     \<And>x s. 0 \<le> p x s; \<And>x. setsum (p x) (S x) \<le> 1 \<rbrakk> \<Longrightarrow>
   sub_conj (wp (SetPC a S p))"
  apply(rule sub_conjI)
  apply(simp add:wp_eval)
  apply(rule le_funI)
  apply(simp add:exp_conj_def)
  apply(simp add:pconj_def)
  apply(rule order_trans)
  apply(rule eq_refl)
  apply(rule_tac f="\<lambda>s. s \<ominus> 1" in arg_cong)
  apply(rule setsum.distrib[symmetric], simp)
  apply(rule order_trans) prefer 2
  apply(rule setsum_mono)
   apply(rule mult_left_mono)
    apply(fold pconj_def)
    apply(subst exp_conj_def[symmetric])
    apply(rule_tac t1="wp (a aa)" in le_funD[OF sub_conjD], simp, assumption+)[]
   apply(assumption)
  apply(simp add:exp_conj_def pconj_def)
  apply(subst tminus_left_distrib, assumption)
  apply(rule order_trans) prefer 2
  apply(rule tminus_setsum_mono, assumption)
  apply(simp add:distrib_left)
  apply(rule tminus_right_antimono)
  apply(assumption)
  done

lemmas sub_conj_lemmas =
  sub_conj_wp_Abort sub_conj_wp_Seq sub_conj_wp_SetPC
  sub_conj_wp_Apply

section "Sub-distributivity"

definition
  sub_distrib_pconj :: "'s prog \<Rightarrow> bool"
where
  "sub_distrib_pconj prog \<equiv>
   \<forall>P Q. sound P \<and> bounded_by 1 P \<and> sound Q \<longrightarrow>
         wlp prog P && wp prog Q \<tturnstile> wp prog (P && Q)"

lemma sub_distrib_pconjI[intro]:
  "\<lbrakk>\<And>P Q. \<lbrakk> sound P; bounded_by 1 P; sound Q \<rbrakk> \<Longrightarrow>
          wlp prog P && wp prog Q \<tturnstile> wp prog (P && Q) \<rbrakk> \<Longrightarrow>
    sub_distrib_pconj prog"
  unfolding sub_distrib_pconj_def by(simp)

lemma sub_distrib_pconjD[dest]:
  "\<And>P Q. \<lbrakk> sub_distrib_pconj prog; sound P; bounded_by 1 P; sound Q \<rbrakk> \<Longrightarrow>
   wlp prog P && wp prog Q \<tturnstile> wp prog (P && Q)"
  unfolding sub_distrib_pconj_def by(simp)

lemma sdp_Abort:
  "sub_distrib_pconj Abort"
  apply(rule sub_distrib_pconjI)
  unfolding wp_eval
  apply(subst exp_conj_rzero)
   apply(blast)
  apply(simp)
  done

lemma sdp_Skip:
  "sub_distrib_pconj Skip"
  apply(rule sub_distrib_pconjI)
  unfolding wp_eval
  by(simp)

lemma sdp_Seq:
  fixes a and b
  assumes sdp_a:   "sub_distrib_pconj a"
      and sdp_b:   "sub_distrib_pconj b"
      and h_wp_a:  "healthy (wp a)"
      and h_wp_b:  "healthy (wp b)"
      and h_wlp_b: "healthy (wlp b)"
  shows "sub_distrib_pconj (a ;; b)"
proof(rule sub_distrib_pconjI, unfold wp_eval o_def)
  fix P::"'a \<Rightarrow> real" and Q::"'a \<Rightarrow> real"
  assume s_P: "sound P" and s_Q: "sound Q" and bby: "bounded_by 1 P"

  with sdp_a and h_wp_b and h_wlp_b
  have "wlp a (wlp b P) && wp a (wp b Q) \<tturnstile> wp a (wlp b P && wp b Q)"
    by(blast intro: sub_distrib_pconjD healthy_bounded_byD2)
  also {
    from sdp_b and s_P and s_Q and bby
    have "wlp b P && wp b Q \<tturnstile> wp b (P && Q)" by(blast)
    with h_wp_a and h_wp_b and h_wlp_b and s_P and s_Q
    have "wp a (wlp b P && wp b Q) \<tturnstile> wp a (wp b (P && Q))"
      by(blast intro:exp_conj_sound mono_transD)
  }
  finally show "wlp a (wlp b P) && wp a (wp b Q) \<tturnstile> wp a (wp b (P && Q))" .
qed

(* XXX - complete these *)
lemmas sdp_intros = sdp_Abort sdp_Skip sdp_Seq

subsection "Sub-additivity"

lemma sub_add_wp_Seq:
  "\<lbrakk> sub_add (wp a); sub_add (wp b);
     healthy (wp a); healthy (wp b) \<rbrakk> \<Longrightarrow> sub_add (wp (a ;; b))"
  apply(rule sub_addI)
  apply(simp add:wp_eval)
  apply(rule order_trans)
   apply(rule_tac t="wp a" in sub_addD, assumption)
    apply(blast)
   apply(blast)
  apply(rule_tac x=s in le_funD)
  apply(rule_tac t="wp a" in mono_transD)
     apply(blast)
    apply(blast)
   apply(blast)
  apply(rule le_funI)
  apply(rule_tac t="wp b" in sub_addD, assumption+)
  done

lemma sub_add_wp_SetPC:
  assumes sub_add_a: "\<And>x s. x \<in> S s \<Longrightarrow> sub_add (wp (a x))"
      and fin_S: "\<And>s. finite (S s)"
      and pnneg: "\<And>s x. 0 \<le> p s x"
  shows "sub_add (wp (SetPC a S p))"
  apply(rule sub_addI)
  apply(simp add:wp_eval)
  apply(rule order_trans)
   apply(rule eq_refl)
   apply(rule setsum.distrib[symmetric])
   apply(rule fin_S)
  apply(rule setsum_mono)
  apply(fold distrib_left)
  apply(rule mult_left_mono)
   apply(rule_tac t="wp (a x)" in sub_addD)
      apply(rule sub_add_a, assumption+)
  apply(rule pnneg)
  done

lemma sub_add_wp_SetDC:
  "\<lbrakk> \<And>x s. x \<in> S s \<Longrightarrow> sub_add (wp (a x));
     \<And>x s. x \<in> S s \<Longrightarrow> healthy (wp (a x));
     \<And>s. S s \<noteq> {} \<rbrakk> \<Longrightarrow>
   sub_add (wp (SetDC a S))"
  apply(rule sub_addI)
  apply(simp add:wp_eval)
  apply(rule glb_greatest)
   apply(rule nonempty_witness, simp)
  apply(clarsimp)
  apply(rule order_trans) prefer 2
   apply(rule_tac t="wp (a x)" in sub_addD)
     apply(blast, assumption+)
   apply(rule add_mono)
    apply(rule glb_lower)
     apply(blast)
    apply(rule_tac x=0 in exI)
    apply(clarsimp)
    apply(rule_tac x=s in nnegD)
    apply(rule_tac P=P in healthy_nnegD)
     apply(blast)
    apply(assumption)
   apply(rule glb_lower)
    apply(blast)
   apply(rule_tac x=0 in exI)
   apply(clarsimp)
   apply(rule_tac x=s in nnegD)
   apply(rule_tac P=Q in healthy_nnegD)
    apply(blast)
   apply(assumption)
   done

lemma sub_add_wp_Skip:
  "sub_add (wp Skip)"
  by(rule sub_addI, simp add:wp_eval)

lemma sub_add_wp_repeat:
  assumes sa_a: "sub_add (wp a)"
      and ha:   "healthy (wp a)"
  shows "sub_add (wp (repeat n a))" (is "?X n")
proof(induct n)
  show "?X 0"
    by(simp add:sub_add_wp_Skip)

  fix n
  assume IH: "?X n"
  show "?X (Suc n)"
    by(simp_all add:IH assms healthy_intros sub_add_wp_Seq)
qed

lemma sub_add_wp_Apply:
  "sub_add (wp (Apply f))"
  by(simp add:sub_addI wp_eval)

(* XXX - complete these *)
lemmas sa_intros =
  sub_add_wp_Seq sub_add_wp_SetPC sub_add_wp_SetDC
  sub_add_wp_Apply sub_add_wp_repeat sub_add_wp_Skip

section "Additivity"

lemma additive_wp_Apply:
  "additive (wp (Apply f))"
  by(rule additiveI, simp add:wp_eval)

lemma additive_wp_SetPC:
  assumes add_a: "\<And>x s. x \<in> S s \<Longrightarrow> additive (wp (a x))"
      and fin_S: "\<And>s. finite (S s)"
  shows "additive (wp (SetPC a S p))"
  by(rule additiveI,
     simp add:wp_eval additiveD[OF add_a] distrib_left
              setsum.distrib fin_S)

lemma additive_wp_Bind:
  "\<lbrakk> \<And>x. additive (wp (a (f x))) \<rbrakk> \<Longrightarrow> additive (wp (Bind f a))"
  by(simp add:wp_eval additive_def)

(* XXX - complete these *)
lemmas fa_intros = additive_wp_Apply additive_wp_SetPC additive_wp_Bind

subsection "Determinism"

lemma determ_wp_Skip:
  "determ (wp Skip)"
  by(safe intro!:determI additiveI maximalI,
     simp_all add:wp_eval)
  
lemma determ_wp_Apply:
  "determ (wp (Apply f))"
  by(safe intro!:determI additiveI maximalI,
     simp_all add:wp_eval o_def)

lemma determ_wp_Seq:
  fixes a::"'b prog"
  assumes da: "determ (wp a)"
      and db: "determ (wp b)"
      and ha: "healthy (wp a)"
      and hb: "healthy (wp b)"
  shows "determ (wp (a ;; b))"
proof(safe intro!:determI additiveI maximalI, simp_all add:wp_eval)
  fix c
  from da and db
  show "wp a (wp b (\<lambda>_. c)) = (\<lambda>_. c)"
    by(simp add:maximalD determ_maximalD)
next
  fix P::"'b \<Rightarrow> real" and Q::"'b \<Rightarrow> real" and s::'b
  assume sP: "sound P" and sQ: "sound Q"

  from sP and hb have sbP: "sound (wp b P)" by(blast)
  from sQ and hb have sbQ: "sound (wp b Q)" by(blast)

  from da and db and sP and sQ and sbP and sbQ
  show "wp a (wp b (\<lambda>s. P s + Q s)) s = wp a (wp b P) s + wp a (wp b Q) s"
    by(simp add:additiveD determ_additiveD)
qed

lemma determ_wp_repeat:
  assumes onestep: "determ (wp a)"
      and ha:      "healthy (wp a)"
  shows "determ (wp (repeat n a))" (is "?X n")
proof(induct n)
  show "?X 0" by(simp add:determ_wp_Skip)

  fix n assume "?X n"
  with assms show "?X (Suc n)"
    by(simp add:determ_wp_Seq healthy_intros)
qed

(* XXX - complete these *)
lemmas determ_intros =
  determ_wp_Skip determ_wp_Apply determ_wp_Seq determ_wp_repeat

definition
  wp_under_wlp :: "'s prog \<Rightarrow> bool"
where
  "wp_under_wlp prog \<equiv> \<forall>P. sound P \<longrightarrow> wp prog P \<tturnstile> wlp prog P"

lemma wp_under_wlpI[intro]:
  "\<lbrakk> \<And>P. sound P \<Longrightarrow> wp prog P \<tturnstile> wlp prog P \<rbrakk> \<Longrightarrow> wp_under_wlp prog"
  unfolding wp_under_wlp_def by(simp)

lemma wp_under_wlpD[dest]:
  "\<lbrakk> wp_under_wlp prog; sound P \<rbrakk> \<Longrightarrow> wp prog P \<tturnstile> wlp prog P"
  unfolding wp_under_wlp_def by(simp)

lemma wp_under_le_trans:
  "wp_under_wlp a \<Longrightarrow> le_trans (wp a) (wlp a)"
  by(blast)

lemma wp_under_wlp_Abort:
  "wp_under_wlp Abort"
  apply(rule wp_under_wlpI)
  unfolding wp_eval by(blast)

lemma wp_under_wlp_Skip:
  "wp_under_wlp Skip"
  apply(rule wp_under_wlpI)
  unfolding wp_eval by(blast)

lemma wp_under_wlp_Seq:
  assumes h_wp_a:  "healthy (wp a)"
      and h_wlp_a: "healthy (wlp a)"
      and h_wp_b:  "healthy (wp b)"
      and h_wlp_b: "healthy (wlp b)"
      and wp_u_a:  "wp_under_wlp a"
      and wp_u_b:  "wp_under_wlp b"
  shows "wp_under_wlp (a ;; b)"
proof(rule wp_under_wlpI, unfold wp_eval o_def)
  fix P::"'a \<Rightarrow> real" assume s_P: "sound P"
  with h_wp_b have "sound (wp b P)" by(blast)
  with wp_u_a have "wp a (wp b P) \<tturnstile> wlp a (wp b P)" by(blast)
  also {
    from wp_u_b and s_P have "wp b P \<tturnstile> wlp b P" by(blast)
    with h_wlp_a and h_wlp_b and h_wp_b and s_P
    have "wlp a (wp b P) \<tturnstile> wlp a (wlp b P)"
      by(blast intro:mono_transD[OF healthy_monoD, OF h_wlp_a])
  }
  finally show "wp a (wp b P) \<tturnstile> wlp a (wlp b P)" .
qed

lemma wp_under_wlp_PC:
  assumes h_wp_a:  "healthy (wp a)"
      and h_wlp_a: "healthy (wlp a)"
      and h_wp_b:  "healthy (wp b)"
      and h_wlp_b: "healthy (wlp b)"
      and wp_u_a:  "wp_under_wlp a"
      and wp_u_b:  "wp_under_wlp b"
      and s_P:     "sound P"
      and u_P:     "bounded_by 1 P"
  shows "wp_under_wlp (a \<^bsub>P\<^esub>\<oplus> b)"
proof(rule wp_under_wlpI, unfold wp_eval, rule le_funI)
  fix Q::"'a \<Rightarrow> real" and s
  assume s_Q: "sound Q"
  from u_P have "P s \<le> 1" by(blast)
  hence "0 \<le> 1 - P s" by(simp)
  moreover
  from s_Q and wp_u_b have "wp b Q s \<le> wlp b Q s" by(blast)
  ultimately
  have "(1 - P s) * wp b Q s \<le> (1 - P s) * wlp b Q s"
    by(blast intro:mult_left_mono)

  moreover {
    from s_Q and wp_u_a have "wp a Q s \<le> wlp a Q s" by(blast)
    with s_P have "P s * wp a Q s \<le> P s * wlp a Q s"
      by(blast intro:mult_left_mono)
  }

  ultimately
  show "P s * wp  a Q s + (1 - P s) * wp  b Q s \<le>
        P s * wlp a Q s + (1 - P s) * wlp b Q s"
    by(blast intro: add_mono)
qed

lemma wp_under_wlp_Embed:
  "wp_under_wlp (Embed t)"
  apply(rule wp_under_wlpI)
  unfolding wp_eval by(blast)

lemma wp_under_wlp_Mu:
  assumes h_wp_pr:  "\<And>t. healthy t \<Longrightarrow> healthy (wp (f (Embed t)))"
      and h_wlp_pr: "\<And>t. healthy t \<Longrightarrow> healthy (wlp (f (Embed t)))"
      and eq_pr: "\<And>t u. \<lbrakk>healthy t; healthy u; equiv_trans t u\<rbrakk> \<Longrightarrow>
                        equiv_trans (wp (f (Embed t))) (wp (f (Embed u)))"
      and wp_u_pr: "\<And>a. \<lbrakk> healthy (wp a); healthy (wlp a); wp_under_wlp a \<rbrakk> \<Longrightarrow> 
                        wp_under_wlp (f a)"
      and mwlp_tt: " mono_ttrans (\<lambda>t. wlp (f (Embed t)))"
  shows "wp_under_wlp (\<mu> t. f t)"
proof(rule wp_under_wlpI, unfold wp_eval)
  fix P::"'a \<Rightarrow> real"
  assume s_P: "sound P"
  have "le_trans (lfp_trans (\<lambda>t. wp (f (Embed t))))
                 (gfp_trans (\<lambda>t. wlp (f (Embed t))))"
    apply(rule lfp_trans_lowerbound)
       apply(rule gfp_trans_healthy)
      apply(rule h_wp_pr, assumption)
     apply(rule eq_pr, assumption+)
    apply(rule le_trans_trans)
     apply(rule wp_under_le_trans)
     apply(rule wp_u_pr)
       apply(rule healthy_wp_Embed)
       apply(rule gfp_trans_healthy)
      apply(rule healthy_wlp_Embed)
      apply(rule gfp_trans_healthy)
     apply(rule wp_under_wlp_Embed)
    apply(rule equiv_imp_le)
    apply(subst equiv_trans_comm)
    apply(rule gfp_trans_unfold)
     apply(rule mwlp_tt)
    apply(rule h_wlp_pr, assumption)
    done
  with s_P show "lfp_trans (\<lambda>t. wp (f (Embed t))) P \<tturnstile>
                 gfp_trans (\<lambda>t. wlp (f (Embed t))) P"
    by(blast intro: le_transD)
qed

lemmas wp_under_intros =
  wp_under_wlp_Abort wp_under_wlp_Skip wp_under_wlp_Seq
  wp_under_wlp_PC wp_under_wlp_Embed wp_under_wlp_Mu

definition
  well_def :: "'s prog \<Rightarrow> bool"
where
  "well_def prog \<equiv> healthy (wp prog) \<and> healthy (wlp prog)
                 \<and> wp_under_wlp prog"

lemma well_defI[intro]:
  "\<lbrakk> healthy (wp prog); healthy (wlp prog);
     wp_under_wlp prog \<rbrakk> \<Longrightarrow>
   well_def prog"
  unfolding well_def_def by(simp)

lemma well_def_wp_healthy[dest]:
  "well_def prog \<Longrightarrow> healthy (wp prog)"
  unfolding well_def_def by(simp)

lemma well_def_wlp_healthy[dest]:
  "well_def prog \<Longrightarrow> healthy (wlp prog)"
  unfolding well_def_def by(simp)

lemma well_def_wp_under[dest]:
  "well_def prog \<Longrightarrow> wp_under_wlp prog"
  unfolding well_def_def by(simp)

lemma well_def_Abort:
  "well_def Abort"
  by(blast intro:healthy_intros wp_under_intros)

lemma well_def_Skip:
  "well_def Skip"
  by(blast intro:healthy_intros wp_under_intros)

lemma well_def_Seq:
  "\<lbrakk> well_def a; well_def b \<rbrakk> \<Longrightarrow> well_def (a ;; b)"
  by(blast intro:healthy_intros wp_under_intros)

(* XXX - complete these *)
lemmas well_def_intros = well_def_Abort well_def_Skip well_def_Seq

lemma wp_under_inv_rel:
  "\<lbrakk> wp_under_wlp body; wp_inv G body I; sound I \<rbrakk> \<Longrightarrow> wlp_inv G body I"
  unfolding wp_inv_def wlp_inv_def
  by(blast intro:order_trans)

lemma wp_loop_unfold:
  assumes s: "sound P"
      and h: "healthy (wp body)"
  shows
  "wp (do G \<longrightarrow> body od) P =
   (\<lambda>s. \<guillemotleft>\<^boverline>G\<^eoverline>\<guillemotright> s * P s + \<guillemotleft>G\<guillemotright> s * wp body (wp (do G \<longrightarrow> body od) P) s)" 
  unfolding wp_eval
proof -
  have "equiv_trans
    (lfp_trans (\<lambda>t. wp (body ;; Embed t \<^bsub>\<guillemotleft> G \<guillemotright>\<^esub>\<oplus> Skip)))
    (wp (body ;;
         Embed (lfp_trans (\<lambda>t. wp (body ;;
                                   Embed t \<^bsub>\<guillemotleft> G \<guillemotright>\<^esub>\<oplus> Skip))) \<^bsub>\<guillemotleft> G \<guillemotright>\<^esub>\<oplus> Skip))"
    apply(rule lfp_trans_unfold)
     apply(force intro:mono_ttrans_intros healthy_intros h)
    apply(force intro:healthy_intros h)
    done
  also have "equiv_trans ...
    (\<lambda>P s. \<guillemotleft>\<^boverline>G\<^eoverline>\<guillemotright> s * P s + \<guillemotleft>G\<guillemotright> s * wp body (wp (do G \<longrightarrow> body od) P) s)"
    apply(rule equiv_transI)
    apply(simp add:wp_eval)
    apply(simp add:algebra_simps negate_embed)
    done

  finally show "lfp_trans (\<lambda>t. wp (body ;; Embed t \<^bsub>\<guillemotleft> G \<guillemotright>\<^esub>\<oplus> Skip)) P =
    (\<lambda>s. \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> s * P s +
         \<guillemotleft> G \<guillemotright> s *
         wp body (lfp_trans (\<lambda>t. wp (body ;; Embed t \<^bsub>\<guillemotleft> G \<guillemotright>\<^esub>\<oplus> Skip)) P) s)"
    using s unfolding wp_eval by(rule equiv_transD)
qed

lemma bounded_byD2[dest]:
  "bounded_by b f \<Longrightarrow> f \<tturnstile> (\<lambda>x. b)"
  unfolding bounded_by_def by(blast)

lemma healthy_pr_bot[simp]:
  "healthy t \<Longrightarrow> t (\<lambda>s. 0) = (\<lambda>s. 0)"
  by(blast intro!:antisym bounded_byD2 nnegD2)

lemma fold_guard: "\<forall>s. I s \<longrightarrow> G s \<Longrightarrow> \<guillemotleft>G\<guillemotright> s * \<guillemotleft>I\<guillemotright> s = \<guillemotleft>I\<guillemotright> s"
  by(case_tac "I s", simp_all)

lemma elim_guard: "\<forall>s. I s \<longrightarrow> G s \<Longrightarrow> \<guillemotleft>\<^boverline>G\<^eoverline>\<guillemotright> s * \<guillemotleft>I\<guillemotright> s = 0"
  by(case_tac "I s", simp_all)

lemma I_loop_fp:
  assumes A: "wlp_inv G body \<guillemotleft>I\<guillemotright>"
      and B: "\<forall>s. I s \<longrightarrow> G s"
      and h: "healthy (wlp body)"
  shows "\<guillemotleft>I\<guillemotright> \<tturnstile> wlp (do G \<longrightarrow> body od) \<guillemotleft>I\<guillemotright>"
  using h apply(subst wlp_Loop1, blast, simp)
  apply(simp add:elim_guard[OF B])
  apply(rule gfp_in_upperbound)
    apply(rule weakly_bounded_bounded_above, blast)
   apply(rule weakly_bounded_self, blast)
  apply(rule le_funI)
  apply(subst fold_guard[OF B, symmetric])
  apply(rule mult_left_mono)
   using A unfolding wlp_inv_def apply(simp add:fold_guard[OF B])
  apply(simp)
  done

lemma lift_Sup_empty:
  "\<lbrakk> S = {}; sound P \<rbrakk> \<Longrightarrow> lift_Sup S P = (\<lambda>s. 0)"
  unfolding lift_Sup_def
  apply(simp add:bot_trans_def)
  apply(rule_tac P=P in equiv_transD, simp_all)
  done

(* XXX - this should be promoted as the canonical proof *)
lemma lub_fun_healthy:
  fixes S::"(('a \<Rightarrow> real) \<Rightarrow> 'a \<Rightarrow> real) set"
  assumes ne: "S \<noteq> {}"
      and h:  "\<forall>t\<in>S. healthy t"
  shows "healthy (lub S)"
proof(safe intro!:healthyI feasibleI mono_transI scalingI bounded_byI
                  nnegI le_funI)
  fix b::real and P::"'a \<Rightarrow> real" and s
  assume bby: "bounded_by b P" and nneg: "nneg P"

  from ne obtain t where tin: "t \<in> S" by(blast)

  from h and bby and nneg have top: "\<forall>t \<in> S. t P \<le> (\<lambda>s. b)" by(blast)
  with tin have "lub S P \<le> (\<lambda>s. b)"
    unfolding lub_fun_def by(blast intro:lub_least)
  thus "lub S P s \<le> b" by(rule le_funD)

  show "0 \<le> lub S P s"
  proof -
    from h and bby and nneg and tin have "0 \<le> t P s" by(blast)
    also {
      from tin and top have "t P \<le> lub S P"
        unfolding lub_fun_def by(blast intro: lub_greater)
      hence "t P s \<le> lub S P s" by(rule le_funD)
    }
    finally show ?thesis .
  qed

next
  fix P::"'a \<Rightarrow> real" and Q and s
  assume sP: "sound P" and sQ: "sound Q" and le: "P \<le> Q"

  from ne obtain t where tin: "t \<in> S" by(blast)

  from sQ and h have top: "\<forall>a\<in>{t Q |t. t \<in> S}. a \<le> (\<lambda>s. bound_of Q)"
    by(blast intro:bounded_byD2)

  show "lub S P s \<le> lub S Q s"
    unfolding lub_fun_def
  proof(rule lub_least, blast intro:tin, safe)
    fix t' assume sin: "t' \<in> S"

    from sP and sQ and h and le and sin have "t' P s \<le> t' Q s"
      by(blast)
    also from top and sin have "... \<le> lub {f Q s |f. f \<in> S}"
      by(blast intro:lub_greater)
    also have "{f Q s |f. f \<in> S} = {f s |f. f \<in> {f Q |f. f \<in> S}}"
      by(blast)
    finally show "t' P s \<le> lub ..." .
  qed

next
  fix P::"'a \<Rightarrow> real" and c::real and s
  assume sP: "sound P" and pos: "0 < c"

  from ne obtain t where tin: "t \<in> S" by(blast)

  show "c * lub S P s = lub S (\<lambda>x. c * P x) s"
  proof -
    from h and sP have top1: "\<forall>t \<in> S. t P \<le> (\<lambda>s. bound_of P)"
      by(blast intro!:bounded_byD2)
    from h and sP and pos
    have top2: "\<And>s. \<forall>x \<in> {c * t P s |t. t \<in> S}. x \<le> c * bound_of P"
      by(blast intro!:mult_left_mono dest:less_imp_le)
    from pos have nneg: "0 \<le> c" by(rule less_imp_le)
    from pos have nz:   "c \<noteq> 0" by(blast)
    from nneg have inneg: "0 \<le> inverse c" by(simp)
    from nz have inz: "inverse c \<noteq> 0" by(simp)
    note c_facts = pos nneg nz inneg inz
    have "\<And>s. c * lub S P s = lub {c * t P s |t. t \<in> S}"
      (is "\<And>s. c * lub S P s = lub (?Y s)")
    proof (rule antisym)
      from c_facts have "\<And>t s. inverse c * (c * t P s) = t P s" by(simp)
      from top2 have "\<And>t s. t \<in> S \<Longrightarrow> c * t P s \<le> lub (?Y s)"
        by(blast intro:lub_greater)
      hence "\<And>t s. t \<in> S \<Longrightarrow> c * t P s \<le> lub (?Y s)" by(blast)
      with c_facts have "\<And>t s. t \<in> S \<Longrightarrow> t P s \<le> inverse c * lub (?Y s)"
        by(simp add:mult_div_mono_right)
      hence "\<And>t. t \<in> S \<Longrightarrow> t P \<le> (\<lambda>s. inverse c * lub (?Y s))"
        by(rule le_funI)
      with tin have "lub S P \<le> (\<lambda>s. inverse c * lub (?Y s))"
        unfolding lub_fun_def
        apply -
        apply(rule le_funI[OF lub_least], blast+)
        done
      hence "\<And>s. lub S P s \<le> inverse c * lub (?Y s)"
        by(rule le_funD)
      with c_facts show "\<And>s. c * lub S P s \<le> lub (?Y s)"
        by(simp add:mult_div_mono_left)

      from tin and c_facts and top1
      show "\<And>s. lub (?Y s) \<le> c * lub S P s"
        unfolding lub_fun_def
        apply -
        apply(rule lub_least)
         apply(blast)
        apply(clarify)
        apply(rule mult_left_mono) prefer 2 apply(blast)
        apply(rule lub_greater)
         apply(blast)
        apply(rule_tac x="bound_of P" in exI)
        apply(clarify)
        apply(rule_tac x=s in le_funD)
        apply(blast)
        done (* Yuck.  Why did blast stop working? *)
    qed

    moreover
    have "{c * t P s |t. t \<in> S} = {f s |f. \<exists>fa. f = fa (\<lambda>x. c * P x) \<and> fa \<in> S}"
    proof(rule set_eqI, simp, safe)
      fix t assume "t \<in> S"
      moreover {
        with h have healthy: "healthy t" by(simp)
        hence "scaling t" by(blast)
        with sP and pos have "c * t P s = t (\<lambda>s. c * P s) s"
          by(simp add:scalingD)
      }
      ultimately
      show "\<exists>f. c * t P s = f s \<and> (\<exists>fa. f = fa (\<lambda>x. c * P x) \<and> fa \<in> S)"
        by(blast)
    next
      fix t assume "t \<in> S"
      moreover {
        with h have healthy: "healthy t" by(simp)
        hence "scaling t" by(blast)
        with sP and pos have "t (\<lambda>s. c * P s) s = c * t P s"
          by(simp add:scalingD)
      }
      ultimately
      show "\<exists>t'. t (\<lambda>x. c * P x) s = c * t' P s \<and> t' \<in> S"
        by(blast)
    qed

    ultimately show ?thesis unfolding lub_fun_def by(simp)
  qed
qed

lemma lift_Sup_point:
  "\<lbrakk> S \<noteq> {}; \<forall>t\<in>S. healthy t; sound P \<rbrakk> \<Longrightarrow> lift_Sup S P = lub {t P |t. t \<in> S}"
  apply(rule antisym)
   prefer 2
   apply(rule lub_least)
    apply(blast)
   apply(clarify)
   apply(rule_tac P=P in le_transD, simp_all)[]
   apply(blast intro:lift_Sup_upper)
  apply(rule_tac P=P in le_transD, simp_all)[]
  apply(rule lift_Sup_least)
    apply(fold lub_fun_def)
    apply(rule lub_fun_healthy, assumption+)
   apply(blast)
  apply(rule le_transI)
  unfolding lub_fun_def
  apply(rule le_funI[OF lub_greater])
   apply(blast)
  apply(rule_tac x="bound_of Pa" in exI)
  apply(clarify)
  apply(blast intro!:bounded_byD)
  done

text {* If the invariant implies the guard, the loop does not
  terminate.  This is the case in which our loop rule fails, and so it
  is proved separately. *}
lemma implies_guard_nt:
  assumes A: "wlp_inv G body \<guillemotleft>I\<guillemotright>"
      and B: "\<forall>s. I s \<longrightarrow> G s"
      and C: "wp_inv G body \<guillemotleft>I\<guillemotright>"
      and h: "healthy (wlp body)"
      and h2: "healthy (wp body)"
      and sdp: "sub_distrib_pconj (do G \<longrightarrow> body od)"
  shows "\<guillemotleft>I\<guillemotright> && wp (do G \<longrightarrow> body od) (\<lambda>s. 1) = (\<lambda>s. 0)"
  apply(rule antisym)
   apply(rule order_trans)
    apply(rule exp_conj_mono_left)
    apply(rule I_loop_fp)
      apply(rule A)
     apply(rule B)
    apply(rule h)
   apply(rule order_trans)
    apply(rule sub_distrib_pconjD[OF sdp], force+)[]
   apply(subst exp_conj_1_right, blast)
   apply(simp add:wp_eval)
   apply(rule lfp_trans_ordinal_induct)
       apply(force intro:mono_ttrans_intros healthy_intros h2)
      apply(force intro:mono_ttrans_intros healthy_intros h2)
     apply(subgoal_tac "S \<guillemotleft>I\<guillemotright> = (\<lambda>s. 0)")
      prefer 2
      apply(rule antisym, assumption)
      apply(force)
     apply(simp add:wp_eval)
     using h2 apply(simp add:algebra_simps fold_guard[OF B])
    apply(case_tac "M={}")
     apply(subst lift_Sup_empty)
       apply(assumption)
      apply(force)
     apply(simp)
    apply(subst lift_Sup_point)
       apply(assumption)
      apply(assumption)
     apply(force)
    apply(rule lub_least)
     apply(blast)
    apply(blast)
   apply(rule order_trans) prefer 2 apply assumption
   apply(force intro:le_transD)
   apply(rule exp_conj_nneg)
   done

text {* The first total correctness lemma for loops which terminate with
        probability 1 (Lemma 7.3.1, p186). *}
lemma wp_Loop1:
  assumes hwp:   "healthy (wp body)"
      and hwlp:  "healthy (wlp body)"
      and sdp:   "sub_distrib_pconj (do G \<longrightarrow> body od)"
      and inv:   "wlp_inv G body I"
      and sound: "sound I"
      and unit:  "bounded_by 1 I"
      (* XXX - this really needs a better name. *)
      and magic: "I \<in> weakly_bounded_by (\<lambda>s. \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> s * I s)"
  shows "I && wp (do G \<longrightarrow> body od) (\<lambda>s. 1) \<tturnstile>
         wp (do G \<longrightarrow> body od) (\<lambda>s. \<guillemotleft>\<^boverline>G\<^eoverline>\<guillemotright> s * I s)"
  (is "I && ?T \<tturnstile> wp ?loop ?X")
proof -
  from assms have "I && ?T \<tturnstile> wlp ?loop ?X && ?T"
    by(blast intro:exp_conj_mono_left wlp_Loop)

  also from hwp and hwlp have "... \<tturnstile> wp ?loop (?X && (\<lambda>s. 1))"
    apply -
    apply(rule sub_distrib_pconjD[OF sdp])
      apply(rule mult_sound)
       apply(force)
      using sound apply(force)
     apply(rule bounded_byI)
     apply(rule mult_mono[where b="1::real" and d=1, simplified])
       apply(force)
      using unit apply(blast)
     apply(rule_tac x=x in nnegD)
     using sound apply(blast dest!:sound_nneg)
    apply(force)
    done

  also from sound have "... = wp ?loop ?X"
    unfolding exp_conj_def pconj_def
      (* Shouldn't unfold, need more lemmata *)
    apply(subst tminus_cancel_1)
     apply(blast intro:mult_nonneg_nonneg)
    apply(simp)
    done

  finally show ?thesis .
qed

text {* The above, specialized for standard expectations. *}
lemma wp_Loop_std:
  assumes wd:     "well_def body"
      and sdp:    "sub_distrib_pconj (do G \<longrightarrow> body od)"
      and wp_inv: "wp_inv G body \<guillemotleft>I\<guillemotright>"
  shows "\<guillemotleft>I\<guillemotright> && wp (do G \<longrightarrow> body od) (\<lambda>s. 1) \<tturnstile>
         wp (do G \<longrightarrow> body od) (\<lambda>s. \<guillemotleft>\<^boverline>G\<^eoverline>\<guillemotright> s * \<guillemotleft>I\<guillemotright> s)"
proof(cases "\<forall>s. I s \<longrightarrow> G s")
  from wd have h_wp: "healthy (wp body)" by(blast)
  from wd have h_wlp: "healthy (wlp body)" by(blast)
  from wd and wp_inv have wlp_inv: "wlp_inv G body \<guillemotleft>I\<guillemotright>"
    by(force intro:wp_under_inv_rel)
  case True with assms and h_wp and h_wlp and wlp_inv show ?thesis
    apply(simp add:implies_guard_nt)
    apply(rule nnegD2)
    apply(rule_tac t="wp (\<mu> x. body ;; x \<^bsub>\<guillemotleft> G \<guillemotright>\<^esub>\<oplus> Skip)" in healthy_nnegD)
     apply(blast intro:healthy_intros)
    apply(simp add:sound_intros)
    done
next
  from wd have h_wp: "healthy (wp body)" by(blast)
  from wd have h_wlp: "healthy (wlp body)" by(blast)
  from wd and wp_inv have wlp_inv: "wlp_inv G body \<guillemotleft>I\<guillemotright>"
    by(force intro:wp_under_inv_rel)
  case False
  hence "\<guillemotleft>I\<guillemotright> \<in> weakly_bounded_by (\<lambda>s. \<guillemotleft> \<^boverline>G\<^eoverline> \<guillemotright> s * \<guillemotleft>I\<guillemotright> s)"
    by(rule nentails_weakly_bounded)
  with sdp and wp_inv and h_wp and h_wlp and wlp_inv show ?thesis
    apply -
    apply(rule wp_Loop1, assumption+, force, force, assumption)
    done
qed

section {* Program Refinement *}

definition
  refines :: "'s prog \<Rightarrow> 's prog \<Rightarrow> bool" (infix "\<sqsubseteq>" 70)
where
  "prog \<sqsubseteq> prog' \<equiv> \<forall>P. sound P \<longrightarrow> wp prog P \<tturnstile> wp prog' P"

lemma refinesI[intro]:
  "\<lbrakk> \<And>P. sound P \<Longrightarrow> wp prog P \<tturnstile> wp prog' P \<rbrakk> \<Longrightarrow> prog \<sqsubseteq> prog'"
  unfolding refines_def by(simp)

lemma refinesD[dest]:
  "\<lbrakk> prog \<sqsubseteq> prog'; sound P \<rbrakk> \<Longrightarrow> wp prog P \<tturnstile> wp prog' P"
  unfolding refines_def by(simp)

section {* Algebra *}

subsection {* Laws following from the basic arithmetic of the
  operators seperately *}

lemma DC_comm[ac_simps]:
  "a \<Sqinter> b = b \<Sqinter> a"
  unfolding DC_def by(simp add:ac_simps)

lemma DC_assoc[ac_simps]:
  "a \<Sqinter> (b \<Sqinter> c) = (a \<Sqinter> b) \<Sqinter> c"
  unfolding DC_def by(simp add:ac_simps)

lemma DC_idem:
  "a \<Sqinter> a = a"
  unfolding DC_def by(simp)

lemma PC_quasi_comm:
  "a \<^bsub>p\<^esub>\<oplus> b = b \<^bsub>(\<lambda>s. 1 - p s)\<^esub>\<oplus> a"
  unfolding PC_def by(simp add:algebra_simps)

lemma PC_idem:
  "a \<^bsub>p\<^esub>\<oplus> a = a"
  unfolding PC_def by(simp add:algebra_simps)

lemma Seq_assoc[ac_simps]:
  "A ;; (B ;; C) = A ;; B ;; C"
  by(simp add:Seq_def o_def)

subsection {* Laws relating demonic choice and refinement *}

lemma left_refines_DC:
  "(a \<Sqinter> b) \<sqsubseteq> a"
  by(force intro:min_max.inf_le1 simp:wp_eval)

lemma right_refines_DC:
  "(a \<Sqinter> b) \<sqsubseteq> b"
  by(subst DC_comm, rule left_refines_DC)

lemma DC_refines:
  "\<lbrakk> a \<sqsubseteq> b; a \<sqsubseteq> c \<rbrakk> \<Longrightarrow> a \<sqsubseteq> (b \<Sqinter> c)"
  apply(rule refinesI)
  apply(simp add:wp_eval)
  apply(drule refinesD, assumption)+
  apply(blast intro:min_max.le_infI)
  done

subsection {* Laws depending on the arithmetic of \<^bsub>p\<^esub>\<oplus> and \<Sqinter> together *}

lemma PC_refines_DC:
  assumes nneg: "nneg p"
      and unit: "bounded_by 1 p"
  shows "(a \<Sqinter> b) \<sqsubseteq> (a \<^bsub>p\<^esub>\<oplus> b)"
proof(rule refinesI, unfold wp_eval, rule le_funI)
  fix s and P::"'a \<Rightarrow> real" assume sound: "sound P"
  from nneg have nn_p: "0 \<le> p s" by(blast)
  from unit have "p s \<le> 1" by(blast)
  hence nn_np: "0 \<le> 1 - p s" by(simp)
  show "min (wp a P s) (wp b P s) \<le> p s * wp a P s + (1 - p s) * wp b P s"
  proof(cases "wp a P s \<le> wp b P s",
      simp_all add:min_max.inf_absorb1 min_max.inf_absorb2)
    case True note le = this
    have "wp a P s = (p s + (1 - p s)) * wp a P s" by(simp)
    also have "... = p s * wp a P s + (1 - p s) * wp a P s"
      by(simp only: distrib_right)
    also {
      from le and nn_np have "(1 - p s) * wp a P s \<le> (1 - p s) * wp b P s"
        by(rule mult_left_mono)
      hence "p s * wp a P s + (1 - p s) * wp a P s \<le>
        p s * wp a P s + (1 - p s) * wp b P s"
        by(rule add_left_mono)
    }
    finally show "wp a P s \<le> p s * wp a P s + (1 - p s) * wp b P s" .
  next
    case False
    then have le: "wp b P s \<le> wp a P s" by(simp)
    have "wp b P s = (p s + (1 - p s)) * wp b P s" by(simp)
    also have "... = p s * wp b P s + (1 - p s) * wp b P s"
      by(simp only: distrib_right)
    also {
      from le and nn_p have "p s * wp b P s \<le> p s * wp a P s"
        by(rule mult_left_mono)
      hence "p s * wp b P s + (1 - p s) * wp b P s \<le>
        p s * wp a P s + (1 - p s) * wp b P s"
        by(rule add_right_mono)
    }
    finally show "wp b P s \<le> p s * wp a P s + (1 - p s) * wp b P s" .
  qed
qed

subsection {* Transitivity of Refinement *}

lemma pr_trans[trans]:
  fixes A::"'a prog"
  assumes prAB: "A \<sqsubseteq> B"
      and prBC: "B \<sqsubseteq> C"
  shows "A \<sqsubseteq> C"
proof
  fix P::"'a \<Rightarrow> real" assume sP: "sound P"
  with prAB have "wp A P \<tturnstile> wp B P" by(blast)
  also from sP and prBC have "... \<tturnstile> wp C P" by(blast)
  finally show "wp A P \<tturnstile> ..." .
qed

section {* Modular Reasoning *}

subsection {* The expectation-entailment form *}

lemma wp_combine:
  assumes wp1: "P \<tturnstile> wp A R"
      and wp2: "Q \<tturnstile> wp A S"
      and sc:  "sub_conj (wp A)"
      and sR:  "sound R"
      and sS:  "sound S"
      and bR:  "bounded_by 1 R"
      and bS:  "bounded_by 1 S"
  shows "P && Q \<tturnstile> wp A (R && S)"
proof -
  from wp1 and wp2 have "P && Q \<tturnstile> wp A R && wp A S"
    by(blast intro:entails_frame)
  also from sc and sR and sS and bR and bS have "... \<le> wp A (R&&S)"
    by(rule sub_conjD)
  finally show ?thesis .
qed

text {* These allow mismatched results to be composed *}

lemma wp_strengthen_post:
  "\<lbrakk>P \<tturnstile> wp pr Q; healthy (wp pr); sound R; Q \<tturnstile> R; sound Q\<rbrakk> \<Longrightarrow> P \<tturnstile> wp pr R"
  apply(rule entailsI)
  apply(rule_tac y="wp pr Q s" in order_trans, blast)
  apply(rule_tac t="wp pr" in le_funD[OF mono_transD, standard], simp_all)
  apply(blast intro:healthy_monoD)
  done

lemma wp_weaken_pre:
  "\<lbrakk>Q \<tturnstile> wp pr R; healthy (wp pr); sound R; P \<tturnstile> Q; sound Q\<rbrakk> \<Longrightarrow> P \<tturnstile> wp pr R"
  by(blast intro:entails_trans)

lemma wlp_strengthen_post:
  "\<lbrakk>P \<tturnstile> wlp pr Q; healthy (wlp pr); sound R; Q \<tturnstile> R; sound Q\<rbrakk> \<Longrightarrow> P \<tturnstile> wlp pr R"
  apply(rule entailsI)
  apply(rule_tac y="wlp pr Q s" in order_trans, blast)
  apply(rule_tac t="wlp pr" in le_funD[OF mono_transD, standard], simp_all)
  apply(blast intro:healthy_monoD)
  done

lemma wlp_weaken_pre:
  "\<lbrakk> Q \<tturnstile> wlp pr R; P \<tturnstile> Q \<rbrakk> \<Longrightarrow> P \<tturnstile> wlp pr R"
  by(blast intro:entails_trans)

text {* This rule is unique to pGCL.  Use it to scale the post-expectation
        of a rule to 'fit under' the precondition you need to satisfy. *}
lemma wp_scale:
  "\<lbrakk> P \<tturnstile> wp prog Q; healthy (wp prog); sound Q; 0 < c \<rbrakk> \<Longrightarrow>
   (\<lambda>s. c * P s) \<tturnstile> wp prog (\<lambda>s. c * Q s)"
  by(rule le_funI,
     subst scalingD[where t="wp prog", symmetric],
     (blast intro:mult_left_mono less_imp_le)+)

lemma wlp_scale:
  "\<lbrakk> P \<tturnstile> wlp prog Q; healthy (wlp prog); sound Q; 0 < c \<rbrakk> \<Longrightarrow>
   (\<lambda>s. c * P s) \<tturnstile> wlp prog (\<lambda>s. c * Q s)"
  by(rule le_funI,
     subst scalingD[where t="wlp prog", symmetric],
     (blast intro:mult_left_mono less_imp_le)+)

text {* The remainder are structural rules *}

lemma wp_Seq:
  assumes ent_a: "P \<tturnstile> wp a Q"
      and ent_b: "Q \<tturnstile> wp b R"
      and h_a:   "healthy (wp a)"
      and h_b:   "healthy (wp b)"
      and s_Q:   "sound Q"
      and s_R:   "sound R"
  shows "P \<tturnstile> wp (a ;; b) R"
proof -
  note ent_a
  also from assms have "wp a Q \<tturnstile> wp a (wp b R)"
    by(blast intro:healthy_monoD2)
  finally show ?thesis by(simp add:wp_eval)
qed

lemma wlp_Seq:
  assumes ent_a: "P \<tturnstile> wlp a Q"
      and ent_b: "Q \<tturnstile> wlp b R"
      and h_a:   "healthy (wlp a)"
      and h_b:   "healthy (wlp b)"
      and s_Q:   "sound Q"
      and s_R:   "sound R"
  shows "P \<tturnstile> wlp (a ;; b) R"
proof -
  note ent_a
  also from assms have "wlp a Q \<tturnstile> wlp a (wlp b R)"
    by(blast intro:healthy_monoD2)
  finally show ?thesis by(simp add:wp_eval)
qed

lemma wp_Skip:
  "P \<tturnstile> wp Skip P"
  unfolding wp_eval by(blast)

lemma wlp_Skip:
  "P \<tturnstile> wlp Skip P"
  unfolding wp_eval by(blast)

lemma wp_DC:
  "\<lbrakk> P \<tturnstile> wp prog Q; P \<tturnstile> wp prog' Q \<rbrakk> \<Longrightarrow> P \<tturnstile> wp (prog \<Sqinter> prog') Q"
  unfolding wp_eval by(blast intro:min_max.inf_greatest)

lemma wlp_DC:
  "\<lbrakk> P \<tturnstile> wlp prog Q; P \<tturnstile> wlp prog' Q \<rbrakk> \<Longrightarrow> P \<tturnstile> wlp (prog \<Sqinter> prog') Q"
  unfolding wp_eval by(blast intro:min_max.inf_greatest)

lemma wp_Apply:
  "P \<tturnstile> Q o f \<Longrightarrow> P \<tturnstile> wp (Apply f) Q"
  unfolding wp_eval by(assumption)

lemma wlp_Apply:
  "P \<tturnstile> Q o f \<Longrightarrow> P \<tturnstile> wlp (Apply f) Q"
  unfolding wp_eval by(assumption)

lemma wp_SetPC:
  fixes f::"'x \<Rightarrow> 'y prog"
    and S::"'y \<Rightarrow> 'x set"
    and p::"'y \<Rightarrow> 'x \<Rightarrow> real"
  assumes rec: "\<And>x s. x \<in> S s \<Longrightarrow> P x \<tturnstile> wp (f x) Q"
      and nnp: "\<And>s. nneg (p s)"
      and sQ:  "\<And>x s. x \<in> S s \<Longrightarrow> sound (P x)"
      and sQ:  "sound Q"
  shows "(\<lambda>s. \<Sum>x \<in> S s. p s x * P x s) \<tturnstile> wp (SetPC f S p) Q"
  unfolding wp_eval pGCL.K_def
proof(rule le_funI)
  fix s
  from rec have "\<And>x. x \<in> S s \<Longrightarrow> P x s \<le> wp (f x) Q s" by(blast)
  moreover from nnp have "\<And>x. 0 \<le> p s x" by(blast)
  ultimately have "\<And>x. x \<in> S s \<Longrightarrow> p s x * P x s \<le> p s x * wp (f x) Q s"
    by(blast intro:mult_left_mono)
  thus "(\<Sum>x \<in> S s. p s x * P x s) \<le> (\<Sum>x \<in> S s. p s x * wp (f x) Q s)"
    by(rule setsum_mono)
qed

(* XXX - complete these *)

subsection {* Standard expectations *}

text {* Use a standard pre-expectation as an assumption. *}
lemma use_premise:
  "\<lbrakk> healthy t; \<And>s. P s \<Longrightarrow> 1 \<le> t \<guillemotleft>Q\<guillemotright> s \<rbrakk> \<Longrightarrow> \<guillemotleft>P\<guillemotright> \<tturnstile> t \<guillemotleft>Q\<guillemotright>"
  apply(rule entailsI)
  apply(case_tac "P s", simp_all)
  apply(rule_tac x=s in nnegD)
  apply(blast)
  done

lemma wp_conj_post:
  "\<lbrakk> P \<tturnstile> wp pr \<guillemotleft>\<lambda>s. Q s \<and> R s\<guillemotright>; healthy (wp pr) \<rbrakk> \<Longrightarrow> P \<tturnstile> wp pr \<guillemotleft>Q\<guillemotright>"
  by(blast intro:wp_strengthen_post implies_entails)

lemma wlp_conj_post:
  "\<lbrakk> P \<tturnstile> wlp pr \<guillemotleft>\<lambda>s. Q s \<and> R s\<guillemotright>; healthy (wlp pr) \<rbrakk> \<Longrightarrow> P \<tturnstile> wlp pr \<guillemotleft>Q\<guillemotright>"
  by(blast intro:wlp_strengthen_post implies_entails)

text {* Similar to @{thm use_premise}, but more general. *}
lemma entails_pconj_assumption:
  "\<lbrakk> feasible t; \<And>s. P s \<Longrightarrow> Q s \<le> t R s;
     bounded_by 1 Q; nneg Q;
     bounded_by 1 R; nneg R \<rbrakk> \<Longrightarrow> \<guillemotleft>P\<guillemotright> && Q \<tturnstile> t R"
  apply(rule entailsI)
  unfolding exp_conj_def
  apply(case_tac "P s", simp_all add:nnegD bounded_byD)
  apply(rule_tac x=s in nnegD)
  apply(rule_tac t="t" and b=1 in feasible_nnegD, simp_all)
  done

subsection {* Refinement *}

lemma wp_refines:
  assumes refines: "a \<sqsubseteq> b"
      and wp_a:    "P \<tturnstile> wp a Q"
      and sound:   "sound Q"
  shows "P \<tturnstile> wp b Q"
proof -
  note wp_a
  also from refines and sound have "wp a Q \<tturnstile> wp b Q"
    unfolding refines_def by(simp)
  finally show ?thesis .
qed

subsection {* Hoare triples *}

text {* The Hoare triple, or validity predicate, is logically equivalent
        to the weakest-precondition entailment form.  The benefit
        is that it allows us to define transitivity rules for
        computational (also/finally) reasoning. *}

definition
  wp_valid :: "('a \<Rightarrow> real) \<Rightarrow> 'a prog \<Rightarrow> ('a \<Rightarrow> real) \<Rightarrow> bool" ("\<lbrace>_\<rbrace> _ \<lbrace>_\<rbrace>p")
where
  "wp_valid P prog Q \<equiv> P \<tturnstile> wp prog Q"

lemma wp_validI:
  "P \<tturnstile> wp prog Q \<Longrightarrow> \<lbrace>P\<rbrace> prog \<lbrace>Q\<rbrace>p"
  unfolding wp_valid_def by(assumption)

lemma wp_validD:
  "\<lbrace>P\<rbrace> prog \<lbrace>Q\<rbrace>p \<Longrightarrow> P \<tturnstile> wp prog Q"
  unfolding wp_valid_def by(assumption)

lemma valid_Seq:
  "\<lbrakk> \<lbrace>P\<rbrace> a \<lbrace>Q\<rbrace>p; \<lbrace>Q\<rbrace> b \<lbrace>R\<rbrace>p;
     healthy (wp a); healthy (wp b); sound Q; sound R \<rbrakk> \<Longrightarrow>
  \<lbrace>P\<rbrace> a ;; b \<lbrace>R\<rbrace>p"
  unfolding wp_valid_def by(rule wp_Seq)

text {* Useful for reasoning with also-finally *}
declare valid_Seq[trans]

end
