(*
 * Copyright (C) 2013 National ICT Australia Ltd ('NICTA')
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Author: David Cock - David.Cock@nicta.com.au *)

(*
  Nondeterministic state and error monads with failure in pGCL
*)

header "Nondeterministic State Monad with Failure in pGCL"

theory pGCL_Monad imports
  pGCL pVCG PCorres
  "../Contrib/NonDetMonad"
  "../Contrib/Corres_UL"
begin

text {*
  The following is an embedding of the nondeterministic state monad
  used in the seL4 specification into pGCL.
*}

section "Running a Monad"

text {*
  A monad is a state transformer, which maps very simply into pGCL.
  The complication is the treatment of failure, divergence
  (nontermination) and nondeterminism.  For consistency with our
  pGCL semantics, we apply a demonic interpretation, giving the
  least possible result over all choices.  This then forces us to
  define `stuck' (no successor states) as success, for consistency.
  Finally, explicit failure is mapped to @{const Abort}, which in turn
  gives either success or failure, under wlp and wp semantics,
  respectively.
*}

definition Exec :: "('s, 'a) nondet_monad \<Rightarrow> 's prog"
where "Exec M \<equiv> \<lambda>ab R s.
  let (SA,f) = M s in                           (* Run the monad *)
  if f then Abort ab R s                        (* Fail is Abort *)
       else if SA = {} then (bound_of R)        (* Stuck is Success *)
                       else let S = snd ` SA in (* Ignore result *)
                            glb (R ` S)         (* Infimum over states *)"

subsection "Healthiness"

text {*
  As a consequence of the shallow embedding, we must show healthiness
  explicitly for the above definition, which is non-trivial.
*}

lemma healthy_Exec:
  "healthy ((Exec prog) ab)" (is "healthy ?T")
proof
  show "feasible ?T"
  proof
    fix b::real and P::"'a \<Rightarrow> real"
    assume bby: "bounded_by b P" and nneg: "nneg P"

    from bby and nneg have "\<And>x. Abort ab P x \<le> b"
      unfolding Abort_def by(case_tac ab, simp_all, blast+)
    moreover from bby have "bound_of P \<le> b" by(blast)
    moreover have "\<And>S. S \<noteq> {} \<Longrightarrow> glb (P ` S) \<le> b" proof -
      fix S::"'a set" assume "S \<noteq> {}"
      from this obtain s where "s \<in> S" by(blast)
      with nneg have "glb (P ` S) \<le> P s"
        by(blast intro:glb_lower)
      also from bby have "... \<le> b" by(blast)
      finally show "glb (P ` S) \<le> b" .
    qed
    ultimately show "bounded_by b (Exec prog ab P)"
      unfolding Exec_def wp_def Let_def split_def
      by(simp add:bounded_byI)

    from bby and nneg have "\<And>x. 0 \<le> Abort ab P x"
      unfolding Abort_def by(case_tac ab, simp_all, blast)
    moreover from nneg and bby have "0 \<le> bound_of P" by(blast)
    moreover from nneg have "\<And>S. S \<noteq> {} \<Longrightarrow> 0 \<le> glb (P ` S)"
      by(blast intro:glb_greatest)
    ultimately show "nneg (Exec prog ab P)"
      unfolding Exec_def wp_def Let_def split_def
      by(simp add:nnegI)
  qed

  show "mono_trans ?T"
  proof(rule mono_transI)
    fix P::"'a \<Rightarrow> real" and Q
    assume sP: "sound P" and sQ: "sound Q" and le: "P \<tturnstile> Q"
    hence "bound_of P \<le> bound_of Q" by(blast)
    moreover from sP and sQ and le
    have "\<And>x. Abort ab P x \<le> Abort ab Q x"
      unfolding Abort_def by(case_tac ab, simp_all, blast)
    moreover have "\<And>S. S \<noteq> {} \<Longrightarrow> glb (P ` S) \<le> glb (Q ` S)"
    proof -
      have lb: "\<And>S. \<forall>a\<in>Q ` S. glb (P ` S) \<le> a" proof
        fix a and S assume "a \<in> Q ` S"
        from this obtain s
          where sin: "s \<in> S" and sim: "a = Q s" by(blast)
        from sP and sin have "glb (P ` S) \<le> P s"
          by(blast intro:glb_lower)
        also from sim and le have "... \<le> a" by(blast)
        finally show "glb (P ` S) \<le> a" .
      qed

      fix S::"'a set" assume "S \<noteq> {}"
      from this obtain s where sin: "s \<in> S" by(blast)

      with lb show "glb (P ` S) \<le> glb (Q ` S)"
        by(blast intro:glb_greatest)
    qed
    ultimately show "?T P \<tturnstile> ?T Q"
      unfolding Exec_def wp_def Let_def split_def
      by(simp add:le_funI)
  qed

  show "scaling ?T"
  proof
    fix P::"'a \<Rightarrow> real" and c::real and s
    assume sound: "sound P" and pos: "0 < c"
    hence "c * Abort ab P s = Abort ab (\<lambda>s. c * P s) s"
      unfolding Abort_def by(case_tac ab, simp_all)
    moreover from pos and sound
    have "c * bound_of P = bound_of (\<lambda>s. c * P s)" by(simp)
    moreover from pos and sound
    have "\<And>S. S \<noteq> {} \<Longrightarrow> c * glb (P ` S) = glb ((\<lambda>s. c * P s) ` S)"
    proof -
      fix S::"'a set" assume sne: "S \<noteq> {}"
      show "c * glb (P ` S) = glb ((\<lambda>s. c * P s) ` S)" (is "?X = ?Y")
      proof(rule antisym)
        have "\<forall>a\<in>(\<lambda>s. c * P s) ` S. c * glb (P ` S) \<le> a"
        proof
          fix a assume "a \<in> (\<lambda>s. c * P s) ` S"
          from this obtain s
            where sin: "s \<in> S" and sim: "a = c * P s" by(blast)
          with pos and sound show "c * glb (P ` S) \<le> a"
            by(simp, blast intro:glb_lower)
        qed
        with sne show "?X \<le> ?Y" by(blast intro:glb_greatest)
        from pos show "?Y \<le> ?X"
          apply -
          apply(rule mult_div_mono_right)
            apply(blast intro:less_imp_le)
           apply(blast)
          apply(rule glb_greatest)
           using sne apply(blast)
          apply(rule ballI)
          apply(rule mult_div_mono_left)
            apply(simp)
           apply(simp)
          apply(simp)
          apply(rule glb_lower)
           apply(blast)
          apply(rule_tac x=0 in exI)
          apply(rule ballI)
          apply(erule imageE)+
          apply(simp)
          apply(rule mult_nonneg_nonneg)
           apply(blast intro:less_imp_le)
          using sound apply(blast)
          done
      qed
    qed
    ultimately show "c * ?T P s = ?T (\<lambda>s. c * P s) s"
      unfolding Exec_def wp_def Let_def split_def by(simp)
  qed
qed

lemma healthy_wp_Exec:
  "healthy (wp (Exec prog))"
  unfolding wp_def by(rule healthy_Exec)

lemma healthy_wlp_Exec:
  "healthy (wlp (Exec prog))"
  unfolding wlp_def by(rule healthy_Exec)

declare healthy_wlp_Exec[health] healthy_wp_Exec[health]

lemma sub_add_Exec:
  "\<lbrakk> no_fail \<top> m; empty_fail m \<rbrakk> \<Longrightarrow> sub_add (wp (Exec m))"
  apply(rule sub_addI)
  apply(simp add:Exec_def wp_def no_fail_def empty_fail_def)
  apply(case_tac "m s", simp add:Abort_def)
  apply(subgoal_tac "a \<noteq> {}", simp)
   prefer 2 apply(erule_tac x=s in allE) back apply(simp)
  apply(clarsimp)
  apply(rule glb_greatest)
   apply(rule nonempty_witness, blast)
  apply(clarsimp)
  apply(rule add_mono)
   apply(rule glb_lower)
    apply(rule imageI)
    apply(rule_tac x="(aa,ba)" in image_eqI, simp_all)[]
   apply(blast)
  apply(rule glb_lower)
   apply(rule imageI)
   apply(rule_tac x="(aa,ba)" in image_eqI, simp_all)[]
  apply(blast)
  done

section "Lifting Results"

text {*
  Given a Hoare triple on the underlying monad, we can derive a
  predicate entailment (wp rule) on the pGCL level.  The property
  @{text nontrivial} is a side-effect of the embedding, which could
  be relaxed with some more work.  XXX - fix this.
*}

lemma wlp_Exec:
  assumes valid: "\<lbrace>P\<rbrace> prog \<lbrace>\<lambda>r s. Q s\<rbrace>"
      and nontrivial: "\<exists>s. Q s"
  shows "\<guillemotleft>P\<guillemotright> \<tturnstile> wlp (Exec prog) \<guillemotleft>Q\<guillemotright>"
proof -
  from nontrivial obtain s' where Qs: "Q s'" by(blast)
  have Q_bound: "1 \<le> bound_of \<guillemotleft>Q\<guillemotright>" proof -
    from Qs have "1 = \<guillemotleft>Q\<guillemotright> s'" by(simp)
    also have "... \<le> bound_of \<guillemotleft>Q\<guillemotright>" by(blast)
    finally show ?thesis .
  qed

  have "\<And>s. P s \<Longrightarrow> 1 \<le> wlp (Exec prog) \<guillemotleft>Q\<guillemotright> s" proof -
    fix s assume Ps: "P s"
    obtain R and f where ev_prog: "prog s = (R,f)"
      by(case_tac "prog s", blast)
    with Ps and valid have QR: "\<forall>(r,s)\<in>R. Q s"
      unfolding valid_def by(simp)
    show "1 \<le> wlp (Exec prog) \<guillemotleft>Q\<guillemotright> s" (is ?X)
    proof(cases f)
      case True then show ?X
        unfolding Exec_def wlp_def ev_prog Abort_def
        by(simp add:Q_bound)
    next case False
      note nf = this
      show ?X
      proof(cases "R={}")
        case True with nf show ?X
          unfolding Exec_def wlp_def ev_prog Abort_def
          by(simp add:Q_bound)
      next
        case False note ne = this
        from this obtain S where Rim: "snd ` R = S" and Sne: "S \<noteq> {}"
          by(blast)
        with QR have "\<forall>s\<in>S. Q s" by(force)
        with Sne have "1 \<le> glb (\<guillemotleft>Q\<guillemotright> ` S)"
          by(force intro!:glb_greatest)
        with nf and ne show ?X
          unfolding Exec_def wlp_def ev_prog Abort_def
          by(simp add:Rim)
      qed
    qed
  qed

  with healthy_wlp_Exec show ?thesis  by(rule use_premise)
qed

lemma wp_Exec:
  assumes valid: "\<lbrace>P\<rbrace> prog \<lbrace>\<lambda>r s. Q s\<rbrace>"
      and nontrivial: "\<exists>s. Q s"
      and no_fail: "no_fail P prog"
  shows "\<guillemotleft>P\<guillemotright> \<tturnstile> wp (Exec prog) \<guillemotleft>Q\<guillemotright>"
proof -
  from nontrivial obtain s' where Qs: "Q s'" by(blast)
  have Q_bound: "1 \<le> bound_of \<guillemotleft>Q\<guillemotright>" proof -
    from Qs have "1 = \<guillemotleft>Q\<guillemotright> s'" by(simp)
    also have "... \<le> bound_of \<guillemotleft>Q\<guillemotright>" by(blast)
    finally show ?thesis .
  qed

  have "\<And>s. P s \<Longrightarrow> 1 \<le> wp (Exec prog) \<guillemotleft>Q\<guillemotright> s" proof -
    fix s assume Ps: "P s"
    obtain R and f where ev_prog: "prog s = (R,f)"
      by(case_tac "prog s", blast)
    with Ps and valid have QR: "\<forall>(r,s)\<in>R. Q s"
      unfolding valid_def by(simp)
    from Ps and ev_prog and no_fail have nf: "\<not>f"
      unfolding no_fail_def by(force)
    show "1 \<le> wp (Exec prog) \<guillemotleft>Q\<guillemotright> s" (is ?X)
    proof(cases "R={}")
      case True with nf show ?X
        unfolding Exec_def wp_def ev_prog Abort_def by(simp add:Q_bound)
    next
      case False note ne = this
      from this obtain S where Rim: "snd ` R = S" and Sne: "S \<noteq> {}"
        by(blast)
      with QR have "\<forall>s\<in>S. Q s" by(force)
      with Sne have "1 \<le> glb (\<guillemotleft>Q\<guillemotright> ` S)"
        by(force intro!:glb_greatest)
      with nf and ne show ?X
        unfolding Exec_def wp_def ev_prog by(simp add:Rim)
    qed
  qed

  with healthy_wp_Exec show ?thesis  by(rule use_premise)
qed

lemma wp_Exec_ef:
  assumes valid: "\<lbrace>P\<rbrace> prog \<lbrace>\<lambda>r s. Q s\<rbrace>"
      and no_fail: "no_fail P prog"
      and empty_fail: "empty_fail prog"
  shows "\<guillemotleft>P\<guillemotright> \<tturnstile> wp (Exec prog) \<guillemotleft>Q\<guillemotright>"
proof(rule use_premise)
  show "healthy (wp (Exec prog))" by(rule healthy_wp_Exec)

  fix s assume Ps: "P s"
  obtain R and f where ev_prog: "prog s = (R,f)"
    by(case_tac "prog s", blast)
  with Ps and valid have QR: "\<forall>(r,s)\<in>R. Q s"
    unfolding valid_def by(simp)
  from Ps and ev_prog and no_fail have nf: "\<not>f"
    unfolding no_fail_def by(force)
  with empty_fail have ne: "R \<noteq> {}"
    unfolding empty_fail_def
    apply(erule_tac x=s in allE)
    apply(auto simp:ev_prog)
    done
  from this
  obtain r and s' where "(r,s') \<in> R" by(auto)
  with Ps and valid
  have Qs: "Q s'" unfolding valid_def
    apply(erule_tac x=s in allE)
    apply(simp add:ev_prog)
    by(blast)
  have Q_bound: "1 \<le> bound_of \<guillemotleft>Q\<guillemotright>" proof -
    from Qs have "1 = \<guillemotleft>Q\<guillemotright> s'" by(simp)
    also have "... \<le> bound_of \<guillemotleft>Q\<guillemotright>" by(blast)
    finally show ?thesis .
  qed

  from ne obtain S where Rim: "snd ` R = S" and Sne: "S \<noteq> {}"
    by(blast)
  with QR have "\<forall>s\<in>S. Q s" by(force)
  with Sne have "1 \<le> glb (\<guillemotleft>Q\<guillemotright> ` S)"
    by(force intro!:glb_greatest)
  with nf and ne show "1 \<le> wp (Exec prog) \<guillemotleft>Q\<guillemotright> s"
    unfolding Exec_def wp_def ev_prog by(simp add:Rim)
qed

text {*
  The difference between wp and wlp is exactly
  @{term no_fail} on the monadic level.
*}

lemma wlp_wp_Exec:
  "\<lbrakk> \<guillemotleft>P\<guillemotright> \<tturnstile> wlp (Exec prog) \<guillemotleft>Q\<guillemotright>; no_fail P prog \<rbrakk> \<Longrightarrow> \<guillemotleft>P\<guillemotright> \<tturnstile> wp (Exec prog) \<guillemotleft>Q\<guillemotright>"
  unfolding Exec_def wp_def wlp_def
  apply(simp)
  apply(rule le_funI)
  apply(drule_tac x=x in le_funD)
  apply(case_tac "prog x", simp)
  unfolding Abort_def apply(simp)
  unfolding no_fail_def
  apply(case_tac "a={}", simp_all)
   apply(case_tac b, simp_all)
   apply(erule_tac x=x in allE)
   apply(simp)
  apply(case_tac b, simp_all)
  apply(erule_tac x=x in allE)
  apply(simp)
  done

section "Useful Results"

lemma Exec_modify:
  "Exec (modify f) = Apply f"
  unfolding Exec_def Apply_def modify_def get_def put_def bind_def
  by(simp)

lemma Exec_gets:
  "Exec (gets f >>= M) = Bind f (\<lambda>x. Exec (M x))"
  unfolding Exec_def Bind_def gets_def get_def return_def
  apply(rule ext)
  apply(simp add:Bind_def bind_def)
  done

text {* Used to reason about preserved values e.g. in partitioned
        state spaces. *}
lemma wp_Exec_const_cong:
  assumes wpM: "\<lbrace>P\<rbrace> M \<lbrace>\<lambda>_. Q\<rbrace>"
      and PS:  "\<And>s. P s \<Longrightarrow> S s = k"
      and QR:  "\<And>s. Q s \<Longrightarrow> R s = k"
      and nfM: "no_fail P M"
      and efM: "empty_fail M"
      and sS:  "sound S"
      and bS:  "bounded_by 1 S"
      and sR:  "sound R"
      and bR:  "bounded_by 1 R"
  shows "\<guillemotleft>P\<guillemotright> && S \<tturnstile> wp (Exec M) R"
proof -
  have "feasible (wp (Exec M))"
    by(blast intro:healthy_wp_Exec healthy_feasibleD)
  moreover
  from sS have "nneg S" by(blast)
  moreover
  from sR have "nneg R" by(blast)
  moreover
  note bS and bR
  moreover
  have "\<And>s. P s \<Longrightarrow> S s \<le> wp (Exec M) R s"
  proof -
    fix s assume Ps: "P s"
    obtain sa and f where split_Ms[simp]: "M s = (sa,f)"
      by(case_tac "M s", force)
    from Ps and PS have "S s = k" by(simp)
    also {
      from Ps and nfM have "\<not>f"
        unfolding no_fail_def by(force)
      moreover
      with efM have "sa \<noteq> {}"
        unfolding empty_fail_def by(erule_tac x=s in allE, simp)
      moreover {
        from Ps and wpM have "\<forall>(a,s)\<in>(fst (M s)). Q s"
          unfolding valid_def by(simp)
        with QR have "\<forall>(a,s)\<in>(fst (M s)). R s = k"
          by(blast)
      }
      ultimately
      have "k \<le> wp (Exec M) R s"
        by(force simp:Exec_def wp_def
                 intro!:glb_greatest nonempty_witness)
    }
    finally show "S s \<le> wp (Exec M) R s" .
  qed
  ultimately
  show ?thesis by(rule_tac R=R in entails_pconj_assumption)
qed

lemma wp_Exec_nofail_unfold:
  "\<lbrakk> no_fail \<top> M; empty_fail M \<rbrakk> \<Longrightarrow>
   wp (Exec M) R = (\<lambda>s. glb (R ` snd ` fst (M s)))"
  unfolding no_fail_def empty_fail_def Exec_def wp_def
  by(simp add:split_def)

lemma wp_Exec_nofail_unfold_guard:
  "\<lbrakk> no_fail G M; empty_fail M; G s \<rbrakk> \<Longrightarrow>
   wp (Exec M) R s = glb (R ` snd ` fst (M s))"
  unfolding no_fail_def empty_fail_def Exec_def wp_def
  by(simp add:split_def Let_def, blast)

text {* Exec is sub-conjunctive if M fails whenever its result is empty.
  This result is important in data refinement. *}
lemma sub_conj_Exec:
  fixes M::"('s,'a) nondet_monad"
  assumes efM: "empty_fail M"
  shows "sub_conj (wp (Exec M))"
proof(rule sub_conjI, rule le_funI)
  fix P::"'s \<Rightarrow> real" and Q::"'s \<Rightarrow> real" and s::'s
  assume sP: "sound P" and sQ: "sound Q"

  obtain f and sa where [simp]: "M s = (sa,f)" by(case_tac "M s", blast)

  show "(wp (Exec M) P && wp (Exec M) Q) s \<le> wp (Exec M) (P && Q) s"
  proof(cases)
    assume "f \<or> sa={}"
    moreover from efM have "fst (M s) = {} \<longrightarrow> snd (M s)"
      unfolding empty_fail_def by(blast)
    ultimately have "f" by(simp, blast)
    thus ?thesis
      unfolding wp_def Exec_def exp_conj_def Abort_def by(simp)
  next
    assume "\<not>(f \<or> sa={})"
    hence "\<not>f" and ne:"sa \<noteq> {}" by(blast)+
    thus ?thesis unfolding wp_def Exec_def exp_conj_def
    proof(simp)
      show "glb (P ` snd ` sa) .& glb (Q ` snd ` sa) \<le>
            glb ((\<lambda>s. P s .& Q s) ` snd ` sa)"
      proof(rule glb_greatest, simp_all)
        from ne obtain x where "x \<in> sa" by(blast)
        hence "snd x \<in> snd ` sa" by(blast)
        hence "(P (snd x) .& Q (snd x)) \<in> (\<lambda>s. P s .& Q s) ` snd ` sa"
          by(blast)
        thus "\<exists>a. a \<in> (\<lambda>s. P s .& Q s) ` snd ` sa" by(blast)
      next
        show "\<forall>x\<in>sa. glb (P ` snd ` sa) .& glb (Q ` snd ` sa) \<le>
                     P (snd x) .& Q (snd x)"
        proof(clarsimp, rule pconj_mono)
          fix a and b
          assume X: "(a,b) \<in> sa"
          hence "P b \<in> P ` snd ` sa" by(simp add:rev_image_eqI)
          with sP show "glb (P ` snd ` sa) \<le> P b" by(blast intro:glb_lower)
          from X have "Q b \<in> Q ` snd ` sa" by(simp add:rev_image_eqI)
          with sQ show "glb (Q ` snd ` sa) \<le> Q b" by(blast intro:glb_lower)
        qed
      qed
    qed
  qed
qed

text {* Given correspondence of two monads, we have probabilistic
  correspondence of their lifted analogues. *}
lemma corres_pcorres:
  fixes M::"('a,'r) nondet_monad" and M'::"('b,'r) nondet_monad"
    and \<phi>::"'b \<Rightarrow> 'a" and Pre::"'a \<Rightarrow> bool"
  -- "Assume correspondence with state relation construction from \<phi>,
      no failure, any result relation, and matching preconditions."
  assumes corres: "corres_underlying {(s,s'). s = \<phi> s'} True rrel
                                     Pre (Pre o \<phi>) M M'"
  -- "For simplicity, insist on totality."
      and nfM: "no_fail Pre M" and efM: "empty_fail M" and efM': "empty_fail M'"
  -- "The specification is deterministic, on the parts that we case about."
      and dspec: "\<And>s. Pre (\<phi> s) \<Longrightarrow> \<exists>s'. fst (M (\<phi> s)) = {s'}"
  shows "pcorres \<phi> (Pre o \<phi>) (Exec M) (Exec M')"
proof(rule pcorresI2, simp_all add:healthy_wp_Exec)
  fix Q::"'a \<Rightarrow> real" and s::'b
  assume sQ: "sound Q" and bQ: "bounded_by 1 Q" and p: "Pre (\<phi> s)"

  from corres have nfM': "no_fail (Pre o \<phi>) M'"
    -- "Totality for M follows from corres"
    by(simp add:corres_underlying_def no_fail_def)

  have "glb (Q ` snd ` fst (M (\<phi> s))) = glb ((Q \<circ> \<phi>) ` snd ` fst (M' s))"
       (is "?X = ?Y")
  proof(rule antisym)
      from p and nfM' have "\<not> snd (M' s)" by(simp add:no_fail_def)
      with efM' have "fst (M' s) \<noteq> {}" unfolding empty_fail_def by(blast)
      from this obtain y where yin: "y \<in> fst (M' s)" by(blast)

      from sQ have "\<forall>x\<in>Q ` snd ` fst (M (\<phi> s)). 0 \<le> x" by(blast)
      hence bM: "\<exists>b. \<forall>y\<in>Q ` snd ` fst (M (\<phi> s)). b \<le> y" by(blast)

    show "?X \<le> ?Y"
    proof(rule glb_under_glb)
      from yin show "\<exists>y. y \<in> ((Q o \<phi>) ` snd ` fst (M' s))" by(blast)

      from bM show "\<exists>b. \<forall>y\<in>Q ` snd ` fst (M (\<phi> s)). b \<le> y" .

      show "\<forall>x\<in>(Q o \<phi>) ` snd ` fst (M' s). \<exists>y\<in>Q ` snd ` fst (M (\<phi> s)). y \<le> x"
      proof
        fix x
        assume "x \<in> (Q o \<phi>) ` snd ` fst (M' s)"
        from this obtain s' where "s' \<in> fst (M' s)"
                              and xrw: "x = (Q o \<phi>) (snd s')"
          by(blast)
        with corres and p obtain t where tin: "t \<in> fst (M (\<phi> s))"
                                     and rel: "\<phi> (snd s') = (snd t)"
          unfolding corres_underlying_def by(force)
        hence "Q (snd t) \<in> Q ` snd ` fst (M (\<phi> s))" by(blast)
        moreover from rel and xrw have "Q (snd t) \<le> x" by(simp)
        ultimately show "\<exists>y\<in>Q ` snd ` fst (M (\<phi> s)). y \<le> x" by(blast)
      qed
    qed

    from p and dspec
    obtain s' where Mrw: "fst (M (\<phi> s)) = {s'}" by(force)

    from sQ have "\<forall>x\<in>(Q o \<phi>) ` snd ` fst (M' s). 0 \<le> x" by(force)
    hence "\<exists>b. \<forall>y\<in>(Q o \<phi>) ` snd ` fst (M' s). b \<le> y" by(blast)
    with yin have "?Y \<le> (Q o \<phi>) (snd y)" by(blast intro:glb_lower)
    also {
      from yin and corres and p
      obtain x where xin: "x \<in> fst (M (\<phi> s))" and yrw: "\<phi> (snd y) = (snd x)"
        unfolding corres_underlying_def by(force)
      with Mrw have "x = s'" by(simp)
      with yrw have "(Q o \<phi>) (snd y) = Q (snd s')" by(simp)
    }
    also from Mrw have "Q (snd s') = glb (Q ` snd ` fst (M (\<phi> s)))" by(simp)
    finally show "?Y \<le> ?X" .
  qed
  moreover
  from p have "(Pre o \<phi>) s" unfolding o_def .
  moreover note nfM nfM' efM efM' p
  ultimately show "wp (Exec M) Q (\<phi> s) = wp (Exec M') (Q o \<phi>) s"
    by(simp add:wp_Exec_nofail_unfold_guard)
qed

section "The Reverse Embedding"

text {*
  By making the philosophical commitment that `true with probability 0'
  is `false', one can embed in the reverse direction also: constructing
  a monad from the implied state transformer, by including all final
  states with probability >0 from a given starting state.
*}

subsection "Forward Transformers"

text {*
  The forward transformer of a program: mapping distributions on initial
  states to distributions on final states.
*}
definition
  prog_fw :: "'s prog \<Rightarrow> ('s \<Rightarrow> real) \<Rightarrow> 's \<Rightarrow> real"
where
  "prog_fw prog \<equiv> \<lambda>P.
     \<lambda>s_final. \<Sum>s_initial\<in>UNIV.  P s_initial * wp prog \<guillemotleft>op = s_final\<guillemotright> s_initial"

lemma Skip_fw:
  assumes "finite (UNIV::'s set)"
  shows "prog_fw (Skip::'s prog) = (\<lambda>P. P)"
  apply(rule ext)
  unfolding prog_fw_def wp_eval
  apply(rule ext)
  unfolding embed_bool_def
  apply(subgoal_tac "\<And>a P. a * (if P then 1::real else 0) =
                               (if P then a else 0)")
   prefer 2
   apply(simp)
  apply(simp)
  apply(subst eq_commute)
  using assms apply(simp add: setsum_delta)
  done

lemma Abort_fw:
  fixes P::"'s \<Rightarrow> real"
  shows "prog_fw Abort = (\<lambda>P s. 0)"
  apply(rule ext)
  unfolding prog_fw_def wp_eval
  by(simp)

subsection "The Embedding Operator"

text {* Map to the set of states reachable with P > 0. *}
term "Collect (\<lambda>t. 0 < wp prog \<guillemotleft>op = s\<guillemotright> t)"
definition
  PDo :: "'s prog \<Rightarrow> ('s,unit) nondet_monad"
where
  "PDo prog \<equiv>
   \<lambda>s. (Pair () ` {t. 0 < wp prog \<guillemotleft>op = t\<guillemotright> s}, False)"

end
