(*
 * Copyright (C) 2013 National ICT Australia Ltd ('NICTA')
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Author: David Cock - David.Cock@nicta.com.au *)

signature pVCG_sig =
sig
        val setup:    theory -> theory;
        val pVCG_tac: Proof.context -> int -> tactic;
end;

structure pVCG =
struct

structure HealthRules = Named_Thms
  (val name = @{binding "health"}
   val description = "Theorems to automatically establish healthiness")

structure CoreWPRules = Named_Thms
  (val name = @{binding "wp_core"}
   val description = "Core wp rules, for evaluating primitive terms")

structure UserWPRules = Named_Thms
  (val name = @{binding "wp"}
   val description = "User-supplied wp rules")

structure UserWLPRules = Named_Thms
  (val name = @{binding "wlp"}
   val description = "User-supplied wlp rules")

(* Match soundness goals on non-schematic expectations *)
fun m_sound t =
  case t of
    @{term "Trueprop"} $ (Const ("Expectations.sound",_) $ Var _) => false
  | @{term "Trueprop"} $ (Const ("Expectations.sound",_) $ _) => true
  | _ => false

(* Match healthiness goals on non-schematic expectations *)
fun m_healthy t =
  case t of
    @{term "Trueprop"} $ (Const ("Transformers.healthy",_) $ Var _) => false
  | @{term "Trueprop"} $ (Const ("Transformers.healthy",_) $ _) => true
  | _ => false

val sound_goal_tac =
  resolve_tac @{thms sound_intros}
fun maybe_sound_tac (g,i) =
  if m_sound g then sound_goal_tac i else no_tac
val sound_tac = SUBGOAL maybe_sound_tac
val all_sound_tac = ALLGOALS sound_tac

fun healthy_goal_tac ctxt =
  resolve_tac (HealthRules.get ctxt)
fun maybe_healthy_tac ctxt (g,i) =
  if m_healthy g then healthy_goal_tac ctxt i else no_tac
fun healthy_tac ctxt = SUBGOAL (maybe_healthy_tac ctxt)

(* Attempt to solve a side-condition (healthiness or soundness) *)
fun sidegoal_tac ctxt =
  REPEAT_ALL_NEW (CHANGED o (healthy_tac ctxt ORELSE'
                  sound_tac ORELSE'
                  asm_full_simp_tac (simpset_of ctxt)))

fun wp_start ctxt =
  (rtac @{thm wp_weaken_pre} ORELSE' rtac @{thm wlp_weaken_pre})
  THEN_ALL_NEW ((sidegoal_tac ctxt) ORELSE' (K all_tac))

fun user_wp_rules ctxt = UserWPRules.get ctxt

fun mod_user_wp_rules ctxt =
  map (fn thm => thm RS @{thm wp_strengthen_post}) (user_wp_rules ctxt)

fun user_wlp_rules ctxt = UserWLPRules.get ctxt

fun mod_user_wlp_rules ctxt =
  map (fn thm => thm RS @{thm wlp_strengthen_post}) (user_wlp_rules ctxt)

fun core_rules ctxt = CoreWPRules.get ctxt

(* Prefer raw user rules, then lifted user rules, then core (safe) rules *)
fun wp_step ctxt =
  (resolve_tac (user_wp_rules ctxt) ORELSE'
   resolve_tac (mod_user_wp_rules ctxt) ORELSE'
   resolve_tac (user_wlp_rules ctxt) ORELSE'
   resolve_tac (mod_user_wlp_rules ctxt) ORELSE'
   resolve_tac (core_rules ctxt)) THEN_ALL_NEW
  (sidegoal_tac ctxt ORELSE' K all_tac)

fun pVCG_tac ctxt =
  (wp_start ctxt THEN'
   REPEAT_ALL_NEW (wp_step ctxt)) THEN_ALL_NEW
   (TRY o sidegoal_tac ctxt)

val setup = HealthRules.setup #> CoreWPRules.setup #>
            UserWPRules.setup #> UserWLPRules.setup

end;

structure pVCGInst : pVCG_sig = pVCG;
